Type: Package
Package: rmapshaper
Title: Client for 'mapshaper' for 'Geospatial' Operations
Version: 0.6.1
Authors@R: 
    c(person(given = "Andy",
             family = "Teucher",
             role = c("aut", "cre"),
             comment = c(ORCID = "0000-0002-7840-692X"),
             email = "andy.teucher@gmail.com"),
      person(given = "Kenton",
             family = "Russell",
             role = "aut",
             email = "kent.russell@timelyportfolio.com",
             comment = "JavaScript support"),
      person(given = "Matthew",
             family = "Bloch",
             role = "cph",
             comment = "mapshaper Javascript library"))
Description: Edit and simplify 'geojson', 'Spatial', and 'sf'
    objects.  This is wrapper around the 'mapshaper' 'JavaScript' library
    by Matthew Bloch <https://github.com/mbloch/mapshaper/> to perform
    topologically-aware polygon simplification, as well as other
    operations such as clipping, erasing, dissolving, and converting
    'multi-part' to 'single-part' geometries.
License: MIT + file LICENSE
URL: https://github.com/ateucher/rmapshaper,
        http://andyteucher.ca/rmapshaper/
BugReports: https://github.com/ateucher/rmapshaper/issues
Depends: R (>= 4.1.0)
Imports: methods, geojsonsf (>= 2.0.5), jsonify (>= 1.2.3), readr (>=
        2.1.0), sf (>= 1.0.0), sp (>= 1.4-0), V8 (>= 6.0.0)
Suggests: knitr, magrittr, rmarkdown, testthat (>= 3.0.0), jsonlite,
        covr, units, withr
VignetteBuilder: knitr
Encoding: UTF-8
Config/testthat/edition: 3
Config/testthat/parallel: TRUE
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-06 23:55:37 UTC; andy
Author: Andy Teucher [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-7840-692X>),
  Kenton Russell [aut] (JavaScript support),
  Matthew Bloch [cph] (mapshaper Javascript library)
Maintainer: Andy Teucher <andy.teucher@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-09 14:41:08 UTC
Built: R 4.5.2; ; 2026-05-09 17:02:14 UTC; unix

Type: Package
Package: rmapshaper
Title: Client for 'mapshaper' for 'Geospatial' Operations
Version: 0.6.1
Authors@R: 
    c(person(given = "Andy",
             family = "Teucher",
             role = c("aut", "cre"),
             comment = c(ORCID = "0000-0002-7840-692X"),
             email = "andy.teucher@gmail.com"),
      person(given = "Kenton",
             family = "Russell",
             role = "aut",
             email = "kent.russell@timelyportfolio.com",
             comment = "JavaScript support"),
      person(given = "Matthew",
             family = "Bloch",
             role = "cph",
             comment = "mapshaper Javascript library"))
Description: Edit and simplify 'geojson', 'Spatial', and 'sf'
    objects.  This is wrapper around the 'mapshaper' 'JavaScript' library
    by Matthew Bloch <https://github.com/mbloch/mapshaper/> to perform
    topologically-aware polygon simplification, as well as other
    operations such as clipping, erasing, dissolving, and converting
    'multi-part' to 'single-part' geometries.
License: MIT + file LICENSE
URL: https://github.com/ateucher/rmapshaper,
        http://andyteucher.ca/rmapshaper/
BugReports: https://github.com/ateucher/rmapshaper/issues
Depends: R (>= 4.1.0)
Imports: methods, geojsonsf (>= 2.0.5), jsonify (>= 1.2.3), readr (>=
        2.1.0), sf (>= 1.0.0), sp (>= 1.4-0), V8 (>= 6.0.0)
Suggests: knitr, magrittr, rmarkdown, testthat (>= 3.0.0), jsonlite,
        covr, units, withr
VignetteBuilder: knitr
Encoding: UTF-8
Config/testthat/edition: 3
Config/testthat/parallel: TRUE
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-06 23:55:37 UTC; andy
Author: Andy Teucher [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-7840-692X>),
  Kenton Russell [aut] (JavaScript support),
  Matthew Bloch [cph] (mapshaper Javascript library)
Maintainer: Andy Teucher <andy.teucher@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-09 14:41:08 UTC
Built: R 4.6.0; ; 2026-05-09 17:51:45 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(curly_brace_na,Spatial)
S3method(curly_brace_na,data.frame)
S3method(curly_brace_na,sf)
S3method(drop_null_geometries,character)
S3method(drop_null_geometries,json)
S3method(ms_clip,SpatialLines)
S3method(ms_clip,SpatialPoints)
S3method(ms_clip,SpatialPolygons)
S3method(ms_clip,character)
S3method(ms_clip,json)
S3method(ms_clip,sf)
S3method(ms_clip,sfc)
S3method(ms_dissolve,SpatialPoints)
S3method(ms_dissolve,SpatialPolygons)
S3method(ms_dissolve,character)
S3method(ms_dissolve,json)
S3method(ms_dissolve,sf)
S3method(ms_dissolve,sfc)
S3method(ms_erase,SpatialLines)
S3method(ms_erase,SpatialPoints)
S3method(ms_erase,SpatialPolygons)
S3method(ms_erase,character)
S3method(ms_erase,json)
S3method(ms_erase,sf)
S3method(ms_erase,sfc)
S3method(ms_explode,SpatialLines)
S3method(ms_explode,SpatialPolygons)
S3method(ms_explode,character)
S3method(ms_explode,json)
S3method(ms_explode,sf)
S3method(ms_explode,sfc)
S3method(ms_filter_fields,SpatialLinesDataFrame)
S3method(ms_filter_fields,SpatialPointsDataFrame)
S3method(ms_filter_fields,SpatialPolygonsDataFrame)
S3method(ms_filter_fields,character)
S3method(ms_filter_fields,json)
S3method(ms_filter_fields,sf)
S3method(ms_filter_islands,SpatialPolygons)
S3method(ms_filter_islands,character)
S3method(ms_filter_islands,json)
S3method(ms_filter_islands,sf)
S3method(ms_filter_islands,sfc)
S3method(ms_innerlines,SpatialPolygons)
S3method(ms_innerlines,character)
S3method(ms_innerlines,json)
S3method(ms_innerlines,sf)
S3method(ms_innerlines,sfc)
S3method(ms_lines,SpatialPolygons)
S3method(ms_lines,character)
S3method(ms_lines,json)
S3method(ms_lines,sf)
S3method(ms_lines,sfc)
S3method(ms_points,SpatialPolygons)
S3method(ms_points,character)
S3method(ms_points,json)
S3method(ms_points,sf)
S3method(ms_points,sfc)
S3method(ms_simplify,SpatialLines)
S3method(ms_simplify,SpatialPolygons)
S3method(ms_simplify,character)
S3method(ms_simplify,json)
S3method(ms_simplify,sf)
S3method(ms_simplify,sfc)
export(apply_mapshaper_commands)
export(check_sys_mapshaper)
export(drop_null_geometries)
export(ms_clip)
export(ms_dissolve)
export(ms_erase)
export(ms_explode)
export(ms_filter_fields)
export(ms_filter_islands)
export(ms_innerlines)
export(ms_lines)
export(ms_points)
export(ms_simplify)
importClassesFrom(sp,SpatialLines)
importClassesFrom(sp,SpatialPoints)
importFrom(V8,v8)
importFrom(methods,.hasSlot)
importFrom(methods,as)
importFrom(methods,is)
importFrom(methods,slot)

<!-- README.md is generated from README.Rmd. Please edit that file -->

<!-- badges: start -->

[![CRAN_Status_Badge](http://www.r-pkg.org/badges/version/rmapshaper)](https://cran.r-project.org/package=rmapshaper)
[![CRAN Downloads per
month](http://cranlogs.r-pkg.org/badges/rmapshaper)](https://cran.r-project.org/package=rmapshaper)
[![CRAN total
downloads](http://cranlogs.r-pkg.org/badges/grand-total/rmapshaper?color=lightgrey)](https://cran.r-project.org/package=rmapshaper)
[![R-CMD-check](https://github.com/ateucher/rmapshaper/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/ateucher/rmapshaper/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/ateucher/rmapshaper/graph/badge.svg)](https://app.codecov.io/gh/ateucher/rmapshaper)
<!-- badges: end -->

# rmapshaper <a href="http://andyteucher.ca/rmapshaper/"><img src="man/figures/logo.png" align="right" height="139" alt="rmapshaper website" /></a>

An R package providing access to the awesome
[mapshaper](https://github.com/mbloch/mapshaper/) tool by Matthew Bloch,
which has both a [Node.js command-line
tool](https://github.com/mbloch/mapshaper/wiki/Introduction-to-the-Command-Line-Tool)
as well as an [interactive web tool](https://mapshaper.org/).

I started this package so that I could use mapshaper’s
[Visvalingam](https://bost.ocks.org/mike/simplify/) simplification
method in R. There is, as far as I know, no other R package that
performs topologically-aware multi-polygon simplification. (This means
that shared boundaries between adjacent polygons are always kept intact,
with no gaps or overlaps, even at high levels of simplification).

But mapshaper does much more than simplification, so I am working on
wrapping most of the core functionality of mapshaper into R functions.

So far, `rmapshaper` provides the following functions:

- `ms_simplify` - simplify polygons or lines
- `ms_clip` - clip an area out of a layer using a polygon layer or a
  bounding box. Works on polygons, lines, and points
- `ms_erase` - erase an area from a layer using a polygon layer or a
  bounding box. Works on polygons, lines, and points
- `ms_dissolve` - aggregate polygon features, optionally specifying a
  field to aggregate on. If no field is specified, will merge all
  polygons into one.
- `ms_explode` - convert multipart shapes to single part. Works with
  polygons, lines, and points in geojson format, but currently only with
  polygons and lines in the `Spatial` classes (not `SpatialMultiPoints`
  and `SpatialMultiPointsDataFrame`).
- `ms_lines` - convert polygons to topological boundaries (lines)
- `ms_innerlines` - convert polygons to shared inner boundaries (lines)
- `ms_points` - create points from a polygon layer
- `ms_filter_fields` - Remove fields from the attributes
- `ms_filter_islands` - Remove small detached polygons

If you run into any bugs or have any feature requests, please file an
[issue](https://github.com/ateucher/rmapshaper/issues/)

### Installation

`rmapshaper` is on CRAN. Install the current version with:

``` r
install.packages("rmapshaper")
```

You can install the development version from github with `remotes`:

``` r
## install.packages("remotes")
library(remotes)
install_github("ateucher/rmapshaper")
```

### Usage

rmapshaper works with `sf` objects as well as geojson strings (character
objects of class `geo_json`). It also works with `Spatial` classes from
the `sp` package, though this will likely be retired in the future;
users are encouraged to use the more modern `sf` package.

We will use the `nc.gpkg` file (North Carolina county boundaries) from
the `sf` package and read it in as an `sf` object:

``` r
library(rmapshaper)
library(sf)
#> Linking to GEOS 3.13.0, GDAL 3.8.5, PROJ 9.5.1; sf_use_s2() is TRUE

file <- system.file("gpkg/nc.gpkg", package = "sf")
nc_sf <- read_sf(file)
```

Plot the original:

``` r
plot(nc_sf["FIPS"])
```

![](man/figures/unnamed-chunk-3-1.png)<!-- -->

Now simplify using default parameters, then plot the simplified North
Carolina counties:

``` r
nc_simp <- ms_simplify(nc_sf)
plot(nc_simp["FIPS"])
```

![](man/figures/unnamed-chunk-4-1.png)<!-- -->

You can see that even at very high levels of simplification, the
mapshaper simplification algorithm preserves the topology, including
shared boundaries. The `keep` parameter specifies what proportion of
vertices to keep:

``` r
nc_very_simp <- ms_simplify(nc_sf, keep = 0.001)
plot(nc_very_simp["FIPS"])
```

![](man/figures/unnamed-chunk-5-1.png)<!-- -->

Compare this to the output using `sf::st_simplify`, where overlaps and
gaps are evident:

``` r
nc_stsimp <- st_simplify(nc_sf, preserveTopology = TRUE, dTolerance = 10000) # dTolerance specified in meters
plot(nc_stsimp["FIPS"])
```

![](man/figures/unnamed-chunk-6-1.png)<!-- -->

This time we’ll demonstrate the `ms_innerlines` function:

``` r
nc_sf_innerlines <- ms_innerlines(nc_sf)
plot(nc_sf_innerlines)
```

![](man/figures/unnamed-chunk-7-1.png)<!-- -->

All of the functions are quite fast with `geojson` character objects.
They are slower with the `sf` and `Spatial` classes due to internal
conversion to/from json. If you are going to do multiple operations on
large `sf` objects, it’s recommended to first convert to json using
`geojsonsf::sf_geojson()`, or `geojsonio::geojson_json()`. All of the
functions have the input object as the first argument, and return the
same class of object as the input. As such, they can be chained
together. For a totally contrived example, using `nc_sf` as created
above:

``` r
library(geojsonsf)
library(rmapshaper)
library(sf)

## First convert 'states' dataframe from geojsonsf pkg to json

nc_sf |>
  sf_geojson() |>
  ms_erase(bbox = c(-80, 35, -79, 35.5)) |> # Cut a big hole in the middle
  ms_dissolve() |> # Dissolve county borders
  ms_simplify(keep_shapes = TRUE, explode = TRUE) |> # Simplify polygon
  geojson_sf() |> # Convert to sf object
  plot(col = "blue") # plot
```

![](man/figures/unnamed-chunk-8-1.png)<!-- -->

### Using the system mapshaper

Sometimes if you are dealing with a very large spatial object in R,
`rmapshaper` functions will take a very long time or not work at all. As
of version `0.4.0`, you can make use of the system `mapshaper` library
if you have it installed. This will allow you to work with very large
spatial objects.

First make sure you have mapshaper installed:

``` r
check_sys_mapshaper()
#> mapshaper version 0.6.113 is installed and on your PATH
#>                     mapshaper-xl 
#> "/opt/homebrew/bin/mapshaper-xl"
```

If you get an error, you will need to install mapshaper. First install
node (<https://nodejs.org/en>) and then install mapshaper in a command
prompt with:

    $ npm install -g mapshaper

Then you can use the `sys` argument in any rmapshaper function:

``` r
nc_simp_internal <- ms_simplify(nc_sf)
nc_simp_sys <- ms_simplify(nc_sf, sys = TRUE, sys_mem = 8) #sys_mem specifies the amount of memory to use in Gb.  It defaults to 8 if omitted.

par(mfrow = c(1, 2))
plot(st_geometry(nc_simp_internal), main = "internal")
plot(st_geometry(nc_simp_sys), main = "system")
```

![](man/figures/unnamed-chunk-10-1.png)<!-- -->

### Thanks

This package uses the [V8](https://cran.r-project.org/package=V8)
package to provide an environment in which to run mapshaper’s javascript
code in R. It relies heavily on all of the great spatial packages that
already exist (especially `sf`), and the `geojsonio` and the `geojsonsf`
packages for converting between `geojson`, `sf` and `Spatial` object.

Thanks to [timelyportfolio](https://github.com/timelyportfolio) for
helping me wrangle the javascript to the point where it works in V8. He
also wrote the [mapshaper
htmlwidget](https://github.com/timelyportfolio/mapshaper_htmlwidget),
which provides access to the mapshaper web interface, right in your R
session. We have plans to combine the two in the future.

### Code of Conduct

Please note that this project is released with a [Contributor Code of
Conduct](https://github.com/ateucher/rmapshaper/blob/master/CONDUCT.md).
By participating in this project you agree to abide by its terms.

### LICENSE

MIT

The file `mapshaper-browserify.min.js` is generated from the `mapshaper` node module (see `LICENSE` file) using the `browserify` node module, as suggested in the [V8 vignette](https://cran.r-project.org/web/packages/V8/vignettes/npm.html).

The license for `mapshaper` and its dependencies are listed in `node_packages.csv`.

## Steps to Update Bundled Mapshaper

### 1. Install required npm packages

```bash
npm install -g browserify uglify-js
npm install mapshaper
```

Note: The modules to include in the `browserify` call can be found in the mapshaper package.json dependencies: https://github.com/mbloch/mapshaper/blob/master/package.json

### 2. Generate the browserified bundle

```bash
echo "global.mapshaper = require('mapshaper');" > in.js
browserify -r sync-request -r mproj -r buffer -r iconv-lite -r fs -r flatbush -r rw -r path -r kdbush -r @tmcw/togeojson -r @placemarkio/tokml -r idb-keyval in.js -o inst/mapshaper/mapshaper-browserify.js
rm in.js
```

### 3. Apply V8 compatibility modifications

To make this work in V8, which does not have all of the capability of Node, three modifications are required to `inst/mapshaper/mapshaper-browserify.js`:

**Note:** Line numbers will vary with each mapshaper version. Use grep/search to locate the functions.

#### 3a. Fix setTimeout and setImmediate in reduceAsync

Because `setTimeout` and `setImmediate` are not available in the V8 engine, the `reduceAsync` function must be modified.

Search for: `function reduceAsync(arr, memo, iter, done)`

In the current version (0.6.113), this is around line 5120.

Thanks to [@timelyportfolio](https://github.com/timelyportfolio) for figuring this out:

```javascript
function reduceAsync(arr, memo, iter, done) {
  // For V8 in R: commented out the next line which looks for setTimeout / setImmediate
  //var call = typeof setImmediate == 'undefined' ? setTimeout : setImmediate;
  var i=0;
  next(null, memo);

  function next(err, memo) {
    // Detach next operation from call stack to prevent overflow
    // Don't use setTimeout(, 0) if setImmediate is available
    // (setTimeout() can introduce a long delay if previous operation was slow,
    //    as of Node 0.10.32 -- a bug?)
    if (err) {
      return done(err, null);
    }
    // For V8 in R: comment out the `call` call, and replace with anonymous function
    /*
    call(function() {
      if (i < arr.length === false) {
        done(null, memo);
      } else {
        iter(memo, arr[i++], next);
      }
    }, 0);
    */
    (function() {
      if (i < arr.length === false) {
        done(null, memo);
      } else {
        iter(memo, arr[i++], next);
      }
    })();
  }
}
```

#### 3b. Change output format from buffer to geojson

In the `exportGeoJSON` function, the output format must be changed from `'buffer'` to `'geojson'`.

Search for: `function exportGeoJSON(dataset, opts)`

Then find the line: `content: exportDatasetAsGeoJSON(d, opts, 'buffer')`

In the current version (0.6.113), this is around line 23213.

Change it to:

```javascript
content: exportDatasetAsGeoJSON(d, opts, 'geojson'),
```

#### 3c. Disable printStartupMessages

The `printStartupMessages()` call throws errors in V8 and must be commented out.

Search for: `printStartupMessages()`

In the current version (0.6.113), this is around line 51837.

Comment out the call:

```javascript
// For V8 in R: Commented out to avoid errors
//if (!runningInBrowser()) {
//  printStartupMessages();
//}
```

### 4. Minify the bundle

Minify (uglify) the javascript to make it smaller and faster:

```bash
cd inst/mapshaper
uglifyjs mapshaper-browserify.js -o mapshaper-browserify.min.js -b beautify=false,ascii_only=true
rm mapshaper-browserify.js
```

**Note:** The `ascii_only=true` option is necessary to make it run on Windows.

### 5. Update the version in R code

The mapshaper version is not programmatically accessible from the bundled JavaScript (as of version 0.6.113), so it must be manually updated in `R/utils.R` in the `bundled_ms_version()` function.

Find the current version:
```bash
npm list mapshaper
```

Then update the version string in `R/utils.R`:
```r
bundled_ms_version <- function() {
  # ms <- ms_make_ctx()
  # ms$get("mapshaper.internal.VERSION")
  "0.6.113"  # Update this version number
}
```
 

<!-- README.md is generated from README.Rmd. Please edit that file -->

<!-- badges: start -->

[![CRAN_Status_Badge](http://www.r-pkg.org/badges/version/rmapshaper)](https://cran.r-project.org/package=rmapshaper)
[![CRAN Downloads per
month](http://cranlogs.r-pkg.org/badges/rmapshaper)](https://cran.r-project.org/package=rmapshaper)
[![CRAN total
downloads](http://cranlogs.r-pkg.org/badges/grand-total/rmapshaper?color=lightgrey)](https://cran.r-project.org/package=rmapshaper)
[![R-CMD-check](https://github.com/ateucher/rmapshaper/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/ateucher/rmapshaper/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/ateucher/rmapshaper/graph/badge.svg)](https://app.codecov.io/gh/ateucher/rmapshaper)
<!-- badges: end -->

# rmapshaper <a href="http://andyteucher.ca/rmapshaper/"><img src="man/figures/logo.png" align="right" height="139" alt="rmapshaper website" /></a>

An R package providing access to the awesome
[mapshaper](https://github.com/mbloch/mapshaper/) tool by Matthew Bloch,
which has both a [Node.js command-line
tool](https://github.com/mbloch/mapshaper/wiki/Introduction-to-the-Command-Line-Tool)
as well as an [interactive web tool](https://mapshaper.org/).

I started this package so that I could use mapshaper’s
[Visvalingam](https://bost.ocks.org/mike/simplify/) simplification
method in R. There is, as far as I know, no other R package that
performs topologically-aware multi-polygon simplification. (This means
that shared boundaries between adjacent polygons are always kept intact,
with no gaps or overlaps, even at high levels of simplification).

But mapshaper does much more than simplification, so I am working on
wrapping most of the core functionality of mapshaper into R functions.

So far, `rmapshaper` provides the following functions:

- `ms_simplify` - simplify polygons or lines
- `ms_clip` - clip an area out of a layer using a polygon layer or a
  bounding box. Works on polygons, lines, and points
- `ms_erase` - erase an area from a layer using a polygon layer or a
  bounding box. Works on polygons, lines, and points
- `ms_dissolve` - aggregate polygon features, optionally specifying a
  field to aggregate on. If no field is specified, will merge all
  polygons into one.
- `ms_explode` - convert multipart shapes to single part. Works with
  polygons, lines, and points in geojson format, but currently only with
  polygons and lines in the `Spatial` classes (not `SpatialMultiPoints`
  and `SpatialMultiPointsDataFrame`).
- `ms_lines` - convert polygons to topological boundaries (lines)
- `ms_innerlines` - convert polygons to shared inner boundaries (lines)
- `ms_points` - create points from a polygon layer
- `ms_filter_fields` - Remove fields from the attributes
- `ms_filter_islands` - Remove small detached polygons

If you run into any bugs or have any feature requests, please file an
[issue](https://github.com/ateucher/rmapshaper/issues/)

### Installation

`rmapshaper` is on CRAN. Install the current version with:

``` r
install.packages("rmapshaper")
```

You can install the development version from github with `remotes`:

``` r
## install.packages("remotes")
library(remotes)
install_github("ateucher/rmapshaper")
```

### Usage

rmapshaper works with `sf` objects as well as geojson strings (character
objects of class `geo_json`). It also works with `Spatial` classes from
the `sp` package, though this will likely be retired in the future;
users are encouraged to use the more modern `sf` package.

We will use the `nc.gpkg` file (North Carolina county boundaries) from
the `sf` package and read it in as an `sf` object:

``` r
library(rmapshaper)
library(sf)
#> Linking to GEOS 3.14.1, GDAL 3.12.0, PROJ 9.7.0; sf_use_s2() is TRUE

file <- system.file("gpkg/nc.gpkg", package = "sf")
nc_sf <- read_sf(file)
```

Plot the original:

``` r
plot(nc_sf["FIPS"])
```

![](man/figures/unnamed-chunk-3-1.png)<!-- -->

Now simplify using default parameters, then plot the simplified North
Carolina counties:

``` r
nc_simp <- ms_simplify(nc_sf)
plot(nc_simp["FIPS"])
```

![](man/figures/unnamed-chunk-4-1.png)<!-- -->

You can see that even at very high levels of simplification, the
mapshaper simplification algorithm preserves the topology, including
shared boundaries. The `keep` parameter specifies what proportion of
vertices to keep:

``` r
nc_very_simp <- ms_simplify(nc_sf, keep = 0.001)
plot(nc_very_simp["FIPS"])
```

![](man/figures/unnamed-chunk-5-1.png)<!-- -->

Compare this to the output using `sf::st_simplify`, where overlaps and
gaps are evident:

``` r

nc_stsimp <- st_simplify(nc_sf, preserveTopology = TRUE, dTolerance = 10000) # dTolerance specified in meters
plot(nc_stsimp["FIPS"])
```

![](man/figures/unnamed-chunk-6-1.png)<!-- -->

This time we’ll demonstrate the `ms_innerlines` function:

``` r
nc_sf_innerlines <- ms_innerlines(nc_sf)
plot(nc_sf_innerlines)
```

![](man/figures/unnamed-chunk-7-1.png)<!-- -->

All of the functions are quite fast with `geojson` character objects.
They are slower with the `sf` and `Spatial` classes due to internal
conversion to/from json. If you are going to do multiple operations on
large `sf` objects, it’s recommended to first convert to json using
`geojsonsf::sf_geojson()`, or `geojsonio::geojson_json()`. All of the
functions have the input object as the first argument, and return the
same class of object as the input. As such, they can be chained
together. For a totally contrived example, using `nc_sf` as created
above:

``` r
library(geojsonsf)
library(rmapshaper)
library(sf)

## First convert 'states' dataframe from geojsonsf pkg to json

nc_sf %>%
  sf_geojson() |>
  ms_erase(bbox = c(-80, 35, -79, 35.5)) |> # Cut a big hole in the middle
  ms_dissolve() |> # Dissolve county borders
  ms_simplify(keep_shapes = TRUE, explode = TRUE) |> # Simplify polygon
  geojson_sf() |> # Convert to sf object
  plot(col = "blue") # plot
```

![](man/figures/unnamed-chunk-8-1.png)<!-- -->

### Using the system mapshaper

Sometimes if you are dealing with a very large spatial object in R,
`rmapshaper` functions will take a very long time or not work at all. As
of version `0.4.0`, you can make use of the system `mapshaper` library
if you have it installed. This will allow you to work with very large
spatial objects.

First make sure you have mapshaper installed:

``` r
check_sys_mapshaper()
#> mapshaper version 0.6.113 is installed and on your PATH
#>                     mapshaper-xl 
#> "/opt/homebrew/bin/mapshaper-xl"
```

If you get an error, you will need to install mapshaper. First install
node (<https://nodejs.org/en>) and then install mapshaper in a command
prompt with:

    $ npm install -g mapshaper

Then you can use the `sys` argument in any rmapshaper function:

``` r
nc_simp_internal <- ms_simplify(nc_sf)
nc_simp_sys <- ms_simplify(nc_sf, sys = TRUE, sys_mem = 8) #sys_mem specifies the amount of memory to use in Gb.  It defaults to 8 if omitted.

par(mfrow = c(1, 2))
plot(st_geometry(nc_simp_internal), main = "internal")
plot(st_geometry(nc_simp_sys), main = "system")
```

![](man/figures/unnamed-chunk-10-1.png)<!-- -->

### Thanks

This package uses the [V8](https://cran.r-project.org/package=V8)
package to provide an environment in which to run mapshaper’s javascript
code in R. It relies heavily on all of the great spatial packages that
already exist (especially `sf`), and the `geojsonio` and the `geojsonsf`
packages for converting between `geojson`, `sf` and `Spatial` object.

Thanks to [timelyportfolio](https://github.com/timelyportfolio) for
helping me wrangle the javascript to the point where it works in V8. He
also wrote the [mapshaper
htmlwidget](https://github.com/timelyportfolio/mapshaper_htmlwidget),
which provides access to the mapshaper web interface, right in your R
session. We have plans to combine the two in the future.

### Code of Conduct

Please note that this project is released with a [Contributor Code of
Conduct](https://github.com/ateucher/rmapshaper/blob/master/CONDUCT.md).
By participating in this project you agree to abide by its terms.

### LICENSE

MIT

# Script to update node_packages.csv and LICENSE files for bundled mapshaper dependencies
# This fetches license information and LICENSE files from npm/GitHub
# Only processes modules explicitly bundled with browserify -r flags

library(jsonlite)
library(httr2)
library(dplyr)
library(readr)
library(purrr)

# Configuration
output_csv <- "inst/mapshaper/node_packages.csv"
license_dir <- "inst/mapshaper"

# Explicitly bundled modules (from browserify -r flags)
# Excludes Node.js core modules (buffer, fs, path)
bundled_modules <- c(
  "sync-request",
  "mproj",
  "iconv-lite",
  "flatbush",
  "rw",
  "kdbush",
  "@tmcw/togeojson",
  "@placemarkio/tokml",
  "idb-keyval"
)

cat(sprintf("Processing %d bundled modules\n", length(bundled_modules)))

# Function to get package info from npm registry
get_npm_info <- function(package_name) {
  cat(sprintf("  Fetching info for %s...\n", package_name))

  # Handle scoped packages (@org/package)
  encoded_name <- URLencode(package_name, reserved = TRUE)

  tryCatch(
    {
      resp <- request(sprintf("https://registry.npmjs.org/%s", encoded_name)) |>
        req_timeout(30) |>
        req_perform()

      info <- resp_body_json(resp)

      list(
        name = package_name,
        version = info$`dist-tags`$latest %||% "unknown",
        license = info$license %||% "unknown",
        repository = info$repository$url %||% NA_character_,
        homepage = info$homepage %||% NA_character_
      )
    },
    error = function(e) {
      cat(sprintf(
        "    WARNING: Failed to fetch npm info for %s: %s\n",
        package_name,
        e$message
      ))
      list(
        name = package_name,
        version = "unknown",
        license = "unknown",
        repository = NA_character_,
        homepage = NA_character_
      )
    }
  )
}

# Function to clean GitHub URL
clean_github_url <- function(url) {
  if (is.na(url)) {
    return(NA_character_)
  }

  # Remove git+, .git suffix, git://, ssh://, etc.
  url <- sub("^git\\+", "", url)
  url <- sub("^git://", "https://", url)
  url <- sub("^ssh://git@", "https://", url)
  url <- sub("git@github\\.com:", "https://github.com/", url)
  url <- sub("\\.git$", "", url)

  # Extract github.com URLs
  if (grepl("github\\.com", url)) {
    # Extract owner/repo
    match <- regmatches(url, regexec("github\\.com[:/]([^/]+/[^/]+)", url))[[1]]
    if (length(match) >= 2) {
      return(sprintf("https://github.com/%s", match[2]))
    }
  }

  url
}

# Function to extract owner/repo from GitHub URL
extract_github_repo <- function(url) {
  if (is.na(url) || !grepl("github\\.com", url)) {
    return(NULL)
  }

  match <- regmatches(url, regexec("github\\.com[:/]([^/]+)/([^/]+)", url))[[1]]
  if (length(match) >= 3) {
    list(owner = match[2], repo = match[3])
  } else {
    NULL
  }
}

# Function to fetch LICENSE from GitHub
fetch_github_license <- function(owner, repo, package_name) {
  cat(sprintf("    Fetching LICENSE from GitHub: %s/%s\n", owner, repo))

  # Try common license file names
  license_files <- c(
    "LICENSE",
    "LICENSE.md",
    "LICENSE.txt",
    "LICENCE",
    "LICENCE.md",
    "LICENCE.txt",
    "LICENSE-MIT",
    "LICENSE.MIT"
  )

  for (filename in license_files) {
    tryCatch(
      {
        url <- sprintf(
          "https://raw.githubusercontent.com/%s/%s/main/%s",
          owner,
          repo,
          filename
        )

        resp <- request(url) |>
          req_timeout(30) |>
          req_perform()

        if (resp_status(resp) == 200) {
          return(resp_body_string(resp))
        }
      },
      error = function(e) {
        # Try master branch
        tryCatch(
          {
            url <- sprintf(
              "https://raw.githubusercontent.com/%s/%s/master/%s",
              owner,
              repo,
              filename
            )

            resp <- request(url) |>
              req_timeout(30) |>
              req_perform()

            if (resp_status(resp) == 200) {
              return(resp_body_string(resp))
            }
          },
          error = function(e2) {
            # Silent fail, will try next filename
          }
        )
      }
    )
  }

  cat(sprintf("    WARNING: Could not fetch LICENSE for %s/%s\n", owner, repo))
  NULL
}

# Function to write LICENSE file
write_license_file <- function(content, package_name, license_dir) {
  # Clean package name for filename (remove @org/ prefix)
  clean_name <- sub("^@[^/]+/", "", package_name)
  filename <- file.path(license_dir, sprintf("LICENSE-%s", clean_name))

  writeLines(content, filename)
  cat(sprintf("    Wrote %s\n", filename))
}

# Collect all package information
cat("\nFetching package information from npm...\n")
pkg_info_list <- map(bundled_modules, get_npm_info)

# Process each package
cat("\nProcessing licenses...\n")
for (info in pkg_info_list) {
  cat(sprintf("\n%s (%s):\n", info$name, info$license))

  # Clean repository URL
  repo_url <- clean_github_url(info$repository)
  if (is.na(repo_url)) {
    repo_url <- clean_github_url(info$homepage)
  }

  info$repository <- repo_url

  # Try to fetch LICENSE from GitHub
  if (!is.na(repo_url)) {
    gh_repo <- extract_github_repo(repo_url)
    if (!is.null(gh_repo)) {
      license_content <- fetch_github_license(
        gh_repo$owner,
        gh_repo$repo,
        info$name
      )

      if (!is.null(license_content)) {
        write_license_file(license_content, info$name, license_dir)
      }
    }
  } else {
    cat(sprintf("    WARNING: No GitHub URL found for %s\n", info$name))
  }
}

# Create data frame and write CSV
cat("\nCreating node_packages.csv...\n")
pkg_df <- map_dfr(
  pkg_info_list,
  ~ {
    # Clean repository URL to standard https://github.com format
    repo <- clean_github_url(.x$repository)

    tibble(
      `module name` = sprintf("%s@%s", .x$name, .x$version),
      license = .x$license,
      repository = repo %||% ""
    )
  }
) |>
  arrange(`module name`)

write_csv(pkg_df, output_csv)
cat(sprintf("Wrote %s with %d packages\n", output_csv, nrow(pkg_df)))

# Summary
cat("\n=== Summary ===\n")
cat(sprintf("Total packages: %d\n", nrow(pkg_df)))
cat(sprintf(
  "Packages with GitHub repos: %d\n",
  sum(!is.na(pkg_df$repository) & pkg_df$repository != "")
))

license_files <- list.files(
  license_dir,
  pattern = "^LICENSE-",
  full.names = FALSE
)
cat(sprintf("LICENSE files created: %d\n", length(license_files)))

cat("\nDone!\n")

library(sf)
library(geojsonsf)
# library(rmapshaper)
devtools::load_all()

check_sys_mapshaper()

MySfLayer <- st_read("~/Downloads/MyLayers/MySfLayer.gpkg")
MySfClippingLayer <- st_read("~/Downloads/MyLayers/MySfClippingLayer.gpkg")

MySfLayerWGS <- st_transform(MySfLayer, crs = 4326)
MySfClippingLayerWGS <- st_transform(MySfClippingLayer, crs = 4326)

# tictoc::tic()
ClipSf <- ms_clip(
  MySfLayerWGS,
  MySfClippingLayerWGS,
  sys = TRUE,
  sys_mem = 16
)
# tictoc::toc()

system.time(
  rmapshaper:::sf_to_GeoJSON(MySfLayer, file = TRUE)
)

MyGeoJLayer <- sf_geojson(MySfLayer, simplify = FALSE)
MyGeoJLayerWGS <- sf_geojson(MySfLayerWGS, simplify = FALSE)
MyGeoJClippingLayerWGS <- sf_geojson(MySfClippingLayerWGS, simplify = FALSE)

tictoc::tic()
ClipGeoJ <- ms_clip(
  MyGeoJLayerWGS,
  MyGeoJClippingLayerWGS,
  sys = TRUE,
  sys_mem = 16
)
tictoc::toc()

library(rnaturalearth)
library(sf)
library(rmapshaper)
library(geojsonsf)
library(dplyr)
library(hexSticker)
library(showtext)
library(ggplot2)
## Loading Google fonts (http://www.google.com/fonts)
font_add_google("Nova Mono", "novamono")
## Automatically use showtext to render text for future devices
showtext_auto()

## Seed for colours (dummyval)
set.seed(32)

world <- ne_countries(returnclass = "sf")
world_simp <- world |>
  mutate(dummyval = runif(nrow(world))) |>
  select(iso_a3, dummyval) |>
  st_transform("ESRI:54030") |>
  ms_simplify(keep = 0.1)


world_centroid <- st_coordinates(st_centroid(ms_dissolve(world_simp)))

poly_centroids <- st_coordinates(st_centroid(world_simp))

xdiff <- poly_centroids[, "X"] - world_centroid[, "X"]
ydiff <- poly_centroids[, "Y"] - world_centroid[, "Y"]

## Percentage shift from world centroid
xshift <- 0.1
yshift <- 0.1

## Seed for rotation
set.seed(36)
rotate <- runif(length(xdiff), -60, 60)

affine_shift_cmd <-
  glue::glue(
    "-affine shift={xdiff * xshift},{ydiff * yshift} rotate={rotate} where='iso_a3 == \"{world_simp$iso_a3}\"'"
  ) |>
  Filter(\(x) !grepl("ATA", x), x = _) |>
  glue::glue_collapse(sep = " ")


crazy_map <- sf_geojson(world_simp) |>
  apply_mapshaper_commands(affine_shift_cmd) |>
  geojson_sf(
    input = st_crs(world_simp)$input,
    wkt = st_crs(world_simp)$wkt
  ) |>
  select(iso_a3, dummyval)

p <- ggplot() +
  geom_sf(
    data = crazy_map,
    aes(fill = dummyval),
    linewidth = 0.1,
    show.legend = FALSE
  ) +
  scale_fill_distiller(palette = "Greens") +
  theme_void()

## hex sticker

write_sticker <- function(p, format, plot_dim) {
  sticker(
    p,
    package = "rmapshaper",
    p_size = 15, # This seems to behave very differently on a Mac vs PC
    p_y = 0.45,
    p_color = "#c2e4a5",
    p_family = "novamono",
    p_fontface = "bold",
    s_x = 1.0,
    s_y = 1.04,
    s_width = plot_dim,
    s_height = plot_dim,
    h_color = "#c2e4a5",
    filename = file.path(paste0("inst/sticker/rmapshaper.", format))
  )
}

(write_sticker(p, "png", 1.85))
# write_sticker(p, "svg")

usethis::use_logo("inst/sticker/rmapshaper.png")

#' Drop features from a \code{geo_json} FeatureCollection with null geometries
#'
#' @param x a \code{geo_json} FeatureCollection
#'
#' @return a \code{geo_json} FeatureCollection with Features with null geometries
#' removed
#' @export
drop_null_geometries <- function(x) {
  UseMethod("drop_null_geometries")
}

#' @export
drop_null_geometries.json <- function(x) {
  apply_mapshaper_commands(x, "-filter remove-empty", TRUE)
}

#' @export
drop_null_geometries.character <- drop_null_geometries.json

test_that("drop_null_geometries_works", {
  poly_geojson <- structure(
    '{"type":"FeatureCollection",
    "features":[{
    "type":"Feature",
    "geometry":{"type":"Point","coordinates":[1,1]},
    "properties":{"x":1}},{
    "type":"Feature",
    "geometry":{},
    "properties":{"x":2}}]}',
    class = c("geojson", "json")
  )
  expect_snapshot_value(drop_null_geometries(poly_geojson), style = "json2")
})

.onAttach <- function(libname, pkgname) {
  # nocov start
  startup_v8_version_check()
  # nocov end
}

startup_v8_version_check <- function() {
  if (v8_version() < package_version("8.1.307.30")) {
    packageStartupMessage(
      "Warning: v8 Engine is version ",
      v8_version(),
      " but version >= 8.1.307.30 is required for full functionality. See",
      " https://github.com/jeroen/V8 for help installing a modern version",
      " of v8 on your operating system."
    )
  }
}

test_that("loading with old v8 engine gives a message", {
  local_mocked_bindings(v8_version = function() package_version("8.0"))
  expect_message(
    startup_v8_version_check(),
    "Warning: v8 Engine is version 8.0"
  )
})

#' Remove small detached polygons (islands)
#'
#' Remove small detached polygons, keeping those with a minimum area and/or a
#' minimum number of vertices. Optionally remove null geometries.
#'
#' @param input spatial object to filter. One of:
#' \itemize{
#'  \item \code{geo_json} or \code{character} polygons;
#'  \item \code{SpatialPolygons*};
#'  \item \code{sf} or \code{sfc} polygons object
#'  }
#' @param min_area minimum area of polygons to retain. Area is calculated using
#'  planar geometry, except for the area of unprojected polygons, which is
#'  calculated using spherical geometry in units of square meters.
#' @param min_vertices minimum number of vertices in polygons to retain.
#' @param drop_null_geometries should features with empty geometries be dropped?
#'   Default \code{TRUE}. Ignored for \code{SpatialPolyons*}, as it is always
#'   \code{TRUE}.
#' @inheritDotParams apply_mapshaper_commands force_FC sys sys_mem quiet gj2008
#'
#' @return object with only specified features retained, in the same class as
#'   the input
#'
#' @examples
#' library(geojsonsf)
#' library(sf)
#'
#' poly <- structure("{\"type\":\"FeatureCollection\",
#'            \"features\":[{\"type\":\"Feature\",\"properties\":{},
#'            \"geometry\":{\"type\":\"Polygon\",
#'            \"coordinates\":[[[102,2],[102,4],[104,4],[104,2],[102,2]]]}},
#'            {\"type\":\"Feature\",\"properties\":{},
#'            \"geometry\":{\"type\":\"Polygon\",
#'            \"coordinates\":[[[100,2],[98,4],[101.5,4],[100,2]]]}},
#'            {\"type\":\"Feature\",\"properties\":{},
#'            \"geometry\":{\"type\":\"Polygon\",
#'            \"coordinates\":[[[100,0],[100,1],[101,1],[101,0],[100,0]]]}}]}",
#'            class = c("geojson", "json"))
#'
#' poly <- geojson_sf(poly)
#' plot(poly)
#'
#' out <- ms_filter_islands(poly, min_area = 12391399903)
#' plot(out)
#'
#' @export
ms_filter_islands <- function(
  input,
  min_area = NULL,
  min_vertices = NULL,
  drop_null_geometries = TRUE,
  ...
) {
  if (!is.null(min_area) && !is.numeric(min_area))
    stop("min_area must be numeric")
  if (!is.null(min_vertices) && !is.numeric(min_vertices))
    stop("min_vertices must be numeric")
  if (!is.logical(drop_null_geometries))
    stop("drop_null_geometries must be TRUE or FALSE")
  UseMethod("ms_filter_islands")
}

#' @export
ms_filter_islands.character <- function(
  input,
  min_area = NULL,
  min_vertices = NULL,
  drop_null_geometries = TRUE,
  ...
) {
  input <- check_character_input(input)

  cmd <- make_filterislands_call(
    min_area = min_area,
    min_vertices = min_vertices,
    drop_null_geometries = drop_null_geometries
  )

  apply_mapshaper_commands(data = input, command = cmd, ...)
}

#' @export
ms_filter_islands.json <- function(
  input,
  min_area = NULL,
  min_vertices = NULL,
  drop_null_geometries = TRUE,
  ...
) {
  cmd <- make_filterislands_call(
    min_area = min_area,
    min_vertices = min_vertices,
    drop_null_geometries = drop_null_geometries
  )

  apply_mapshaper_commands(data = input, command = cmd, ...)
}

#' @export
ms_filter_islands.SpatialPolygons <- function(
  input,
  min_area = NULL,
  min_vertices = NULL,
  drop_null_geometries = TRUE,
  ...
) {
  ms_filter_islands_sp(
    input,
    min_area = min_area,
    min_vertices = min_vertices,
    ...
  )
}


ms_filter_islands_sp <- function(
  input,
  min_area = NULL,
  min_vertices = NULL,
  ...
) {
  cmd <- make_filterislands_call(
    min_area = min_area,
    min_vertices = min_vertices,
    drop_null_geometries = TRUE
  )
  ms_sp(input = input, call = cmd, ...)
}

#' @export
ms_filter_islands.sf <- function(
  input,
  min_area = NULL,
  min_vertices = NULL,
  drop_null_geometries = TRUE,
  ...
) {
  cmd <- make_filterislands_call(
    min_area = min_area,
    min_vertices = min_vertices,
    drop_null_geometries = TRUE
  )
  ms_sf(input = input, call = cmd, ...)
}

#' @export
ms_filter_islands.sfc <- ms_filter_islands.sf

make_filterislands_call <- function(
  min_area,
  min_vertices,
  drop_null_geometries
) {
  if (!is.null(min_area)) min_area <- paste0("min-area=", min_area)
  if (!is.null(min_vertices))
    min_vertices <- paste0("min-vertices=", min_vertices)

  if (drop_null_geometries) rem <- "remove-empty" else rem <- NULL

  call <- list("-filter-islands", min_area, min_vertices, rem)
  call
}

islands_poly <- structure(
  "{\"type\":\"FeatureCollection\",
\"features\":[{\"type\":\"Feature\",\"properties\":{},
\"geometry\":{\"type\":\"Polygon\",
\"coordinates\":[[[102,2],[102,4],[104,4],[104,2],[102,2]]]}},
{\"type\":\"Feature\",\"properties\":{},
\"geometry\":{\"type\":\"Polygon\",
\"coordinates\":[[[100,2],[98,4],[101.5,4],[100,2]]]}},
{\"type\":\"Feature\",\"properties\":{},
\"geometry\":{\"type\":\"Polygon\",
\"coordinates\":[[[100,0],[100,1],[101,1],[101,0],[100,0]]]}}]}",
  class = c("geojson", "json")
)

islands_poly_spdf <- GeoJSON_to_sp(islands_poly)
islands_poly_sp <- as(islands_poly_spdf, "SpatialPolygons")


islands_poly_sf <- st_as_sf(islands_poly_spdf)
islands_poly_sfc <- st_geometry(islands_poly_sf)

test_that("ms_filter_islands works with min_area", {
  expect_s3_class(
    ms_filter_islands(islands_poly, min_area = 12391399903),
    "geojson"
  )
  expect_s3_class(
    ms_filter_islands(unclass(islands_poly), min_area = 12391399903),
    "geojson"
  )

  expect_snapshot_value(
    ms_filter_islands(islands_poly, min_area = 12391399903),
    style = "json2"
  )

  out_spdf <- ms_filter_islands(islands_poly_spdf, min_area = 12391399903)
  out_sp <- ms_filter_islands(islands_poly_sp, min_area = 12391399903)
  expect_equal(out_spdf@polygons, out_sp@polygons)
  expect_equal(length(out_spdf@polygons), 2)
  expect_equal(
    out_spdf@polygons[[1]]@Polygons[[1]]@coords,
    structure(c(102, 102, 104, 104, 102, 2, 4, 4, 2, 2), .Dim = c(5L, 2L))
  )
  expect_equal(
    out_spdf@polygons[[2]]@Polygons[[1]]@coords,
    structure(c(100, 98, 101.5, 100, 2, 4, 4, 2), .Dim = c(4L, 2L))
  )

  out_sf <- ms_filter_islands(islands_poly_sf, min_area = 12391399903)
  out_sfc <- ms_filter_islands(islands_poly_sfc, min_area = 12391399903)
  expect_equal(length(out_sfc), 2)
  expect_equal(st_geometry(out_sf), out_sfc)
})

test_that("ms_filter_islands works with min_vertoces", {
  expect_s3_class(ms_filter_islands(islands_poly, min_vertices = 4), "geojson")
  expect_snapshot_value(
    ms_filter_islands(islands_poly, min_vertices = 4),
    style = "json2"
  )

  expect_s3_class(
    ms_filter_islands(unclass(islands_poly), min_vertices = 4),
    "geojson"
  )
  out_spdf <- ms_filter_islands(islands_poly_spdf, min_vertices = 4)
  expect_equal(length(out_spdf@polygons[[1]]@Polygons), 1)
  expect_equal(
    out_spdf@polygons[[1]]@Polygons[[1]]@coords,
    structure(c(102, 102, 104, 104, 102, 2, 4, 4, 2, 2), .Dim = c(5L, 2L))
  )

  out_sfc <- ms_filter_islands(islands_poly_sfc, min_vertices = 4)
  expect_equal(length(out_sfc), 2)
})

test_that("ms_filter_islands works drop_null_geometries = FALSE", {
  expect_s3_class(
    ms_filter_islands(
      islands_poly,
      min_area = 43310462718,
      drop_null_geometries = FALSE
    ),
    "geojson"
  )
  expect_snapshot_value(
    ms_filter_islands(
      islands_poly,
      min_area = 43310462718,
      drop_null_geometries = FALSE
    ),
    style = "json2"
  )

  expect_s3_class(
    ms_filter_islands(
      unclass(islands_poly),
      min_area = 43310462718,
      drop_null_geometries = FALSE
    ),
    "geojson"
  )
  out_spdf <- ms_filter_islands(
    islands_poly_spdf,
    min_area = 43310462718,
    drop_null_geometries = FALSE
  )
  expect_equal(length(out_spdf@polygons[[1]]@Polygons), 1)
  expect_equal(
    out_spdf@polygons[[1]]@Polygons[[1]]@coords,
    structure(c(102, 102, 104, 104, 102, 2, 4, 4, 2, 2), .Dim = c(5L, 2L))
  )
})

test_that("specifying min_vertices and min_area works", {
  skip(
    "possible bug in mapshaper https://github.com/mbloch/mapshaper/issues/487"
  )
  expect_snapshot_value(
    ms_filter_islands(
      geojson_list(islands_poly),
      min_area = 12391399902,
      min_vertices = 4
    ),
    style = "json2"
  )
})

test_that("ms_filter_islands fails correctly", {
  expect_error(
    ms_filter_islands(islands_poly, min_area = "foo"),
    "min_area must be numeric"
  )
  expect_error(
    ms_filter_islands(islands_poly, min_vertices = "foo"),
    "min_vertices must be numeric"
  )
  expect_error(
    ms_filter_islands(islands_poly, drop_null_geometries = "foo"),
    "drop_null_geometries must be TRUE or FALSE"
  )
  expect_error(
    ms_filter_islands(islands_poly, force_FC = "foo"),
    "force_FC must be TRUE or FALSE"
  )
})

test_that("ms_filter_islands works with sys = TRUE", {
  skip_if_not(has_sys_mapshaper())
  expect_s3_class(
    ms_filter_islands(islands_poly, min_area = 12391399902, sys = TRUE),
    "geojson"
  )
  expect_snapshot_value(
    ms_filter_islands(islands_poly, min_area = 12391399902, sys = TRUE),
    style = "json2"
  )
  expect_s4_class(
    ms_filter_islands(islands_poly_spdf, min_area = 12391399902, sys = TRUE),
    "SpatialPolygonsDataFrame"
  )
  expect_s3_class(
    ms_filter_islands(islands_poly_sf, min_area = 12391399902, sys = TRUE),
    "sf"
  )
})

#' Apply a mapshaper command string to a geojson object
#'
#' @param data character containing geojson or path to geojson file. If a file
#'   path, \code{sys} must be true.
#' @param command valid mapshaper command string
#' @param force_FC should the output be forced to be a FeatureCollection (or sf
#'   object or Spatial*DataFrame) even if there are no attributes? Default
#'   \code{TRUE}. If FALSE and there are no attributes associated with the
#'   geometries, a GeometryCollection (or Spatial object with no dataframe, or
#'   sfc) will be output.
#' @param sys Should the system mapshaper be used instead of the bundled
#'   mapshaper? Gives better performance on large files. Requires the mapshaper
#'   node package to be installed and on the PATH.
#' @param sys_mem How much memory (in GB) should be allocated if using the
#'   system mapshaper (`sys = TRUE`)? Default 8. Ignored if `sys = FALSE`. This
#'   can also be set globally with the option `"mapshaper.sys_mem"`
#' @param quiet If `sys = TRUE`, should the mapshaper messages be silenced?
#'   Default `FALSE`. This can also be set globally with the option
#'   `"mapshaper.sys_quiet"`
#' @param gj2008 Generate output that is consistent with the pre-RFC 7946
#'   GeoJSON spec (dating to 2008). Polygon rings are CW and holes are CCW,
#'   which is the opposite of the default RFC 7946-compatible output. This should
#'   be rarely needed, though may be useful when preparing data for D3-based
#'   data visualizations (such as `plotly::plot_ly()`). Default `FALSE`
#'
#' @return geojson
#' @export
#' @examples
#'
#' nc <- sf::read_sf(system.file("gpkg/nc.gpkg", package = "sf"))
#' rmapshaper::apply_mapshaper_commands(geojsonsf::sf_geojson(nc), "-clean")
#'
apply_mapshaper_commands <- function(
  data,
  command,
  force_FC = TRUE,
  sys = FALSE,
  sys_mem = getOption("mapshaper.sys_mem", default = 8),
  quiet = getOption("mapshaper.sys_quiet", default = FALSE),
  gj2008 = FALSE
) {
  if (!is.logical(force_FC)) {
    stop("force_FC must be TRUE or FALSE", call. = FALSE)
  }
  if (!is.logical(sys)) {
    stop("sys must be TRUE or FALSE", call. = FALSE)
  }
  if (!is.numeric(sys_mem)) {
    stop("sys_mem must be numeric", call. = FALSE)
  }
  if (!is.logical(gj2008)) {
    stop("gj2008 must be TRUE or FALSE", call. = FALSE)
  }

  data <- as.character(data)

  if (!is.numeric(sys_mem)) {
    if (nchar(data) < 1000L && file.exists(data) && !sys) {
      stop(
        "'data' points to a file on disk but you did not specify to use
         the system mapshaper. To do so set sys = TRUE"
      )
    }
  }

  # command <- paste(ms_compact(command), collapse = " ")

  if (sys) {
    ret <- sys_mapshaper(
      data = data,
      command = command,
      force_FC = force_FC,
      sys_mem = sys_mem,
      quiet = quiet,
      gj2008 = gj2008
    )
  } else {
    out_flags <- NULL

    if (force_FC || gj2008) {
      add_id <- if (force_FC) fc_command() else NULL
      gj2008 <- if (gj2008) "gj2008" else NULL
      out_flags <- paste("-o", add_id, gj2008)
    }

    command <- c(
      command,
      out_flags
    )

    command <- paste(ms_compact(command), collapse = " ")

    ms <- ms_make_ctx()

    ## Create a JS object to hold the returned data
    ms$eval("var return_data;")

    ms$call("mapshaper.applyCommands", command, data, V8::JS(callback()))

    # TODO: New syntax for applyCommands:
    # Useful for clip and erase so can define two inputs
    # cmd <- paste("-i input.geojson", command, "-o output.geojson")
    #
    # ms$call("mapshaper.applyCommands", cmd, V8::JS("{'input.geojson': input}"),
    #         V8::JS("function(err, output) {
    # 	return_data = output['output.geojson'];
    # }"))
    ret <- ms$get("return_data")
    ret <- class_geo_json(ret)
  }
  ret
}

ms_make_ctx <- function() {
  ctx <- V8::v8()
  ctx$source(system.file(
    "mapshaper/mapshaper-browserify.min.js",
    package = "rmapshaper"
  ))
  ctx
}

sys_mapshaper <- function(
  data,
  data2 = NULL,
  command,
  force_FC = TRUE, # default FALSE as in most cases it is added in apply_mapshaper_commands
  sys_mem = getOption("mapshaper.sys_mem", default = 8),
  quiet = getOption("mapshaper.sys_quiet", default = FALSE),
  gj2008 = FALSE
) {
  # Get full path to sys mapshaper, use mapshaper-xl
  ms_path <- paste0(check_sys_mapshaper("mapshaper-xl", verbose = FALSE))

  # Check if need to read/write the file or if it's been written already
  # by write_sf or writeOGR
  read_write <- !file.exists(data)

  in_data_file <- data
  in_data_file2 <- data2

  if (read_write) {
    in_data_file <- temp_geojson()
    readr::write_file(data, in_data_file)
    on.exit(unlink(in_data_file))

    if (!is.null(data2)) {
      in_data_file2 <- temp_geojson()
      readr::write_file(data2, in_data_file2)
      on.exit(unlink(in_data_file2), add = TRUE)
    }
  }

  out_data_file <- temp_geojson()

  out_fc <- if (force_FC) fc_command() else NULL
  gj2008 <- if (gj2008) "gj2008" else NULL

  cmd_args <- c(
    sys_mem,
    shQuote(in_data_file),
    command <- paste(ms_compact(command), collapse = " "),
    shQuote(in_data_file2), # will be NULL if no data2/in_data_file2
    "-o",
    shQuote(out_data_file),
    out_fc, # will be NULL if force_FC is FALSE
    gj2008, # will be NULL if gj2008 is FALSE
    if (quiet) "-quiet"
  )

  system2(ms_path, cmd_args)

  if (read_write) {
    on.exit(unlink(out_data_file), add = TRUE)
    # Read the geojson object and return it
    ret <- class_geo_json(readr::read_file(out_data_file))
  } else {
    # Return the path to the file
    ret <- out_data_file
  }

  ret
}

ms_get_raw <- function(x) {
  rawToChar(as.raw(x[["data"]]))
}

## create a JS callback function
callback <- function() {
  "function(Error, data) {
if (Error) console.error('Error in V8 context: ' + Error.stack);
return_data = data;
}"
}

ms_sp <- function(
  input,
  call,
  sys = FALSE,
  force_FC = TRUE,
  sys_mem = getOption("mapshaper.sys_mem", default = 8),
  quiet = getOption("mapshaper.sys_quiet", default = FALSE),
  gj2008 = FALSE
) {
  has_data <- .hasSlot(input, "data")
  if (has_data) {
    classes <- col_classes(input@data)
  }

  geojson <- sp_to_GeoJSON(input, file = sys)

  ret <- apply_mapshaper_commands(
    data = geojson,
    command = call,
    force_FC = force_FC,
    sys = sys,
    sys_mem = sys_mem,
    quiet = quiet,
    gj2008 = gj2008
  )

  ret <- GeoJSON_to_sp(ret, crs = attr(geojson, "crs"))

  # remove data slot if input didn't have one (default out_class is the class of the input)
  if (!has_data) {
    ret <- as(ret, gsub("DataFrame$", "", class(ret)[1]))
  } else {
    ret@data <- restore_classes(ret@data, classes)
  }

  ret
}

GeoJSON_to_sp <- function(geojson, crs = NULL) {
  x_sf <- GeoJSON_to_sf(geojson, crs)
  if (nrow(x_sf) == 0) {
    stop(
      "Cannot convert result to a Spatial* object.
         It is likely too much simplification was applied and all features
         were reduced to null.",
      call. = FALSE
    )
  }
  as(x_sf, "Spatial")
}

sp_to_GeoJSON <- function(sp, file = FALSE) {
  crs <- methods::slot(sp, "proj4string")

  if (file) {
    js <- sp_to_tempfile(sp)
  } else {
    js_tmp <- sp_to_tempfile(sp)
    js <- readr::read_file(js_tmp, locale = readr::locale())
    on.exit(unlink(js_tmp))
  }
  structure(js, crs = crs)
}

## Utilties for sf
ms_sf <- function(
  input,
  call,
  sys = FALSE,
  force_FC = TRUE,
  sys_mem = getOption("mapshaper.sys_mem", default = 8),
  quiet = getOption("mapshaper.sys_quiet", default = FALSE),
  gj2008 = FALSE
) {
  has_data <- inherits(input, "sf")
  if (has_data) {
    classes <- col_classes(input)
    geom_name <- attr(input, "sf_column")
    input <- ms_de_unit(input)
  } else {
    input <- unname(input)
  }

  geojson <- sf_to_GeoJSON(input, file = sys)

  ret <- apply_mapshaper_commands(
    data = geojson,
    command = call,
    force_FC = force_FC,
    sys = sys,
    sys_mem = sys_mem,
    quiet = quiet,
    gj2008 = gj2008
  )

  if (!sys & grepl('^\\{"type":"GeometryCollection"', ret)) {
    stop(
      "Cannot convert result to an sf object.
         It is likely too much simplification was applied and all features
         were reduced to null.",
      call. = FALSE
    )
  }

  ret <- GeoJSON_to_sf(ret, crs = attr(geojson, "crs"))

  ## Only return sfc if that's all that was input
  if (!has_data) {
    ret <- sf::st_geometry(ret)
  } else {
    ret <- restore_classes(ret, classes)
    names(ret)[names(ret) == attr(ret, "sf_column")] <- geom_name
    sf::st_geometry(ret) <- geom_name
  }
  ##maintain tbl_df
  if (all(class(input) == c("sf", "tbl_df", "tbl", "data.frame"))) {
    class(ret) <- c("sf", "tbl_df", "tbl", "data.frame")
  }
  ret
}

GeoJSON_to_sf <- function(geojson, crs = NULL) {
  sf <- geojsonsf::geojson_sf(geojson)
  if (!is.null(crs)) {
    suppressWarnings(sf::st_crs(sf) <- crs)
  }
  curly_brace_na(sf)
}

sf_to_GeoJSON <- function(sf, file = FALSE) {
  crs <- sf::st_crs(sf)

  js <- if (inherits(sf, "sf")) {
    geojsonsf::sf_geojson(sf, simplify = FALSE)
  } else {
    json <- geojsonsf::sfc_geojson(sf)
    paste0(
      "{\"type\":\"GeometryCollection\",\"geometries\":[",
      paste(json, collapse = ","),
      "]}"
    )
  }

  if (file) {
    path <- tempfile(fileext = ".geojson")
    writeLines(js, con = path)
    js <- path
  }
  structure(js, crs = crs)
}


sp_to_tempfile <- function(obj) {
  obj <- sf::st_as_sf(sp_to_spdf(obj))
  path <- tempfile(fileext = ".geojson")
  sf::st_write(obj, path, driver = "GeoJSON", quiet = TRUE, delete_dsn = TRUE)
  normalizePath(path, winslash = "/", mustWork = TRUE)
}

sp_to_spdf <- function(obj) {
  non_df_classes <- c("SpatialLines", "SpatialPolygons", "SpatialPoints")
  cls <- inherits(obj, non_df_classes, which = TRUE)
  if (!any(cls)) {
    return(obj)
  }
  as(obj, paste0(non_df_classes[as.logical(cls)], "DataFrame"))
}

#' Check the system mapshaper
#'
#' @param command either "mapshaper-xl" (default) or "mapshaper"
#' @param verbose Print a message stating mapshaper's current version? Default `TRUE`
#'
#' @return character path to mapshaper executable if appropriate version is installed, otherwise throws an error
#' @export
check_sys_mapshaper <- function(command = "mapshaper-xl", verbose = TRUE) {
  if (!command %in% c("mapshaper-xl", "mapshaper")) {
    stop("command must be one of 'mapshaper-xl' or 'mapshaper'", call. = FALSE)
  }

  ms_path <- sys_ms_path(command)

  sys_ms_version <- package_version(sys_ms_version())
  min_ms_version <- package_version(bundled_ms_version()) # Update when updating bundled mapshaper.js

  if (sys_ms_version < min_ms_version) {
    stop(
      "You need to upgrade your system mapshaper library.\n",
      "Update it with: 'npm update -g mapshaper"
    )
  }
  if (verbose) {
    message(
      "mapshaper version ",
      sys_ms_version,
      " is installed and on your PATH"
    )
  }
  ms_path
}

sys_ms_path <- function(command) {
  err_msg <- paste0(
    "The mapshaper node library must be installed and on your PATH.\n",
    "Install node.js (https://nodejs.org/en/) and then install mapshaper with:\n",
    "    npm install -g mapshaper"
  )

  ms_path <- Sys.which(command)

  if (!nzchar(ms_path)) {
    # try to find it:
    if (nzchar(Sys.which("npm"))) {
      npm_prefix <- system2("npm", "get prefix", stdout = TRUE)
      ms_path <- file.path(npm_prefix, "bin", command)
      if (!file.exists(ms_path)) stop(err_msg, call. = FALSE)
    } else {
      stop(err_msg, call. = FALSE)
    }
  }
  ms_path
}

sys_ms_version <- function() {
  system2(sys_ms_path("mapshaper"), "--version", stdout = TRUE)
}

bundled_ms_version <- function() {
  # ms <- ms_make_ctx()
  # ms$get("mapshaper.internal.VERSION")
  "0.6.113"
}

ms_compact <- function(l) Filter(Negate(is.null), l)

fc_command <- function() {
  "geojson-type=FeatureCollection"
}

class_geo_json <- function(x) {
  if (is.null(x)) {
    warning(
      "The command returned an empty response. Please check your inputs",
      call. = FALSE
    )
    x <- list()
  }
  if (is.raw(x)) {
    x <- rawToChar(x)
  }
  structure(x, class = c("geojson", "json"))
}

check_character_input <- function(x) {
  ## Collapse to character vector of length one if many lines (e.g., if used readLines)
  if (length(x) > 1) {
    x <- paste0(x, collapse = "")
  }
  if (!jsonify::validate_json(x)) {
    stop("Input is not valid geojson")
  }
  x
}

## Convert empty curly braces that come out of V8 to NA (that is what they go in as)
curly_brace_na <- function(x) {
  UseMethod("curly_brace_na")
}

#' @export
curly_brace_na.data.frame <- function(x) {
  chr_or_factor <- vapply(
    x,
    inherits,
    c("character", "factor"),
    FUN.VALUE = logical(1)
  )
  if (any(chr_or_factor)) {
    x[, chr_or_factor][x[, chr_or_factor] == "{ }"] <- NA
  }
  x
}

#' @export
curly_brace_na.Spatial <- function(x) {
  x@data <- curly_brace_na(x@data)
  x
}

# This method basically just removes the sf class and then
# restores it after the data.frame method does its work, because
# the sf column is 'sticky' with `[`.sf methods, so would be
# included in the { } substitution if the sf class was kept

#' @export
curly_brace_na.sf <- function(x) {
  sf_col <- attr(x, "sf_column")
  class(x) <- setdiff(class(x), "sf")
  x <- curly_brace_na(x)
  sf::st_as_sf(x, sf_column_name = sf_col)
}

col_classes <- function(df) {
  classes <- lapply(df, function(x) {
    out <- list()
    if (inherits(x, "POSIXlt")) {
      stop(
        "POSIXlt classes not supported. Please convert to POSIXct",
        call. = FALSE
      )
    }
    out$class <- class(x)
    if (is.factor(x)) {
      out$levels <- levels(x)
      if (is.ordered(x)) {
        out$ordered <- TRUE
      } else {
        out$ordered <- FALSE
      }
    }
    out
  })
  classes
}

restore_classes <- function(df, classes) {
  in_classes <- lapply(df, class)

  keep_in_both <- intersect(names(in_classes), names(classes))
  in_classes <- in_classes[keep_in_both]

  class_matches <- vapply(
    names(in_classes),
    function(n) {
      if ("sfc" %in% classes[[n]]$class) {
        return(TRUE)
      }
      isTRUE(all.equal(classes[[n]]$class, in_classes[[n]]))
    },
    FUN.VALUE = logical(1)
  )

  mismatches <- which(!class_matches)
  if (length(mismatches) == 0) {
    return(df)
  }

  for (n in names(mismatches)) {
    cls <- classes[[n]]$class
    if ("factor" %in% cls) {
      df[[n]] <- factor(
        df[[n]],
        levels = classes[[n]]$levels,
        ordered = classes[[n]]$ordered
      )
    } else if (!"units" %in% cls) {
      # Skip units columns... TODO: Fix units parsing on return
      as_fun <- paste0("as.", cls[1])
      tryCatch(
        {
          df[[n]] <- eval(call(as_fun, df[[n]]))
        },
        error = function(e) {
          warning(
            "Could not convert column ",
            names(df)[n],
            " to class ",
            cls,
            ". Returning as ",
            paste(in_classes[[n]], collapse = ", ")
          )
        }
      )
    }
  }
  df
}

stop_for_old_v8 <- function() {
  if (v8_version() < "6") {
    # nocov start
    stop(
      "Warning: v8 Engine is version ",
      V8::engine_info()[["version"]],
      " but version >=6 is required for this function. See",
      " https://github.com/jeroen/V8 for help installing a modern version",
      " of v8 on your operating system."
    )
  }
  # nocov end
}

v8_version <- function() {
  V8::engine_info()[["numeric_version"]]
}

temp_geojson <- function() {
  # This option is really just to allow testing strange paths like #107
  tmpdir <- getOption("ms_tempdir", default = tempdir())
  dir.create(tmpdir, showWarnings = FALSE)
  normalizePath(
    tempfile(
      tmpdir = tmpdir,
      fileext = ".geojson"
    ),
    mustWork = FALSE
  )
}

upkeep_bullets <- function() "Update bundled mapshaper node library."

has_sys_mapshaper <- function() {
  nzchar(Sys.which("mapshaper"))
}

skip_on_old_v8 <- function() {
  if (v8_version() < "6") {
    testthat::skip("Skipping due to old V8 engine")
  }
}

expect_equivalent <- function(object, expected, ...) {
  if (inherits(object, "Spatial") && .hasSlot(object, "proj4string")) {
    comment(object@proj4string) <- NULL
    comment(expected@proj4string) <- NULL
  }
  testthat::expect_equal(object, expected, ignore_attr = TRUE, ...)
}

test_df <- data.frame(
  char = letters[1:3],
  dbl = c(1.1, 2.2, 3.3),
  int = c(1L, 2L, 3L),
  fct = factor(LETTERS[4:6]),
  ord = ordered(LETTERS[7:9], levels = rev(LETTERS[7:9])),
  date = as.Date(c("2016-01-25", "2017-01-27", "1979-03-09")),
  posix_ct = as.POSIXct(c(
    "2016-01-25 11:25:03",
    "2017-01-27 23:24:56",
    "1979-03-09 10:25:15"
  )),
  "column 6" = 1:3,
  stringsAsFactors = FALSE,
  check.names = FALSE
)

test_that("Restore column works", {
  cls <- col_classes(test_df)
  expect_type(cls, "list")
  expect_equal(length(cls), ncol(test_df))

  back_in <- fromJSON(toJSON(test_df))
  expect_equal(
    unname(sapply(back_in, class)),
    c(
      "character",
      "numeric",
      "integer",
      "character",
      "character",
      "character",
      "character",
      "integer"
    )
  )
  restored <- restore_classes(df = back_in, classes = cls)
  expect_equal(lapply(test_df, class), lapply(restored, class))
  expect_equal(names(test_df), names(restored)) # retain special names, https://github.com/ateucher/rmapshaper/issues/91
  expect_equal(test_df, restored)
  test_df$posix_lt <- as.POSIXlt(test_df$posix_ct)
  expect_error(col_classes(test_df), "POSIXlt classes not supported")
})

test_that("dealing with encoding works", {
  test_name <- "örtchen"
  test_value <- "Münster"
  Encoding(test_name) <- "UTF-8"
  Encoding(test_value) <- "UTF-8"

  pts <- "{\"type\":\"FeatureCollection\",\"features\":[\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"Point\",\"coordinates\":[-78,-53]},\"properties\":{\"örtchen\":\"Münster\"}}]}"
  Encoding(pts) <- "UTF-8"

  out <- GeoJSON_to_sp(pts)
  expect_s4_class(out, "SpatialPointsDataFrame")
  expect_equal(out[1][[1]], test_value)
  expect_equal(names(out), test_name)

  out <- GeoJSON_to_sf(pts)
  expect_s3_class(out, "sf")
  expect_equal(out[1][[1]], test_value)
  expect_equal(names(out)[1], test_name)
})

test_that("geometry column name is preserved", {
  d <- data.frame(a = 1:2)
  d$SHAPE <- c("POINT(0 0)", "POINT(0 1)")
  df <- sf::st_as_sf(d, wkt = "SHAPE")
  out <- ms_filter_fields(df, "a")
  expect_equal(attr(out, "sf_column"), "SHAPE")
})

test_that("NA values dealt with in sf_to_GeoJSON and GeoJSON_to_sf", {
  sf_obj <- sf::st_sf(
    a = c(1.0, NA_real_),
    geometry = sf::st_as_sfc(c("POINT(0 0)", "POINT(1 1)"))
  )
  geojson <- sf_to_GeoJSON(sf_obj)
  expect_snapshot_value(geojson, style = "json2")

  back_to_sf <- GeoJSON_to_sf(geojson, crs = attr(geojson, "crs"))
  expect_equal(sf_obj, back_to_sf)
})

test_that("utilities for checking v8 engine work", {
  expect_s3_class(v8_version(), "numeric_version")
})


test_that("an sf data frame with no columns works", {
  points <- geojsonsf::geojson_sf(
    "{\"type\":\"FeatureCollection\",\"features\":[{\"type\":\"Feature\",\"properties\":{},\"geometry\":{\"type\":\"Point\",\"coordinates\":[53.7702,-40.4873]}},{\"type\":\"Feature\",\"properties\":{},\"geometry\":{\"type\":\"Point\",\"coordinates\":[58.0202,-43.634]}}]}"
  )
  expect_silent(ms_sf(points, "-info"))
  expect_true(grepl("\"type\":\"FeatureCollection\"", sf_to_GeoJSON(points)))
})

test_that("sys_mapshaper works with spaces in path (#107)", {
  skip_if_not(has_sys_mapshaper())
  geojson <- "{\"type\":\"FeatureCollection\",\"features\":[{\"type\":\"Feature\",\"properties\":{\"a\":1},\"geometry\":{\"type\":\"Point\",\"coordinates\":[0,0]}},{\"type\":\"Feature\",\"properties\":{\"a\":null},\"geometry\":{\"type\":\"Point\",\"coordinates\":[1,1]}}]}"
  poly <- '{
"type": "Feature",
"properties": {},
"geometry": {
"type": "Polygon",
"coordinates": [
[
[0.5, 1.5],
[1.5, 1.5],
[1.5, 0.5],
[0.5, 0.5],
[0.5, 1.5]
]
]
}
}'
  withr::local_options(ms_tempdir = file.path(tempdir(), "path with. spaces"))
  expect_match(temp_geojson(), "path with. spaces")
  expect_silent(sys_mapshaper(geojson, poly, command = "-clip"))
})

test_that("ms_de_unit works", {
  skip_if_not_installed("units")
  df <- datasets::iris
  df$Sepal.Length <- units::as_units(df$Sepal.Length, "mm")
  expect_true(inherits(df$Sepal.Length, "units"))
  expect_warning(out <- ms_de_unit(df), "units")
  expect_false(any(vapply(out, inherits, "units", FUN.VALUE = logical(1))))
})

test_that("ms_lines works with all classes", {
  out_json <- ms_lines(innerlines_poly)
  expect_s3_class(out_json, "json")
  expect_snapshot_value(out_json, style = "json2")
  expect_s3_class(ms_lines(unclass(innerlines_poly)), "geojson")

  expected_sp <- GeoJSON_to_sp(out_json)
  expect_equivalent(ms_lines(innerlines_poly_spdf), expected_sp)
  expect_equivalent(
    ms_lines(innerlines_poly_sp),
    as(expected_sp, "SpatialLines")
  )

  expected_sf <- read_sf(unclass(out_json))
  expected_sfc <- st_geometry(expected_sf)

  expect_equivalent(st_geometry(ms_lines(innerlines_poly_sf)), expected_sfc)
  expect_equivalent(ms_lines(innerlines_poly_sfc), expected_sfc)
})

test_that("ms_lines works with fields specified", {
  out_json <- ms_lines(innerlines_poly, "foo")
  expect_s3_class(out_json, "geojson")
  expect_snapshot_value(out_json, style = "json2")

  expected_sp <- GeoJSON_to_sp(out_json)

  expect_equivalent(ms_lines(innerlines_poly_spdf, "foo"), expected_sp)
  expect_equivalent(
    ms_lines(innerlines_poly_sf, "foo")$RANK,
    c(2, 2, 1, 1, 0, 0, 0, 0)
  )
})

test_that("ms_lines errors correctly", {
  expect_error(ms_lines('{foo: "bar"}'), "Input is not valid geojson")
  # Don't test this as the V8 error throws a warning
  expect_warning(
    ms_lines(innerlines_poly, "bar"),
    "The command returned an empty response"
  )
  expect_error(
    ms_lines(innerlines_poly_spdf, "bar"),
    "not all fields specified exist in input data"
  )
  expect_error(
    ms_lines(innerlines_poly, 1),
    "fields must be a character vector"
  )
  expect_error(
    ms_lines(innerlines_poly, force_FC = "true"),
    "force_FC must be TRUE or FALSE"
  )

  expect_error(
    ms_lines(innerlines_poly_sfc, "foo"),
    "Do not specify fields for sfc classes"
  )
})

test_that("ms_innerlines works with sys = TRUE", {
  skip_if_not(has_sys_mapshaper())
  expect_s3_class(ms_lines(innerlines_poly, sys = TRUE), "geojson")
  expect_snapshot_value(ms_lines(innerlines_poly, sys = TRUE), style = "json2")
  expect_s4_class(
    ms_lines(innerlines_poly_spdf, sys = TRUE),
    "SpatialLinesDataFrame"
  )
  expect_s3_class(ms_lines(innerlines_poly_sf, sys = TRUE), "sf")
})

test_that("ms_clip.geojson works", {
  skip_on_old_v8()
  default_clip_json <- ms_clip(ce_poly, ce_inner_poly)

  expect_s3_class(default_clip_json, "geojson")
  expect_snapshot_value(default_clip_json, style = "json2")
  expect_true(jsonify::validate_json(default_clip_json))

  expect_snapshot_value(
    ms_clip(ce_poly, ce_inner_poly, force_FC = FALSE),
    style = "json2"
  )

  skip_if_not(has_sys_mapshaper())
  out_sys <- ms_clip(ce_poly, ce_inner_poly, sys = TRUE)
  expect_s3_class(out_sys, "geojson")
  expect_snapshot_value(out_sys, style = "json2")

  out_sys_nofc <- ms_clip(ce_poly, ce_inner_poly, sys = TRUE, force_FC = FALSE)
  expect_s3_class(out_sys_nofc, "geojson")
  expect_snapshot_value(out_sys_nofc, style = "json2")
})

test_that("ms_clip.character works", {
  skip_on_old_v8()
  default_clip_json <- ms_clip(unclass(ce_poly), unclass(ce_inner_poly))

  expect_s3_class(default_clip_json, "geojson")
  expect_snapshot_value(default_clip_json, style = "json2")
  expect_true(jsonify::validate_json(default_clip_json))
})

test_that("ms_erase.geojson works", {
  skip_on_old_v8()
  default_erase_json <- ms_erase(ce_poly, ce_inner_poly)

  expect_s3_class(default_erase_json, "geojson")
  expect_snapshot_value(default_erase_json, style = "json2")
  expect_true(jsonify::validate_json(default_erase_json))

  expect_snapshot_value(
    ms_erase(ce_poly, ce_inner_poly, force_FC = FALSE),
    style = "json2"
  )

  skip_if_not(has_sys_mapshaper())
  expect_s3_class(ms_erase(ce_poly, ce_inner_poly, sys = TRUE), "geojson")
})

test_that("ms_erase.character works", {
  skip_on_old_v8()
  default_erase_json <- ms_erase(unclass(ce_poly), unclass(ce_inner_poly))

  expect_s3_class(default_erase_json, "geojson")
  expect_snapshot_value(default_erase_json, style = "json2")
  expect_true(jsonify::validate_json(default_erase_json))
})

## Spatial Classes
test_that("ms_clip.SpatialPolygons works", {
  skip_on_old_v8()
  default_clip_spdf <- ms_clip(ce_poly_spdf, ce_inner_poly_spdf)

  expect_s4_class(default_clip_spdf, "SpatialPolygonsDataFrame")
  expect_equivalent(
    sapply(
      default_clip_spdf@polygons[[1]]@Polygons,
      function(x) length(x@coords)
    ),
    10
  )
  expect_true(sf::st_is_valid(sf::st_as_sf(default_clip_spdf)))

  default_clip_sp <- ms_clip(ce_poly_sp, ce_inner_poly_spdf)
  expect_equivalent(as(default_clip_spdf, "SpatialPolygons"), default_clip_sp)

  skip_if_not(has_sys_mapshaper())
  expect_s4_class(
    ms_clip(ce_poly_spdf, ce_inner_poly_spdf, sys = TRUE),
    "SpatialPolygonsDataFrame"
  )
})

test_that("ms_erase.SpatialPolygons works", {
  skip_on_old_v8()
  default_erase_spdf <- ms_erase(ce_poly_spdf, ce_inner_poly_spdf)

  expect_s4_class(default_erase_spdf, "SpatialPolygonsDataFrame")
  expect_equivalent(
    sapply(
      default_erase_spdf@polygons[[1]]@Polygons,
      function(x) length(x@coords)
    ),
    14
  )
  expect_true(sf::st_is_valid(sf::st_as_sf(default_erase_spdf)))

  default_erase_sp <- ms_erase(ce_poly_sp, ce_inner_poly_spdf)
  expect_equivalent(as(default_erase_spdf, "SpatialPolygons"), default_erase_sp)

  skip_if_not(has_sys_mapshaper())
  expect_s4_class(
    ms_erase(ce_poly_spdf, ce_inner_poly_spdf, sys = TRUE),
    "SpatialPolygonsDataFrame"
  )
})

test_that("ms_clip works with lines", {
  skip_on_old_v8()
  expected_out <- structure(
    '{"type":"FeatureCollection", "features": [
{"type":"Feature","geometry":{"type":"LineString","coordinates":[[55,-40.125],[52,-42]]},"properties":null}
]}',
    class = c("json", "geojson")
  )

  expect_snapshot_value(ms_clip(ce_line, ce_inner_poly), style = "json2")
  expect_equivalent(
    ms_clip(ce_line_spdf, GeoJSON_to_sp(ce_inner_poly)),
    GeoJSON_to_sp(expected_out)
  )
  expect_equivalent(
    ms_clip(ce_line_sp, GeoJSON_to_sp(ce_inner_poly)),
    as(GeoJSON_to_sp(expected_out), "SpatialLines")
  )
  expect_snapshot_value(
    ms_clip(ce_line, bbox = c(51, -45, 55, -40)),
    style = "json2"
  )
  expect_equivalent(
    ms_clip(ce_line_spdf, bbox = c(51, -45, 55, -40)),
    GeoJSON_to_sp(expected_out)
  )
  expect_equivalent(
    ms_clip(ce_line_sp, bbox = c(51, -45, 55, -40)),
    as(GeoJSON_to_sp(expected_out), "SpatialLines")
  )
})

test_that("ms_erase works with lines", {
  skip_on_old_v8()
  expected_out <- structure(
    '{"type":"FeatureCollection", "features": [
{"type":"Feature","geometry":{"type":"LineString","coordinates":[[60,-37],[55,-40.125]]},"properties":null}
]} ',
    class = c("geojson", "json")
  )

  expect_snapshot_value(ms_erase(ce_line, ce_inner_poly), style = "json2")
  expect_equivalent(
    ms_erase(ce_line_spdf, GeoJSON_to_sp(ce_inner_poly)),
    GeoJSON_to_sp(expected_out)
  )
  expect_equivalent(
    ms_erase(ce_line_sp, GeoJSON_to_sp(ce_inner_poly)),
    as(GeoJSON_to_sp(expected_out), "SpatialLines")
  )
  expect_snapshot_value(
    ms_erase(ce_line, bbox = c(51, -45, 55, -40)),
    style = "json2"
  )
  expect_equivalent(
    ms_erase(ce_line_spdf, bbox = c(51, -45, 55, -40)),
    GeoJSON_to_sp(expected_out)
  )
  expect_equivalent(
    ms_erase(ce_line_sp, bbox = c(51, -45, 55, -40)),
    as(GeoJSON_to_sp(expected_out), "SpatialLines")
  )
})

test_that("ms_clip works with points", {
  skip_on_old_v8()
  expected_out <- structure(
    '{"type":"FeatureCollection", "features": [
{"type":"Feature","geometry":{"type":"Point","coordinates":[53,-42]},"properties":null}
]}',
    class = c("geojson", "json")
  )

  expect_snapshot_value(ms_clip(ce_points, ce_inner_poly), style = "json2")
  expect_equivalent(
    ms_clip(ce_points_spdf, GeoJSON_to_sp(ce_inner_poly)),
    GeoJSON_to_sp(expected_out)
  )
  expect_equivalent(
    ms_clip(ce_points_sp, GeoJSON_to_sp(ce_inner_poly)),
    as(GeoJSON_to_sp(expected_out), "SpatialPoints")
  )
  expect_snapshot_value(
    ms_clip(ce_points, bbox = c(51, -45, 55, -40)),
    style = "json2"
  )
  expect_equivalent(
    ms_clip(ce_points_spdf, bbox = c(51, -45, 55, -40)),
    GeoJSON_to_sp(expected_out)
  )
  expect_equivalent(
    ms_clip(ce_points_sp, bbox = c(51, -45, 55, -40)),
    as(GeoJSON_to_sp(expected_out), "SpatialPoints")
  )
})

test_that("ms_erase works with points", {
  skip_on_old_v8()
  expected_out <- structure(
    '{"type":"FeatureCollection", "features": [
{"type":"Feature","geometry":{"type":"Point","coordinates":[57,-42]},"properties":null}
]}',
    class = c("geojson", "json")
  )

  expect_snapshot_value(ms_erase(ce_points, ce_inner_poly), style = "json2")
  expect_equivalent(
    ms_erase(ce_points_spdf, GeoJSON_to_sp(ce_inner_poly)),
    GeoJSON_to_sp(expected_out)
  )
  expect_equivalent(
    ms_erase(ce_points_sp, GeoJSON_to_sp(ce_inner_poly)),
    as(GeoJSON_to_sp(expected_out), "SpatialPoints")
  )
  expect_snapshot_value(
    ms_erase(ce_points, bbox = c(51, -45, 55, -40)),
    style = "json2"
  )
  expect_equivalent(
    ms_erase(ce_points_spdf, bbox = c(51, -45, 55, -40)),
    GeoJSON_to_sp(expected_out)
  )
  expect_equivalent(
    ms_erase(ce_points_sp, bbox = c(51, -45, 55, -40)),
    as(GeoJSON_to_sp(expected_out), "SpatialPoints")
  )
})

test_that("bbox works", {
  skip_on_old_v8()
  out <- ms_erase(ce_poly, bbox = c(51, -45, 55, -40))
  expect_s3_class(out, "geojson")
  expect_snapshot_value(out, style = "json2")
  out <- ms_clip(ce_poly, bbox = c(51, -45, 55, -40))
  expect_s3_class(out, "geojson")
  expect_snapshot_value(out, style = "json2")

  expect_error(ms_erase(ce_poly), "You must specificy either a bounding box")
  expect_error(
    ms_erase(ce_poly, "foo", c(1, 2, 3, 4)),
    "Please only specify either a bounding box"
  )
  expect_error(
    ms_clip(ce_poly, bbox = c(1, 2, 3)),
    "bbox must be a numeric vector of length 4"
  )
  expect_error(
    ms_clip(ce_poly, bbox = c("a", "b", "c", "d")),
    "bbox must be a numeric vector of length 4"
  )

  skip_if_not(has_sys_mapshaper())
  expect_s3_class(
    ms_clip(ce_poly, bbox = c(51, -45, 55, -40), sys = TRUE),
    "geojson"
  )
  expect_s3_class(
    ms_erase(ce_poly, bbox = c(51, -45, 55, -40), sys = TRUE),
    "geojson"
  )
})

## test sf classes

test_that("clip works with sf objects", {
  skip_on_old_v8()
  expect_s3_class(ms_clip(ce_poly_sf, ce_inner_poly_sf), "sf")
  expect_equivalent(names(ms_clip(ce_poly_sf, ce_inner_poly_sf)), "geometry")
  expect_s3_class(ms_clip(ce_poly_sfc, ce_inner_poly_sf), "sfc")
  expect_s3_class(ms_clip(ce_lines_sf, ce_inner_poly_sf), "sf")
  expect_s3_class(ms_clip(ce_points_sf, ce_inner_poly_sf), "sf")
  expect_s3_class(ms_clip(ce_poly_sf, bbox = c(51, -45, 55, -40)), "sf")

  skip_if_not(has_sys_mapshaper())
  expect_s3_class(ms_clip(ce_poly_sf, ce_inner_poly_sf, sys = TRUE), "sf")
})

test_that("erase works with sf objects", {
  skip_on_old_v8()
  expect_s3_class(ms_erase(ce_poly_sf, ce_inner_poly_sf), "sf")
  expect_equivalent(names(ms_erase(ce_poly_sf, ce_inner_poly_sf)), "geometry")
  expect_s3_class(ms_erase(ce_poly_sfc, ce_inner_poly_sf), "sfc")
  expect_s3_class(ms_erase(ce_lines_sf, ce_inner_poly_sf), "sf")
  expect_s3_class(ms_erase(ce_points_sf, ce_inner_poly_sf), "sf")
  expect_s3_class(ms_erase(ce_poly_sf, bbox = c(51, -45, 55, -40)), "sf")

  skip_if_not(has_sys_mapshaper())
  expect_s3_class(ms_erase(ce_poly_sf, ce_inner_poly_sf, sys = TRUE), "sf")
})

test_that("clip and erase fail properly", {
  skip_on_old_v8()
  err_msg <- "must be an sf or sfc object with POLYGON or MULTIPLOYGON geometry"
  expect_error(ms_clip(ce_points_sf, ce_inner_poly_spdf), err_msg)
  expect_error(ms_erase(ce_points_sf, ce_inner_poly_spdf), err_msg)
  expect_error(ms_clip(ce_poly_sf, ce_points_sf), err_msg)
  expect_error(ms_erase(ce_poly_sf, ce_points_sf), err_msg)
})

test_that("ms_clip and ms_erase fail with old v8", {
  skip_if_not(v8_version() < "6", "Not old v8")
  expect_error(ms_clip(ce_poly, ce_inner_poly))
  expect_error(ms_erase(ce_poly, ce_inner_poly))
})

test_that("error occurs when non-identical crs in sf", {
  skip_on_old_v8()
  diff_crs <- sf::st_transform(ce_inner_poly_sf, 3005)
  expect_error(
    ms_clip(ce_poly_sf, diff_crs),
    "target and clip do not have identical CRS"
  )
  expect_error(
    ms_erase(ce_poly_sf, diff_crs),
    "target and erase do not have identical CRS"
  )
})

multi_poly <- structure(
  '{"type":"FeatureCollection","features":[{"type":"Feature","properties":{},"geometry":{
"type": "MultiPolygon",
"coordinates": [[[[102.0, 2.0], [103.0, 2.0], [103.0, 3.0], [102.0, 3.0],
[102.0, 2.0]]], [[[100.0, 0.0], [101.0, 0.0], [101.0, 1.0], [100.0, 1.0],
[100.0, 0.0]]]]
}}]}',
  class = c("geojson", "json")
)

multi_line <- structure(
  '{"type":"FeatureCollection","features":[{"type":"Feature","properties":{},"geometry":{
"type": "MultiLineString",
"coordinates": [[[102.0, 2.0], [103.0, 2.0], [103.0, 3.0], [102.0, 3.0]],
[[100.0, 0.0], [101.0, 0.0], [101.0, 1.0], [100.0, 1.0]]]
}}]}',
  class = c("geojson", "json")
)

multi_point <- structure(
  '{"type":"FeatureCollection","features":[{"type":"Feature","properties":{},"geometry":{
  "type": "MultiPoint","coordinates": [ [100.0, 0.0], [101.0, 1.0] ]}}]}',
  class = c("geojson", "json")
)

test_that("ms_explode.geojson works", {
  out <- ms_explode(multi_poly)
  expect_s3_class(out, "geojson")
  expect_equal(nrow(geojson_sf(out)), 2)
  expect_snapshot_value(out, style = "json2")
  skip_if_not(has_sys_mapshaper())
  expect_s3_class(ms_explode(multi_poly, sys = TRUE), "geojson")
})

test_that("ms_explode.geojson errors correctly", {
  expect_error(ms_explode('{foo: "bar"}'), "Input is not valid geojson")
})

test_that("ms_explode.character works", {
  out <- ms_explode(unclass(multi_poly))
  expect_s3_class(out, "geojson")
  expect_equal(nrow(geojson_sf(out)), 2)
  expect_snapshot_value(out, style = "json2")
})

test_that("ms_explode.SpatialPolygonsDataFrame works", {
  spdf <- GeoJSON_to_sp(multi_poly)
  out <- ms_explode(spdf)
  expect_s4_class(out, "SpatialPolygonsDataFrame")
  # Temporarily remove due to bug in GDAL 2.1.0
  expect_equal(length(out@polygons), 2)
  skip_if_not(has_sys_mapshaper())
  expect_s4_class(ms_explode(spdf, sys = TRUE), "SpatialPolygonsDataFrame")
})

test_that("ms_explode works with lines", {
  expect_snapshot_value(ms_explode(multi_line), style = "json2")

  sp_lines <- GeoJSON_to_sp(multi_line)
  out <- ms_explode(sp_lines)
  out_disagg <- sp::disaggregate(sp_lines)
  expect_equivalent(
    lapply(out@lines, function(x) x@Lines[[1]]@coords),
    lapply(out_disagg@lines, function(x) x@Lines[[1]]@coords)
  )
  skip_if_not(has_sys_mapshaper())
  expect_s4_class(ms_explode(sp_lines, sys = TRUE), "SpatialLinesDataFrame")
})

test_that("ms_explode works with points", {
  expect_snapshot_value(ms_explode(multi_point), style = "json2")
})


test_that("ms_explode works with sf", {
  mp_sf <- read_sf(multi_point)
  out_sf <- ms_explode(mp_sf)
  expect_s3_class(out_sf, "sf")
  expect_equal(nrow(out_sf), 2)
  expect_s3_class(st_geometry(out_sf), "sfc_POINT")

  mp_sfc <- st_geometry(mp_sf)
  out_sfc <- ms_explode(mp_sfc)
  expect_s3_class(out_sfc, "sfc_POINT")
  expect_equal(length(out_sfc), 2)

  skip_if_not(has_sys_mapshaper())
  expect_s3_class(ms_explode(mp_sf, sys = TRUE), "sf")
})

test_that("apply_mapshaper_commands doesn't delete local file", {
  skip_if_not(has_sys_mapshaper())
  geojson <- basic_poly()
  testfile <- "test.geojson"
  withr::local_file(testfile)
  writeLines(geojson, testfile)
  apply_mapshaper_commands(testfile, "--dissolve", force_FC = FALSE, sys = TRUE)
  expect_true(file.exists(testfile))
})

test_that("ms_innerlines works with all classes", {
  out_json <- ms_innerlines(innerlines_poly)
  expect_s3_class(out_json, "geojson")
  expect_snapshot_value(out_json, style = "json2")
  expect_s3_class(ms_innerlines(unclass(innerlines_poly)), "geojson")

  expected_sp <- as(GeoJSON_to_sp(out_json), "SpatialLines")

  expect_equivalent(ms_innerlines(innerlines_poly_spdf), expected_sp)
  expect_equivalent(ms_innerlines(innerlines_poly_sp), expected_sp)

  expected_sf <- read_sf(unclass(out_json))
  expect_equivalent(ms_innerlines(innerlines_poly_sf), expected_sf)
  expect_equivalent(
    ms_innerlines(innerlines_poly_sfc),
    st_geometry(expected_sf)
  )
})

test_that("ms_innerlines errors correctly", {
  expect_error(ms_innerlines('{foo: "bar"}'), "Input is not valid geojson")
  expect_error(
    ms_innerlines(innerlines_poly, force_FC = "true"),
    "force_FC must be TRUE or FALSE"
  )
  # Don't test this as the V8 error throws a warning
  expect_warning(
    ms_innerlines(ms_lines(innerlines_poly)),
    "The command returned an empty response"
  )
})

test_that("ms_innerlines works with sys = TRUE", {
  skip_if_not(has_sys_mapshaper())
  expect_s3_class(ms_innerlines(innerlines_poly, sys = TRUE), "geojson")
  expect_snapshot_value(
    ms_innerlines(innerlines_poly, sys = TRUE),
    style = "json2"
  )
  expect_snapshot_value(
    ms_innerlines(innerlines_poly, sys = TRUE, force_FC = FALSE),
    style = "json2"
  )

  expect_s4_class(
    ms_innerlines(innerlines_poly_spdf, sys = TRUE),
    "SpatialLines"
  )
  expect_s3_class(ms_innerlines(innerlines_poly_sf, sys = TRUE), "sf")
})

poly_geojson <- structure(
  "{\"type\":\"FeatureCollection\",\"features\":[{\"type\":\"Feature\",\"geometry\":{\"type\":\"Polygon\",\"coordinates\":[[[-76.3,-49.68],[-75.53,-51.13],[-74.71,-56.89],[-84.11,-57.09],[-77.9,-50.62],[-84.12,-49.59],[-76.3,-49.68]]]},\"properties\":{\"x\": -78, \"y\": -53}},{\"type\":\"Feature\",\"geometry\":{\"type\":\"Polygon\",\"coordinates\":[[[-68.77,69.82],[-66.26,62.96],[-74.22,60.87],[-74.12,65.22],[-74.55,65.81],[-75.66,67.03],[-68.77,69.82]]]},\"properties\":{\"x\": -71, \"y\": 65}},{\"type\":\"Feature\",\"geometry\":{\"type\":\"Polygon\",\"coordinates\":[[[136.27,65.8],[137.78,64.03],[140.03,59.56],[139.48,56.48],[133.64,62.44],[129.67,69.6],[136.27,65.8]]]},\"properties\":{\"x\": 135, \"y\": 65}}]}",
  class = c("geojson", "json")
)

poly_spdf <- GeoJSON_to_sp(poly_geojson)
poly_sp <- as(poly_spdf, "SpatialPolygons")


poly_sf <- st_as_sf(poly_spdf)
poly_sfc <- st_geometry(poly_sf)

test_that("ms_points works with defaults", {
  out_json <- ms_points(poly_geojson)
  expect_s3_class(out_json, "geojson")
  expect_snapshot_value(out_json, style = "json2")

  expect_s3_class(ms_points(unclass(poly_geojson)), "geojson")

  expected_sp <- GeoJSON_to_sp(out_json)

  expect_equivalent(ms_points(poly_spdf), expected_sp)
  expect_equivalent(ms_points(poly_sp), as(expected_sp, "SpatialPoints"))

  skip_if_not(has_sys_mapshaper())
  out_sys_json <- ms_points(poly_geojson, sys = TRUE)
  expect_s3_class(out_sys_json, "geojson")
  expect_snapshot_value(out_sys_json, style = "json2")
  expect_s3_class(ms_points(unclass(poly_geojson), sys = TRUE), "geojson")
  expect_s4_class(ms_points(poly_spdf, sys = TRUE), "SpatialPointsDataFrame")
  expect_s4_class(ms_points(poly_sp, sys = TRUE), "SpatialPoints")
})

test_that("ms_points works with defaults with sf", {
  out_sf <- ms_points(poly_sf)
  expect_s3_class(out_sf, "sf")
  expect_equal(as.character(st_geometry_type(out_sf)), rep("POINT", 3))
  expect_named(out_sf, c("x", "y", "geometry"))

  expect_equivalent(ms_points(poly_sfc), st_geometry(out_sf))

  skip_if_not(has_sys_mapshaper())
  out_sys_sf <- ms_points(poly_sf, sys = TRUE)
  expect_s3_class(out_sys_sf, "sf")
  expect_equal(out_sys_sf, out_sf)
  expect_s3_class(ms_points(poly_sfc, sys = TRUE), "sfc")
})

test_that("ms_points works with location=centroid", {
  out_json <- ms_points(poly_geojson, location = "centroid")
  expect_s3_class(out_json, "geojson")
  expect_snapshot_value(out_json, "json2")

  expected_sp <- GeoJSON_to_sp(out_json)

  expect_equal(out_json, ms_points(poly_geojson))
  expect_equivalent(ms_points(poly_spdf, location = "centroid"), expected_sp)

  expected_sf <- st_read(out_json, quiet = TRUE, stringsAsFactors = FALSE)[1:2]
  expect_equivalent(ms_points(poly_sf, location = "centroid"), expected_sf)
  expect_equivalent(
    ms_points(poly_sfc, location = "centroid"),
    st_geometry(expected_sf)
  )
})

test_that("ms_points works with location=inner", {
  out_json <- ms_points(poly_geojson, location = "inner")
  expect_s3_class(out_json, "geojson")
  expect_snapshot_value(out_json, style = "json2")

  expected_sp <- GeoJSON_to_sp(out_json)

  expect_equivalent(ms_points(poly_spdf, location = "inner"), expected_sp)

  expected_sf <- st_read(out_json, quiet = TRUE, stringsAsFactors = FALSE)[1:2]
  expect_equal(ms_points(poly_sf, location = "inner"), expected_sf)
  expect_equal(
    ms_points(poly_sfc, location = "inner"),
    st_geometry(expected_sf)
  )

  skip_if_not(has_sys_mapshaper())
  out_sys_sf <- ms_points(poly_sf, location = "inner", sys = TRUE)
  expect_s3_class(out_sys_sf, "sf")
  expect_equal(out_sys_sf, expected_sf)
  expect_s3_class(ms_points(poly_sfc, location = "inner", sys = TRUE), "sfc")
})

test_that("ms_points works with x and y", {
  out_json <- ms_points(poly_geojson, x = "x", y = "y")
  expect_s3_class(out_json, "geojson")
  expect_snapshot_value(out_json, "json2")

  expected_sp <- GeoJSON_to_sp(out_json)

  expect_equivalent(ms_points(poly_spdf, x = "x", y = "y"), expected_sp)

  out_sf <- ms_points(poly_sf, x = "x", y = "y")
  expect_equivalent(
    out_sf,
    st_read(out_json, quiet = TRUE, stringsAsFactors = FALSE)[1:2]
  )

  skip_if_not(has_sys_mapshaper())
  out_sys_sf <- ms_points(poly_sf, x = "x", y = "y", sys = TRUE)
  expect_s3_class(out_sys_sf, "sf")
  expect_equal(out_sys_sf, out_sf)
})

test_that("ms_points fails correctly", {
  expect_error(
    ms_points(poly_geojson, location = "foo"),
    "location must be 'centroid' or 'inner'"
  )
  expect_error(
    ms_points(poly_geojson, location = "inner", x = "x", y = "y"),
    "You have specified both a location and x/y for point placement"
  )
  expect_error(
    ms_points(poly_geojson, location = "inner", x = "x"),
    "You have specified both a location and x/y for point placement"
  )
  expect_error(
    ms_points(poly_geojson, location = "inner", y = "y"),
    "You have specified both a location and x/y for point placement"
  )
  expect_error(ms_points(poly_geojson, x = "x"), "Only one of x/y pair found")
  expect_error(ms_points(poly_geojson, y = "y"), "Only one of x/y pair found")
  expect_error(
    ms_points(poly_geojson, force_FC = "true"),
    "force_FC must be TRUE or FALSE"
  )

  # sfc and SpatialPolygons (i.e., no attributes)
  expect_error(
    ms_points(poly_sfc, x = "x", y = "y"),
    "Objects of class sfc have no columns"
  )

  expect_error(
    ms_points(poly_sp, x = "x", y = "y"),
    "SpatialPolygons objects do not have columns"
  )
})

poly_simp <- structure(
  '{"type":"FeatureCollection","features":[{"type":"Feature","geometry":{"type":"Polygon","coordinates":[[[-7.1549869,45.4449053],[-7.6245498,37.9890775],[-7.5290969,38.0423402],[-3.3235845,40.588151],[-7.344442,37.6863061],[1.8042184,41.0097841],[3.7578538,38.7756389],[1.8629117,35.5400723],[-6.3787009,28.8026166],[-8.3144042,35.6271496],[-9.3413257,34.4122375],[-7.8818739,37.2784218],[-10.970619,35.0652943],[-7.855486,37.303094],[-17.6800154,33.0680873],[-11.4987062,37.7759151],[-16.8542278,41.7896373],[-9.6292336,41.0325088],[-8.3619054,39.5168442],[-8.1027301,39.7855456],[-7.1549869,45.4449053]]]},"properties":{}}]}',
  class = c("geojson", "json")
)
line_simp <- structure(
  '{"type":"FeatureCollection", "features": [{"type":"Feature","geometry":{"type":"LineString","coordinates":[[-146.030845,-17.697398],[-138.8493372,-17.938697],[-137.5671055,-18.9589785],[-146.3153242,-20.8865269],[-142.9518755,-24.359833],[-147.6422817,-20.9477376],[-146.6957993,-24.7101963],[-147.696223,-21.1162469],[-156.2250727,-21.2045764],[-150.6399109,-15.4286993],[-146.030845,-17.697398]]},"properties":{}}]}',
  class = c("geojson", "json")
)

line_simp_spdf <- GeoJSON_to_sp(line_simp)
line_simp_sp <- as(line_simp_spdf, "SpatialLines")

poly_simp_spdf <- GeoJSON_to_sp(poly_simp)
poly_simp_sp <- as(poly_simp_spdf, "SpatialPolygons")

multiple_poly_simp <- structure(
  '
{
  "type": "FeatureCollection",
  "features": [
  {
  "type": "Feature",
  "geometry": {
  "type": "Polygon",
  "coordinates":[[[-152.3433185,-51.1400329],[-144.9301966,-51.453939],[-151.7435349,-55.6886215],[-147.8559534,-56.2197224],[-136.9430457,-58.449077],[-149.8608625,-56.9673288],[-142.1320457,-59.7694693],[-145.8290265,-64.392506],[-153.3221574,-65.1360902],[-159.8131297,-63.4417056],[-168.2902719,-61.7109737],[-159.4759954,-57.921224],[-156.0772551,-56.6289602],[-168.1285495,-57.1318526],[-160.5989367,-56.1920838],[-168.8213077,-55.2001918],[-167.0773342,-52.2966425],[-153.1392916,-55.3074558],[-156.5065133,-47.5779291],[-153.0625187,-50.9189549],[-152.3433185,-51.1400329]]]
  },
  "properties": {}
  },{
  "type": "Feature",
  "geometry": {
  "type": "Polygon",
  "coordinates": [[[70.6011604,-46.7259373],[73.7751031,-51.6831692],[84.6589604,-51.5355248],[86.6562583,-52.0080875],[85.27707,-55.8454444],[73.8683796,-55.8846499],[67.4990576,-59.4842469],[67.4470766,-53.7361917],[62.1855019,-53.3722205],[65.6152052,-48.4648198],[70.6011604,-46.7259373]]]
  },
  "properties": {}
  },{
  "type": "Feature",
  "geometry": {
  "type": "Polygon",
  "coordinates": [[[-167.6460128,-2.4870059],[-164.0218557,4.0942137],[-164.3485321,1.2289566],[-162.3875504,3.219013],[-161.9762018,-7.9484803],[-167.6460128,-2.4870059]]]
  },
  "properties": {}
  }
  ]
  }',
  class = c("geojson", "json")
)

multiple_poly_simp_spdf <- GeoJSON_to_sp(multiple_poly_simp)

multiple_poly_simp_sf <- st_as_sf(multiple_poly_simp_spdf)

test_that("ms_simplify.geojson and character works with defaults", {
  default_simplify_json <- ms_simplify(poly_simp)

  expect_s3_class(default_simplify_json, "geojson")
  expect_snapshot_value(default_simplify_json, style = "json2")
  expect_true(jsonify::validate_json(default_simplify_json))

  expect_equal(ms_simplify(poly_simp), default_simplify_json)

  skip_if_not(has_sys_mapshaper())
  expect_s3_class(ms_simplify(poly_simp, sys = TRUE), "geojson")
})

test_that("ms_simplify.geojson works with different methods", {
  vis_simplify_json <- ms_simplify(poly_simp, method = "vis", weighting = 0)
  dp_simplify_json <- ms_simplify(poly_simp, method = "dp")

  expect_s3_class(vis_simplify_json, "geojson")
  expect_snapshot_value(vis_simplify_json, style = "json2")
  expect_s3_class(dp_simplify_json, "geojson")
  expect_snapshot_value(dp_simplify_json, style = "json2")
})

test_that("ms_simplify.SpatialPolygons works with defaults", {
  default_simplify_spdf <- ms_simplify(poly_simp_spdf)
  default_simplify_sp <- ms_simplify(poly_simp_sp)

  expect_s4_class(default_simplify_spdf, "SpatialPolygonsDataFrame")
  expect_s4_class(default_simplify_sp, "SpatialPolygons")
  expect_equivalent(
    default_simplify_sp,
    as(default_simplify_spdf, "SpatialPolygons")
  )
  expect_true(sf::st_is_valid(sf::st_as_sf(default_simplify_spdf)))

  skip_if_not(has_sys_mapshaper())
  expect_s4_class(
    ms_simplify(poly_simp_spdf, sys = TRUE),
    "SpatialPolygonsDataFrame"
  )
  expect_s4_class(ms_simplify(poly_simp_sp, sys = TRUE), "SpatialPolygons")
})

test_that("simplify.SpatialPolygonsDataFrame works with other methods", {
  vis_simplify_spdf <- ms_simplify(
    poly_simp_spdf,
    method = "vis",
    weighting = 0
  )
  dp_simplify_spdf <- ms_simplify(poly_simp_spdf, method = "dp")

  expect_s4_class(vis_simplify_spdf, "SpatialPolygonsDataFrame")
  expect_true(sf::st_is_valid(sf::st_as_sf(vis_simplify_spdf)))

  expect_s4_class(dp_simplify_spdf, "SpatialPolygonsDataFrame")
  expect_true(sf::st_is_valid(sf::st_as_sf(dp_simplify_spdf)))
})

test_that("exploding works with geojson", {
  multipoly <- structure(
    '{
"type": "MultiPolygon",
"coordinates": [[[[102.0, 2.0], [103.0, 2.0], [103.0, 3.0], [102.0, 3.0],
[102.0, 2.0]]], [[[100.0, 0.0], [101.0, 0.0], [101.0, 1.0], [100.0, 1.0],
[100.0, 0.0]]]]
} ',
    class = c("geojson", "json")
  )
  multi_spdf <- GeoJSON_to_sp(multipoly)

  out <- ms_simplify(multipoly, keep_shapes = TRUE, explode = FALSE)
  expect_snapshot_value(out, style = "json2")
  out <- ms_simplify(multipoly, keep_shapes = TRUE, explode = TRUE)
  expect_snapshot_value(out, style = "json2")

  #SPDF
  out <- ms_simplify(multi_spdf, keep_shapes = TRUE)
  expect_equal(length(out@polygons), 1)
  out <- ms_simplify(multi_spdf, keep_shapes = TRUE, explode = TRUE)
  expect_equal(length(out@polygons), 2)
})

test_that("ms_simplify fails with invalid geojson", {
  expect_error(ms_simplify('{foo: "bar"}'), "Input is not valid geojson")
})

test_that("ms_simplify fails correctly", {
  expect_error(ms_simplify(poly_simp, keep = 0), "keep must be > 0 and <= 1")
  expect_error(ms_simplify(poly_simp, keep = 1.01), "keep must be > 0 and <= 1")
  expect_error(
    ms_simplify(poly_simp, method = "foo"),
    "method should be one of"
  )
})

test_that("ms_simplify works with drop_null_geometries", {
  out_drop <- ms_simplify(
    multiple_poly_simp,
    keep_shapes = FALSE,
    drop_null_geometries = TRUE
  )
  expect_snapshot_value(out_drop, style = "json2")
  out_nodrop <- ms_simplify(
    multiple_poly_simp,
    keep_shapes = FALSE,
    drop_null_geometries = FALSE
  )
  expect_snapshot_value(out_nodrop, style = "json2")
})

test_that("ms_simplify.SpatialPolygonsDataFrame works keep_shapes = FALSE and ignores drop_null_geometries", {
  out <- ms_simplify(
    multiple_poly_simp_spdf,
    keep_shapes = FALSE,
    drop_null_geometries = TRUE
  )
  expect_equal(length(out@polygons), 1)
  out_nodrop <- ms_simplify(
    multiple_poly_simp_spdf,
    keep_shapes = FALSE,
    drop_null_geometries = FALSE
  )
  expect_equivalent(out, out_nodrop)
})

test_that("ms_simplify works with lines", {
  out_json <- ms_simplify(line_simp, keep = 0.1)

  expect_snapshot_value(out_json, style = "json2")
  expect_equivalent(
    ms_simplify(line_simp_spdf, keep = 0.1),
    GeoJSON_to_sp(out_json)
  )
  expect_equivalent(
    ms_simplify(line_simp_sp, keep = 0.1),
    as(ms_simplify(line_simp_spdf, keep = 0.1), "SpatialLines")
  )
})

test_that("ms_simplify works correctly when all geometries are dropped", {
  expect_error(
    ms_simplify(multiple_poly_simp_spdf, keep = 0.001),
    "Cannot convert result to a Spatial\\* object"
  )
  expect_snapshot_value(
    ms_simplify(multiple_poly_simp, keep = 0.001),
    style = "json2"
  )
  expect_snapshot_value(
    ms_simplify(multiple_poly_simp, keep = 0.001, force_FC = FALSE),
    style = "json2"
  )

  skip_if_not(has_sys_mapshaper())
  expect_snapshot_value(
    ms_simplify(multiple_poly_simp, keep = 0.001, sys = TRUE),
    style = "json2"
  )
})

test_that("snap_interval works", {
  poly <- structure(
    '{"type":"FeatureCollection",
  "features":[
                  {"type":"Feature",
                  "properties":{},
                  "geometry":{"type":"Polygon","coordinates":[[
                  [101,2],[101,3],[103,3],[103,2],[102,2],[101,2]
                  ]]}}
                  ,{"type":"Feature",
                  "properties":{},
                  "geometry":{"type":"Polygon","coordinates":[[
                  [101,1],[101,2],[102,1.9],[103,2],[103,1],[101,1]
                  ]]}}]}',
    class = c("geojson", "json")
  )

  poly_not_snapped <- ms_simplify(
    poly,
    keep = 0.8,
    snap = TRUE,
    snap_interval = 0.09
  )
  expect_snapshot_value(poly_not_snapped, style = "json2")

  poly_snapped <- ms_simplify(
    poly,
    keep = 0.8,
    snap = TRUE,
    snap_interval = 0.11
  )
  expect_snapshot_value(poly_snapped, style = "json2")
})

test_that("ms_simplify works with very small values of 'keep", {
  expect_s3_class(ms_simplify(poly_simp, keep = 0.0001), "geojson")
})


# SF ----------------------------------------------------------------------

test_that("ms_simplify works with sf", {
  line_sf <- st_as_sf(line_simp_spdf)
  expect_s3_class(ms_simplify(multiple_poly_simp_sf), "sf")
  expect_s3_class(ms_simplify(line_sf), "sf")

  skip_if_not(has_sys_mapshaper())
  expect_s3_class(ms_simplify(multiple_poly_simp_sf, sys = TRUE), "sf")
})


test_that("ms_simplify works with sfc", {
  poly_sfc <- st_as_sfc(poly_simp_sp)
  line_sfc <- st_as_sfc(line_simp_sp)
  expect_s3_class(ms_simplify(poly_sfc), "sfc_POLYGON")
  # don't simplify too much or goes to empty point
  expect_s3_class(ms_simplify(line_sfc, keep = 0.5), "sfc_LINESTRING")

  skip_if_not(has_sys_mapshaper())
  expect_s3_class(ms_simplify(poly_sfc, sys = TRUE), "sfc_POLYGON")
})


test_that("ms_simplify works with various column types", {
  xs <- st_polygon(list(cbind(
    approx(c(0, 0, 1, 1, 0))$y,
    approx(c(0, 1, 1, 0, 0))$y
  )))
  xsf <- st_sf(geometry = st_sfc(xs, xs + 2, xs + 3), a = 1:3)

  nr <- dim(xsf)[1]
  various_types <- list(
    date = Sys.Date() + seq_len(nr)
    # time = Sys.time() + seq_len(nr)
    # complex(nr),
    # rw = raw(nr),
    # lst = replicate(nr, "a", simplify = FALSE)
  )
  for (itype in seq_along(various_types)) {
    xsf$check_me <- various_types[[itype]]
    simp_xsf <- ms_simplify(xsf)
    expect_s3_class(simp_xsf, "sf")
    ## not currently working for POSIXct
    expect_equal(simp_xsf$check_me, various_types[[itype]], tolerance = 1)
  }

  ## raw special case
  # xsf$check_me <- raw(nr)
  # expect_warning(simp_xsf <- ms_simplify(xsf), "NAs introduced by coercion")
})


# units -------------------------------------------------------------------

test_that("ms_simplify works with sf objects containing units", {
  multiple_poly_simp_sf$area = sf::st_area(multiple_poly_simp_sf)
  expect_warning(
    multipoly_sf_simple <- ms_simplify(multiple_poly_simp_sf),
    "units"
  )
  expect_s3_class(multipoly_sf_simple, "sf")
  expect_type(multipoly_sf_simple$area, "double")
})

test_that("gj2008 flag reverses winding order as expected", {
  # https://github.com/ateucher/rmapshaper/issues/167
  poly_with_hole <- structure(
    '{"type":"FeatureCollection","features":[
                         {"type":"Feature",
                         "geometry":{
                           "type": "Polygon",
                           "coordinates": [
                         [[100.0, 0.0], [100.0, 10.0], [110.0, 10.0], [110.0, 0.0], [100.0, 0.0]],
                         [[101.0, 1.0], [109.0, 1.0], [109.0, 9.0], [101.0, 9.0], [101.0, 1.0]]
                           ]
                         },
                         "properties":{}
                         }]
                         }',
    class = c("geojson", "json")
  )
  expect_snapshot_value(
    ms_simplify(poly_with_hole, keep = 1, gj2008 = FALSE),
    style = "json2"
  )
  expect_snapshot_value(
    ms_simplify(poly_with_hole, keep = 1, gj2008 = TRUE),
    style = "json2"
  )
})

test_that("gj2008 flag reverses winding order as expected with sf", {
  # https://github.com/ateucher/rmapshaper/issues/167
  poly_with_hole <- geojsonsf::geojson_sf(
    structure(
      '{"type":"FeatureCollection","features":[
                         {"type":"Feature",
                         "geometry":{
                           "type": "Polygon",
                           "coordinates": [
                         [[100.0, 0.0], [100.0, 10.0], [110.0, 10.0], [110.0, 0.0], [100.0, 0.0]],
                         [[101.0, 1.0], [109.0, 1.0], [109.0, 9.0], [101.0, 9.0], [101.0, 1.0]]
                           ]
                         },
                         "properties":{}
                         }]
                         }',
      class = c("geojson", "json")
    )
  )
  expect_snapshot_value(
    geojsonsf::sf_geojson(
      ms_simplify(
        poly_with_hole,
        keep = 1,
        gj2008 = FALSE
      )
    ),
    style = "json2"
  )
  expect_snapshot_value(
    geojsonsf::sf_geojson(
      ms_simplify(
        poly_with_hole,
        keep = 1,
        gj2008 = TRUE
      )
    ),
    style = "json2"
  )
})

#' Create a line layer consisting of shared boundaries with no attribute data
#'
#' @param input input polygons object to convert to inner lines. One of:
#' \itemize{
#'  \item \code{geo_json} or \code{character} polygons;
#'  \item \code{SpatialPolygons*};
#'  \item \code{sf} or \code{sfc} polygons object
#'  }
#' @inheritDotParams apply_mapshaper_commands force_FC sys sys_mem quiet gj2008
#'
#' @return lines in the same class as the input layer, but without attributes
#'
#' @examples
#' library(geojsonsf)
#' library(sf)
#'
#' poly <- structure('{"type":"FeatureCollection",
#'             "features":[
#'               {"type":"Feature",
#'                 "properties":{"foo": "a"},
#'                 "geometry":{"type":"Polygon","coordinates":[[
#'                   [102,2],[102,3],[103,3],[103,2],[102,2]
#'                   ]]}}
#'               ,{"type":"Feature",
#'                 "properties":{"foo": "a"},
#'                 "geometry":{"type":"Polygon","coordinates":[[
#'                   [103,3],[104,3],[104,2],[103,2],[103,3]
#'                   ]]}},
#'               {"type":"Feature",
#'                 "properties":{"foo": "b"},
#'                 "geometry":{"type":"Polygon","coordinates":[[
#'                   [102,1],[102,2],[103,2],[103,1],[102,1]
#'                   ]]}},
#'               {"type":"Feature",
#'                 "properties":{"foo": "b"},
#'                 "geometry":{"type":"Polygon","coordinates":[[
#'                   [103,1],[103,2],[104,2],[104,1],[103,1]
#'                   ]]}}]}', class = c("geojson", "json"))
#'
#' poly <- geojson_sf(poly)
#' plot(poly)
#'
#' out <- ms_innerlines(poly)
#' plot(out)
#'
#' @export
ms_innerlines <- function(input, ...) {
  UseMethod("ms_innerlines")
}

#' @export
ms_innerlines.character <- function(input, ...) {
  input <- check_character_input(input)

  apply_mapshaper_commands(data = input, command = "-innerlines", ...)
}

#' @export
ms_innerlines.json <- function(input, ...) {
  apply_mapshaper_commands(data = input, command = "-innerlines", ...)
}

#' @export
ms_innerlines.SpatialPolygons <- function(input, ...) {
  ms_sp(as(input, "SpatialPolygons"), "-innerlines", ...)
}

#' @export
ms_innerlines.sf <- function(input, ...) {
  ms_sf(input, "-innerlines", ...)
}


#' @export
ms_innerlines.sfc <- function(input, ...) {
  ms_sf(input, "-innerlines", ...)
}

#' Delete fields in the attribute table
#'
#' Removes all fields except those listed in the \code{fields} parameter
#'
#' @param input spatial object to filter fields on. One of:
#' \itemize{
#'  \item \code{geo_json} or \code{character} points, lines, or polygons;
#'  \item \code{SpatialPolygonsDataFrame}, \code{SpatialLinesDataFrame}, \code{SpatialPointsDataFrame};
#'  \item \code{sf} object
#'  }
#' @param fields character vector of fields to retain.
#' @inheritDotParams apply_mapshaper_commands sys sys_mem quiet
#'
#' @return object with only specified attributes retained, in the same class as
#'   the input
#'
#' @examples
#' library(geojsonsf)
#' library(sf)
#'
#' poly <- structure("{\"type\":\"FeatureCollection\",
#'                   \"features\":[{\"type\":\"Feature\",
#'                   \"properties\":{\"a\": 1, \"b\":2, \"c\": 3},
#'                   \"geometry\":{\"type\":\"Polygon\",
#'                   \"coordinates\":[[[102,2],[102,4],[104,4],[104,2],[102,2]]]}}]}",
#'                   class = c("geojson", "json"))
#' poly <- geojson_sf(poly)
#' poly
#'
#' # Filter (keep) fields a and b, drop c
#' out <- ms_filter_fields(poly, c("a", "b"))
#' out
#'
#' @export
ms_filter_fields <- function(input, fields, ...) {
  if (!is.character(fields)) stop("fields must be a character vector")
  UseMethod("ms_filter_fields")
}

#' @export
ms_filter_fields.character <- function(input, fields, ...) {
  input <- check_character_input(input)

  cmd <- make_filterfields_call(fields)

  apply_mapshaper_commands(data = input, command = cmd, force_FC = FALSE, ...)
}

#' @export
ms_filter_fields.json <- function(input, fields, ...) {
  cmd <- make_filterfields_call(fields)

  apply_mapshaper_commands(data = input, command = cmd, force_FC = FALSE, ...)
}

#' @export
ms_filter_fields.SpatialPolygonsDataFrame <- function(input, fields, ...) {
  ms_filter_fields_sp(input, fields, ...)
}

#' @export
ms_filter_fields.SpatialPointsDataFrame <- function(input, fields, ...) {
  ms_filter_fields_sp(input, fields, ...)
}

#' @export
ms_filter_fields.SpatialLinesDataFrame <- function(input, fields, ...) {
  ms_filter_fields_sp(input, fields, ...)
}

#' @export
ms_filter_fields.sf <- function(input, fields, ...) {
  if (!all(fields %in% names(input))) {
    stop("Not all fields are in input")
  }

  # call <- make_filterfields_call(fields)
  #
  # ms_sf(input = input, call = call)

  input[, fields, drop = FALSE]
}

ms_filter_fields_sp <- function(input, fields, ...) {
  # cmd <- make_filterfields_call(fields)
  #
  # geojson <- sp_to_GeoJSON(input)
  #
  # ret <- apply_mapshaper_commands(data = geojson, command = cmd, force_FC = FALSE)
  #
  # GeoJSON_to_sp(ret, proj = attr(geojson, "proj"))

  if (!(all(is.element(fields, names(input@data))))) {
    stop("Not all named fields exist in input data", call. = FALSE)
  }

  input@data <- input@data[, fields, drop = FALSE]

  input
}

make_filterfields_call <- function(fields) {
  call <- list("-filter-fields", paste0(fields, collapse = ","))
  call
}

#' Convert polygons to topological boundaries (lines)
#'
#' @param input input polygons object to convert to inner lines. One of:
#' \itemize{
#'  \item \code{geo_json} or \code{character} polygons;
#'  \item \code{SpatialPolygons*};
#'  \item \code{sf} or \code{sfc} polygons object
#'  }
#' @param fields character vector of field names. If left as \code{NULL}
#'   (default), external (unshared) boundaries are attributed as TYPE 0 and
#'   internal (shared) boundaries are TYPE 1. Giving a field name adds an
#'   intermediate level of hierarchy at TYPE 1, with the lowest-level internal
#'   boundaries set to TYPE 2. Supplying a character vector of field names adds
#'   additional levels of hierarchy.
#' @inheritDotParams apply_mapshaper_commands force_FC sys sys_mem quiet gj2008
#'
#' @return topological boundaries as lines, in the same class as the input
#'
#' @examples
#'
#' library(geojsonsf)
#' library(sf)
#'
#' poly <- structure('{"type":"FeatureCollection",
#'              "features":[
#'              {"type":"Feature",
#'              "properties":{"foo": "a"},
#'              "geometry":{"type":"Polygon","coordinates":[[
#'              [102,2],[102,3],[103,3],[103,2],[102,2]
#'              ]]}}
#'              ,{"type":"Feature",
#'              "properties":{"foo": "a"},
#'              "geometry":{"type":"Polygon","coordinates":[[
#'              [103,3],[104,3],[104,2],[103,2],[103,3]
#'              ]]}},
#'              {"type":"Feature",
#'              "properties":{"foo": "b"},
#'              "geometry":{"type":"Polygon","coordinates":[[
#'              [102.5,1],[102.5,2],[103.5,2],[103.5,1],[102.5,1]
#'              ]]}}]}', class = c("geojson", "json"))
#'
#' poly <- geojson_sf(poly)
#' summary(poly)
#' plot(poly)
#'
#' out <- ms_lines(poly)
#' summary(out)
#' plot(out)
#'
#' @export
ms_lines <- function(input, fields = NULL, ...) {
  if (!is.null(fields) && !is.character(fields))
    stop("fields must be a character vector of field names")
  UseMethod("ms_lines")
}

#' @export
ms_lines.character <- function(input, fields = NULL, ...) {
  input <- check_character_input(input)

  command <- make_lines_call(fields)

  apply_mapshaper_commands(data = input, command = command, ...)
}

#' @export
ms_lines.json <- function(input, fields = NULL, ...) {
  command <- make_lines_call(fields)

  apply_mapshaper_commands(data = input, command = command, ...)
}

#' @export
ms_lines.SpatialPolygons <- function(input, fields = NULL, ...) {
  if (.hasSlot(input, "data")) {
    if (!all(fields %in% names(input@data))) {
      stop("not all fields specified exist in input data")
    }
  }

  command <- make_lines_call(fields)

  ms_sp(input, command, ...)
}

#' @export
ms_lines.sf <- function(input, fields = NULL, ...) {
  if (!all(fields %in% names(input))) {
    stop("not all fields specified exist in input data")
  }

  lines_sf(input = input, fields = fields, ...)
}

#' @export
ms_lines.sfc <- function(input, fields = NULL, ...) {
  if (!is.null(fields)) {
    stop("Do not specify fields for sfc classes", call. = FALSE)
  }

  lines_sf(input = input, fields = fields, ...)
}

lines_sf <- function(input, fields, ...) {
  if (!all(sf::st_is(input, c("POLYGON", "MULTIPOLYGON")))) {
    stop("ms_lines only works with (MULTI)POLYGON")
  }

  command <- make_lines_call(fields)

  ms_sf(input, command, ...)
}

make_lines_call <- function(fields) {
  if (!is.null(fields) && !is.character(fields))
    stop("fields must be a character vector of field names")

  fields <- paste0(fields, collapse = ",")

  call <- list("-lines", fields)
  call
}

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @importFrom methods is .hasSlot as slot
#' @importFrom V8 v8
#' @importClassesFrom sp SpatialPoints SpatialLines
## usethis namespace: end
NULL

#' Create points from a polygon layer
#'
#' Can be generated from the polygons by specifying \code{location} to be
#' \code{"centroid"} or \code{"inner"}, OR by specifying fields in the
#' attributes of the layer containing \code{x} and \code{y} coordinates.
#'
#' @param input input polygons object to convert to points. One of:
#' \itemize{
#'  \item \code{geo_json} or \code{character} polygons;
#'  \item \code{SpatialPolygons*};
#'  \item \code{sf} or \code{sfc} polygons object
#'  }
#' @param location either \code{"centroid"} or \code{"inner"}. If
#'   \code{"centroid"}, creates points at the centroid of the largest ring of
#'   each polygon feature. if \code{"inner"}, creates points in the interior of
#'   the largest ring of each polygon feature. Inner points are located away
#'   from polygon boundaries. Must be \code{NULL} if \code{x} and \code{y} are
#'   specified. If left as \code{NULL} (default), will use centroids.
#' @param x name of field containing x coordinate values. Must be \code{NULL} if
#'   \code{location} is specified.
#' @param y name of field containing y coordinate values. Must be \code{NULL} if
#'   \code{location} is specified.
#' @inheritDotParams apply_mapshaper_commands force_FC sys sys_mem quiet gj2008
#'
#' @return points in the same class as the input.
#'
#' @examples
#' library(geojsonsf)
#' library(sf)
#'
#' poly <- structure("{\"type\":\"FeatureCollection\",
#'            \"features\":[{\"type\":\"Feature\",\"properties\":
#'            {\"x_pos\": 1, \"y_pos\": 2},
#'            \"geometry\":{\"type\":\"Polygon\",
#'            \"coordinates\":[[[102,2],[102,4],[104,4],[104,2],[102,2]]]}},
#'            {\"type\":\"Feature\",\"properties\":{\"x_pos\": 3, \"y_pos\": 4},
#'            \"geometry\":{\"type\":\"Polygon\",
#'            \"coordinates\":[[[100,2],[98,4],[101.5,4],[100,2]]]}},
#'            {\"type\":\"Feature\",\"properties\":{\"x_pos\": 5, \"y_pos\": 6},
#'            \"geometry\":{\"type\":\"Polygon\",
#'            \"coordinates\":[[[100,0],[100,1],[101,1],[101,0],[100,0]]]}}]}",
#'            class = c("geojson", "json"))
#'
#' poly <- geojson_sf(poly)
#' summary(poly)
#' plot(poly)
#'
#' # Convert to points using centroids
#' out <- ms_points(poly, location = "centroid")
#' summary(out)
#' plot(out)
#'
#' # Can also specify locations using attributes in the data
#' out <- ms_points(poly, x = "x_pos", y = "y_pos")
#' summary(out)
#' plot(out)
#'
#' @export
ms_points <- function(input, location = NULL, x = NULL, y = NULL, ...) {
  UseMethod("ms_points")
}

#' @export
ms_points.character <- function(
  input,
  location = NULL,
  x = NULL,
  y = NULL,
  ...
) {
  input <- check_character_input(input)

  cmd <- make_points_call(location = location, x = x, y = y)

  apply_mapshaper_commands(data = input, command = cmd, ...)
}

#' @export
ms_points.json <- function(input, location = NULL, x = NULL, y = NULL, ...) {
  cmd <- make_points_call(location = location, x = x, y = y)

  apply_mapshaper_commands(data = input, command = cmd, ...)
}

#' @export
ms_points.SpatialPolygons <- function(
  input,
  location = NULL,
  x = NULL,
  y = NULL,
  ...
) {
  if (!.hasSlot(input, "data") && (!is.null(x) || !is.null(y))) {
    stop(
      "SpatialPolygons objects do not have columns from which to specify x and y coordinates.",
      call. = FALSE
    )
  }

  cmd <- make_points_call(location = location, x = x, y = y)

  ms_sp(input, cmd, ...)
}

#' @export
ms_points.sf <- function(input, location = NULL, x = NULL, y = NULL, ...) {
  if (inherits(input, "sfc") && (!is.null(x) || !is.null(y))) {
    stop(
      "Objects of class sfc have no columns from which to specify x and y coordinates.",
      call. = FALSE
    )
  }

  cmd <- make_points_call(location = location, x = x, y = y)

  ms_sf(input, cmd, ...)
}

#' @export
ms_points.sfc <- ms_points.sf

make_points_call <- function(location, x, y) {
  if (!is.null(location)) {
    if (!location %in% c("centroid", "inner")) {
      stop("location must be 'centroid' or 'inner'")
    }
    if (!is.null(x) || !is.null(y)) {
      stop("You have specified both a location and x/y for point placement")
    }
    call <- list("-points", location)
  } else if (xor(is.null(x), is.null(y))) {
    stop("Only one of x/y pair found")
  } else if (!is.null(x) && !is.null(y)) {
    call <- list("-points", paste0("x=", x), paste0("y=", y))
  } else {
    call <- "-points"
  }

  call
}

## Borrowed verbatim from geojsonio:
## https://github.com/ropensci/geojsonio/blob/321c679057e2dde282e564f3c845dea7f280ccf9/R/as_spatial_methods.R

## SpatialPoints to SpatialPointsDataFrame
as.SpatialPointsDataFrame.SpatialPoints <- function(from) {
  if (!requireNamespace("sp", quietly = TRUE)) {
    stop("Package sp required. Please install.")
  }
  ids <- rownames(slot(from, "coords"))
  if (is.null(ids)) {
    ids <- 1:NROW(slot(from, "coords"))
  }
  df <- data.frame(dummy = rep(0, length(ids)), row.names = ids)
  sp::SpatialPointsDataFrame(from, df)
}

setAs(
  "SpatialPoints",
  "SpatialPointsDataFrame",
  as.SpatialPointsDataFrame.SpatialPoints
)


## SpatialLines to SpatialLinesDataFrame
as.SpatialLinesDataFrame.SpatialLines <- function(from) {
  if (!requireNamespace("sp", quietly = TRUE)) {
    stop("Package sp required. Please install.")
  }
  IDs <- sapply(slot(from, "lines"), function(x) slot(x, "ID"))
  df <- data.frame(dummy = rep(0, length(IDs)), row.names = IDs)
  sp::SpatialLinesDataFrame(from, df)
}

setAs(
  "SpatialLines",
  "SpatialLinesDataFrame",
  as.SpatialLinesDataFrame.SpatialLines
)

#' Aggregate shapes in a polygon or point layer.
#'
#' Aggregates using specified field, or all shapes if no field is given. For point layers,
#' replaces a group of points with their centroid.
#'
#' @param input spatial object to dissolve. One of:
#' \itemize{
#'  \item \code{geo_json} or \code{character} points or polygons;
#'  \item \code{SpatialPolygons}, or \code{SpatialPoints}
#'  }
#' @param snap Snap together vertices within a small distance threshold to fix
#'   small coordinate misalignment in adjacent polygons. Default \code{TRUE}.
#' @param field the field to dissolve on
#' @param sum_fields fields to sum
#' @param copy_fields fields to copy. The first instance of each field will be
#'   copied to the aggregated feature.
#' @param weight Name of an attribute field for generating weighted centroids (points only).
#' @inheritDotParams apply_mapshaper_commands force_FC sys sys_mem quiet gj2008
#'
#' @return the same class as the input
#'
#' @examples
#' library(geojsonsf)
#' library(sf)
#'
#' poly <- structure('{"type":"FeatureCollection",
#'   "features":[
#'   {"type":"Feature",
#'   "properties":{"a": 1, "b": 2},
#'   "geometry":{"type":"Polygon","coordinates":[[
#'   [102,2],[102,3],[103,3],[103,2],[102,2]
#'   ]]}}
#'   ,{"type":"Feature",
#'   "properties":{"a": 5, "b": 3},
#'   "geometry":{"type":"Polygon","coordinates":[[
#'   [100,0],[100,1],[101,1],[101,0],[100,0]
#'   ]]}}]}', class = c("geojson", "json"))
#' poly <- geojson_sf(poly)
#' plot(poly)
#' length(poly)
#' poly
#'
#' # Dissolve the polygon
#' out <- ms_dissolve(poly)
#' plot(out)
#' length(out)
#' out
#'
#' # Dissolve and summing columns
#' out <- ms_dissolve(poly, sum_fields = c("a", "b"))
#' plot(out)
#' out
#'
#' @export
ms_dissolve <- function(
  input,
  field = NULL,
  sum_fields = NULL,
  copy_fields = NULL,
  weight = NULL,
  snap = TRUE,
  ...
) {
  UseMethod("ms_dissolve")
}

#' @export
ms_dissolve.character <- function(
  input,
  field = NULL,
  sum_fields = NULL,
  copy_fields = NULL,
  weight = NULL,
  snap = TRUE,
  ...
) {
  input <- check_character_input(input)

  call <- make_dissolve_call(
    field = field,
    sum_fields = sum_fields,
    weight = weight,
    copy_fields = copy_fields,
    snap = snap
  )

  apply_mapshaper_commands(data = input, command = call, ...)
}

#' @export
ms_dissolve.json <- function(
  input,
  field = NULL,
  sum_fields = NULL,
  copy_fields = NULL,
  weight = NULL,
  snap = TRUE,
  ...
) {
  call <- make_dissolve_call(
    field = field,
    sum_fields = sum_fields,
    weight = weight,
    copy_fields = copy_fields,
    snap = snap
  )

  apply_mapshaper_commands(data = input, command = call, ...)
}

#' @export
ms_dissolve.SpatialPolygons <- function(
  input,
  field = NULL,
  sum_fields = NULL,
  copy_fields = NULL,
  weight = NULL,
  snap = TRUE,
  ...
) {
  dissolve_sp(
    input = input,
    field = field,
    sum_fields = sum_fields,
    copy_fields = copy_fields,
    weight = weight,
    snap = snap,
    ...
  )
}

#' @export
ms_dissolve.SpatialPoints <- function(
  input,
  field = NULL,
  sum_fields = NULL,
  copy_fields = NULL,
  weight = NULL,
  snap = TRUE,
  ...
) {
  dissolve_sp(
    input = input,
    field = field,
    sum_fields = sum_fields,
    copy_fields = copy_fields,
    weight = weight,
    snap = snap,
    ...
  )
}

#' @export
ms_dissolve.sf <- function(
  input,
  field = NULL,
  sum_fields = NULL,
  copy_fields = NULL,
  weight = NULL,
  snap = TRUE,
  ...
) {
  if (!is.null(weight) && !(weight %in% names(input))) {
    stop("specified 'weight' column not present in input data", call. = FALSE)
  }

  dissolve_sf(
    input = input,
    field = field,
    sum_fields = sum_fields,
    copy_fields = copy_fields,
    weight = weight,
    snap = snap,
    ...
  )
}

#' @export
ms_dissolve.sfc <- function(
  input,
  field = NULL,
  sum_fields = NULL,
  copy_fields = NULL,
  weight = NULL,
  snap = TRUE,
  ...
) {
  if (!is.null(weight)) {
    warning(
      "'weight' cannot be used with sfc objects. Ignoring it and proceeding..."
    )
  }

  dissolve_sf(
    input = input,
    field = field,
    sum_fields = sum_fields,
    copy_fields = copy_fields,
    weight = NULL,
    snap = snap,
    ...
  )
}

make_dissolve_call <- function(field, sum_fields, copy_fields, weight, snap) {
  if (is.null(sum_fields)) {
    sum_fields_string <- NULL
  } else {
    sum_fields_string <- paste0(
      "sum-fields=",
      paste0(sum_fields, collapse = ",")
    )
  }

  if (is.null(copy_fields)) {
    copy_fields_string <- NULL
  } else {
    copy_fields_string <- paste0(
      "copy-fields=",
      paste0(copy_fields, collapse = ",")
    )
  }

  if (is.null(weight)) {
    weight_string <- NULL
  } else {
    weight_string <- paste0("weight=", weight)
  }

  if (snap) snap <- "snap" else snap <- NULL

  call <- list(
    snap,
    "-dissolve",
    field,
    sum_fields_string,
    copy_fields_string,
    weight_string
  )

  call
}

dissolve_sp <- function(
  input,
  field,
  sum_fields,
  copy_fields,
  weight,
  snap,
  ...
) {
  if (!inherits(input, "SpatialPointsDataFrame") && !is.null(weight)) {
    stop(
      "weight arguments only applies to points with attributes",
      call. = FALSE
    )
  }

  if (!is.null(weight) && !(weight %in% names(input))) {
    stop("specified 'weight' column not present in input data", call. = FALSE)
  }

  call <- make_dissolve_call(
    field = field,
    sum_fields = sum_fields,
    copy_fields = copy_fields,
    weight = weight,
    snap = snap
  )

  ms_sp(input = input, call = call, ...)
}

dissolve_sf <- function(
  input,
  field,
  sum_fields,
  copy_fields,
  weight,
  snap,
  ...
) {
  if (
    !all(sf::st_is(input, c("POINT", "MULTIPOINT", "POLYGON", "MULTIPOLYGON")))
  ) {
    stop(
      "ms_dissolve only works with (MULTI)POINT or (MULTI)POLYGON",
      call. = FALSE
    )
  }

  if (!all(sf::st_is(input, c("POINT", "MULTIPOINT"))) && !is.null(weight)) {
    stop("weights arguments only applies to points", call. = FALSE)
  }

  call <- make_dissolve_call(
    field = field,
    sum_fields = sum_fields,
    copy_fields = copy_fields,
    weight = weight,
    snap = snap
  )

  ms_sf(input = input, call = call, ...)
}

suppressPackageStartupMessages({
  library("jsonify", quietly = TRUE)
  library("geojsonsf")
  library("sp")
  library("sf", quietly = TRUE)
  library("jsonlite")
})

withr::local_options(
  "rgdal_show_exportToProj4_warnings" = "none",
  "mapshaper.sys_quiet" = TRUE
)

## Objects for testing

basic_poly <- function(attr = FALSE) {
  structure(
    paste0(
      '{"type":"FeatureCollection",
  "features":[
  {"type":"Feature",',
      ifelse(attr, '"properties":{"a": 1, "b": 2},', '"properties":{},'),
      '"geometry":{"type":"Polygon","coordinates":[[
  [102,2],[102,3],[103,3],[103,2],[102,2]
  ]]}}
  ,{"type":"Feature",',
      ifelse(attr, '"properties":{"a": 5, "b": 3},', '"properties":{},'),
      '"geometry":{"type":"Polygon","coordinates":[[
  [100,0],[100,1],[101,1],[101,0],[100,0]
  ]]}}]}'
    ),
    class = c("geojson", "json")
  )
}

basic_points <- function() {
  structure(
    '{"type":"FeatureCollection",
    "features":[{
    "type":"Feature",
    "geometry":{"type":"Point","coordinates":[-78.41,-53.95]},
    "properties":{"x":-78,"y":-53,"foo":0}},{
    "type":"Feature",
    "geometry":{"type":"Point","coordinates":[-70.86,65.19]},
    "properties":{"x":-71,"y":65,"foo":1}},{
    "type":"Feature",
    "geometry":{"type":"Point","coordinates":[135.65,63.10]},
    "properties":{"x":135,"y":65,"foo":2}}]}',
    class = c("geojson", "json")
  )
}

basic_lines <- function() {
  structure(
    "{\"type\":\"FeatureCollection\",
\"features\":[{\"type\":\"Feature\",
\"properties\":{\"a\":1,\"b\":2,\"c\":3},
\"geometry\":{\"type\":\"LineString\",
\"coordinates\":[[102,2],[102,4],[104,4],[104,2],[102,2]]}}]}",
    class = c("geojson", "json")
  )
}

ce_poly <- structure(
  '{"type":"FeatureCollection","features":[{
"type": "Feature",
"properties": {},
"geometry": {
"type": "Polygon",
"coordinates": [
[
[53, -42],
[57, -42],
[57, -47],
[53, -47],
[53, -42]
]
]
}}]
}',
  class = c("geojson", "json")
)

ce_line <- structure(
  '{"type":"FeatureCollection","features":[
{ "type": "Feature",
"geometry": {
"type": "LineString",
"coordinates": [
[60, -37], [52, -42]
]
},
"properties": {}
}]
}',
  class = c("geojson", "json")
)

ce_points <- structure(
  '{"type":"FeatureCollection","features":[{"type":"Feature","geometry":{"type":"Point","coordinates":[53,-42]},"properties":{}},{"type":"Feature","geometry":{"type":"Point","coordinates":[57,-42]},"properties":{}}]}',
  class = c("geojson", "json")
)

ce_inner_poly <- structure(
  '{
"type": "Feature",
"properties": {},
"geometry": {
"type": "Polygon",
"coordinates": [
[
[51, -40],
[55, -40],
[55, -45],
[51, -45],
[51, -40]
]
]
}
}',
  class = c("geojson", "json")
)

ce_poly_spdf <- GeoJSON_to_sp(ce_poly)
ce_poly_sp <- as(ce_poly_spdf, "SpatialPolygons")

ce_line_spdf <- GeoJSON_to_sp(ce_line)
ce_line_sp <- as(ce_line_spdf, "SpatialLines")

ce_points_spdf <- GeoJSON_to_sp(ce_points)
ce_points_sp <- as(ce_points_spdf, "SpatialPoints")

ce_inner_poly_spdf <- GeoJSON_to_sp(ce_inner_poly)

ce_poly_sf <- sf::st_as_sf(ce_poly_spdf)
ce_poly_sfc <- sf::st_as_sfc(ce_poly_sp)
ce_lines_sf <- sf::st_as_sf(ce_line_spdf)
ce_points_sf <- sf::st_as_sf(ce_points_spdf)

ce_inner_poly_sf <- sf::read_sf(unclass(ce_inner_poly))

innerlines_poly <- structure(
  '{"type":"FeatureCollection",
  "features":[
{"type":"Feature",
"properties":{"foo": "a"},
"geometry":{"type":"Polygon","coordinates":[[
[102,2],[102,3],[103,3],[103,2],[102,2]
]]}}
,{"type":"Feature",
"properties":{"foo": "a"},
"geometry":{"type":"Polygon","coordinates":[[
[103,3],[104,3],[104,2],[103,2],[103,3]
]]}},
{"type":"Feature",
"properties":{"foo": "b"},
"geometry":{"type":"Polygon","coordinates":[[
[102,1],[102,2],[103,2],[103,1],[102,1]
]]}},
{"type":"Feature",
"properties":{"foo": "b"},
"geometry":{"type":"Polygon","coordinates":[[
[103,1],[103,2],[104,2],[104,1],[103,1]
]]}}]}',
  class = c("geojson", "json")
)

innerlines_poly_spdf <- GeoJSON_to_sp(innerlines_poly)
innerlines_poly_sp <- as(innerlines_poly_spdf, "SpatialPolygons")

innerlines_poly_sf <- read_sf(unclass(innerlines_poly))
innerlines_poly_sfc <- st_geometry(innerlines_poly_sf)

poly <- basic_poly()

poly_attr <- basic_poly(attr = TRUE)

points <- basic_points()
poly_spdf <- GeoJSON_to_sp(poly)
poly_sp <- as(poly_spdf, "SpatialPolygons")

points_spdf <- GeoJSON_to_sp(points)
points_sp <- as(points_spdf, "SpatialPoints")

test_that("ms_dissolve.geojson works", {
  out_poly <- ms_dissolve(poly)
  expect_s3_class(out_poly, "geojson")
  expect_equal(nrow(geojson_sf(out_poly)), 1)
  expect_snapshot_value(out_poly, style = "json2")

  out_points <- ms_dissolve(points)
  expect_s3_class(out_points, "geojson")
  expect_snapshot_value(out_points, style = "json2")

  skip_if_not(has_sys_mapshaper())
  expect_s3_class(ms_dissolve(points, sys = TRUE), "geojson")
})

test_that("ms_dissolve.geojson errors correctly", {
  expect_error(ms_dissolve('{foo: "bar"}'), "Input is not valid geojson")
})

test_that("ms_dissolve.character works", {
  out_poly <- ms_dissolve(unclass(poly))
  expect_s3_class(out_poly, "geojson")
  expect_equal(nrow(geojson_sf(out_poly)), 1)
  expect_snapshot_value(out_poly, style = "json2")
  out_points <- ms_dissolve(unclass(points))
  expect_s3_class(out_points, "geojson")
  expect_snapshot_value(out_points, style = "json2")
})

test_that("ms_dissolve.SpatialPolygons works", {
  out_poly <- ms_dissolve(poly_spdf)
  expect_s4_class(out_poly, "SpatialPolygonsDataFrame")
  expect_equal(length(out_poly@polygons), 1)

  out_points <- ms_dissolve(points_spdf)
  expect_s4_class(out_points, "SpatialPointsDataFrame")
  expect_equal(nrow(out_points@coords), 1)

  out_poly <- ms_dissolve(poly_sp)
  expect_s4_class(out_poly, "SpatialPolygons")
  expect_equal(length(out_poly@polygons), 1)

  out_points <- ms_dissolve(points_sp)
  expect_s4_class(out_points, "SpatialPoints")
  expect_equal(nrow(out_points@coords), 1)

  skip_if_not(has_sys_mapshaper())
  expect_s4_class(
    ms_dissolve(poly_spdf, sys = TRUE),
    "SpatialPolygonsDataFrame"
  )
})

test_that("copy_fields and sum_fields works", {
  expect_snapshot_value(
    ms_dissolve(poly_attr, copy_fields = c("a", "b")),
    style = "json2"
  )

  expect_snapshot_value(
    ms_dissolve(poly_attr, sum_fields = c("a", "b")),
    style = "json2"
  )
})

## sf classes
points_sf <- st_as_sf(points_spdf)
poly_sf <- st_as_sf(poly_spdf)

test_that("ms_dissolve.sf works with points", {
  expect_s3_class(ms_dissolve(points_sf), "sf")
  expect_s3_class(ms_dissolve(st_geometry(points_sf)), "sfc")
})

test_that("ms_dissolve.sf works with polygons", {
  expect_s3_class(ms_dissolve(poly_sf), "sf")
  expect_s3_class(ms_dissolve(st_geometry(poly_sf)), "sfc")
  skip_if_not(has_sys_mapshaper())
  expect_s3_class(ms_dissolve(poly_sf, sys = TRUE), "sf")
})

test_that("weight argument works", {
  # Don't test this as the V8 error throws a warning
  expect_warning(
    ms_dissolve(points, weight = "w"),
    "The command returned an empty response."
  )
  expect_error(
    ms_dissolve(points_sf, weight = "w"),
    "specified 'weight' column not present in input data"
  )
  expect_gt(
    sum(sf::st_coordinates(ms_dissolve(points_sf, weight = "foo"))),
    sum(sf::st_coordinates(ms_dissolve(points_sf)))
  )
  expect_gt(
    sum(sp::coordinates(ms_dissolve(points_spdf, weight = "foo"))),
    sum(sp::coordinates(ms_dissolve(points_spdf)))
  )
})

pts <- structure(
  "{\"type\":\"FeatureCollection\",
\"features\":[{\"type\":\"Feature\",
\"properties\":{\"a\":1,\"b\":2,\"c\":3},
\"geometry\":{\"type\":\"Point\",
\"coordinates\":[103,3]}}]}",
  class = c("geojson", "json")
)

lines <- structure(
  "{\"type\":\"FeatureCollection\",
\"features\":[{\"type\":\"Feature\",
\"properties\":{\"a\":1,\"b\":2,\"c\":3},
\"geometry\":{\"type\":\"LineString\",
\"coordinates\":[[102,2],[102,4],[104,4],[104,2],[102,2]]}}]}",
  class = c("geojson", "json")
)

test_that("ms_filter_fields works with polygons", {
  poly <- basic_poly(attr = TRUE)

  expect_s3_class(ms_filter_fields(poly, "a"), "geojson")
  expect_s3_class(ms_filter_fields(unclass(poly), c("a", "b")), "geojson")
  expect_snapshot_value(ms_filter_fields(poly, "a"), style = "json2")
  out_sp <- ms_filter_fields(GeoJSON_to_sp(poly), c("a", "b"))
  expect_s4_class(out_sp, "SpatialPolygonsDataFrame")
  expect_equal(
    out_sp@data,
    data.frame(a = c(1, 5), b = c(2, 3), row.names = 1:2)
  )

  skip_if_not(has_sys_mapshaper())
  expect_s3_class(ms_filter_fields(poly, "a", sys = TRUE), "geojson")
  expect_s3_class(
    ms_filter_fields(GeoJSON_to_sf(poly), c("a", "b"), sys = TRUE),
    "sf"
  )
})

test_that("ms_filter_fields works with points", {
  pts <- basic_points()

  expect_s3_class(ms_filter_fields(pts, c("x", "y")), "geojson")
  expect_snapshot_value(ms_filter_fields(pts, c("x", "y")), style = "json2")
  out_sp <- ms_filter_fields(GeoJSON_to_sp(pts), "x")
  expect_s4_class(out_sp, "SpatialPointsDataFrame")
  expect_equal(out_sp@data, data.frame(x = c(-78, -71, 135)))
})

test_that("ms_filter_fields works with lines", {
  lines <- basic_lines()

  expect_s3_class(ms_filter_fields(lines, c("a", "b")), "geojson")
  expect_snapshot_value(ms_filter_fields(lines, c("a", "b")), style = "json2")
  out_sp <- ms_filter_fields(GeoJSON_to_sp(lines), c("a", "b"))
  expect_s4_class(out_sp, "SpatialLinesDataFrame")
  expect_equal(out_sp@data, data.frame(a = 1, b = 2, row.names = 1L))
})

test_that("ms_filter_fields fails correctly", {
  expect_error(
    ms_filter_fields('{foo: "bar"}', "a"),
    "Input is not valid geojson"
  )
  poly <- basic_poly(attr = TRUE)
  expect_warning(
    ms_filter_fields(poly, "d"),
    "The command returned an empty response"
  )
  expect_error(
    ms_filter_fields(GeoJSON_to_sp(poly), "d"),
    "Not all named fields exist in input data"
  )
  expect_error(ms_filter_fields(poly, 1), "fields must be a character vector")
})

test_that("ms_filter_fields works with sf", {
  lines <- basic_lines()

  lines_sf <- read_sf(unclass(lines))
  out_sf <- ms_filter_fields(lines_sf, c("a", "b"))
  expect_s3_class(out_sf, "sf")
  expect_equal(names(out_sf), c("a", "b", "geometry"))
  expect_error(ms_filter_fields(lines_sf, "d"), "Not all fields are in input")

  skip_if_not(has_sys_mapshaper())
  expect_s3_class(ms_filter_fields(lines_sf, c("a", "b"), sys = TRUE), "sf")
})

#' Convert multipart lines or polygons to singlepart
#'
#' For objects of class \code{Spatial} (e.g., \code{SpatialPolygonsDataFrame}),
#' you may find it faster to use \code{sp::disaggregate}.
#'
#' There is currently no method for SpatialMultiPoints
#'
#' @param input One of:
#' \itemize{
#'  \item \code{geo_json} or \code{character} multipart lines, or polygons;
#'  \item multipart \code{SpatialPolygons}, \code{SpatialLines};
#'  \item \code{sf} or \code{sfc} multipart lines, or polygons object
#' }
#' @inheritDotParams apply_mapshaper_commands force_FC sys sys_mem quiet gj2008
#'
#' @return same class as input
#'
#' @examples
#' library(geojsonsf)
#' library(sf)
#'
#' poly <- "{\"type\":\"FeatureCollection\",\"features\":
#'           [\n{\"type\":\"Feature\",\"geometry\":{\"type\":
#'           \"MultiPolygon\",\"coordinates\":[[[[102,2],[102,3],
#'           [103,3],[103,2],[102,2]]],[[[100,0],[100,1],[101,1],
#'           [101,0],[100,0]]]]},\"properties\":{\"a\":0}}\n]}"
#'
#' poly <- geojson_sf(poly)
#' plot(poly)
#' length(poly)
#' poly
#'
#' # Explode the polygon
#' out <- ms_explode(poly)
#' plot(out)
#' length(out)
#' out
#'
#' @export
ms_explode <- function(input, ...) {
  UseMethod("ms_explode")
}

#' @export
ms_explode.character <- function(input, ...) {
  input <- check_character_input(input)

  apply_mapshaper_commands(data = input, command = "-explode", ...)
}

#' @export
ms_explode.json <- function(input, ...) {
  apply_mapshaper_commands(data = input, command = "-explode", ...)
}

## The method using mapshaper's explode works, but is waaaay slower than
## sp::disaggregate due to converstion to/from geojson

#' @export
ms_explode.SpatialPolygons <- function(input, ...) {
  explode_sp(input, ...)
}

#' @export
ms_explode.SpatialLines <- function(input, ...) {
  explode_sp(input, ...)
}

# #' @describeIn ms_explode Method for SpatialPoints
# #' @export
# ms_explode.SpatialPoints <- function(input, force_FC = TRUE) {
#   explode_sp(input, force_FC)
# }

explode_sp <- function(input, ...) {
  ms_sp(input = input, call = "-explode", ...)
}

#' @export
ms_explode.sf <- function(input, ...) {
  explode_sf(input = input, ...)
}

#' @export
ms_explode.sfc <- function(input, ...) {
  explode_sf(input = input, ...)
}

explode_sf <- function(input, ...) {
  ms_sf(input = input, call = "-explode", ...)
}

#' Topologically-aware geometry simplification.
#'
#' Uses \href{https://github.com/mbloch/mapshaper}{mapshaper} to simplify
#' polygons.
#'
#' @param input spatial object to simplify. One of:
#' \itemize{
#'  \item \code{geo_json} or \code{character} polygons or lines;
#'  \item \code{SpatialPolygons*} or \code{SpatialLines*};
#'  \item \code{sf} or \code{sfc} polygons or lines object
#'  }
#' @param keep proportion of points to retain (0-1; default 0.05)
#' @param method simplification method to use: \code{"vis"} for Visvalingam
#'   algorithm, or \code{"dp"} for Douglas-Peuker algorithm. If left as
#'   \code{NULL} (default), uses Visvalingam simplification but modifies the
#'   area metric by underweighting the effective area of points at the vertex of
#'   more acute angles, resulting in a smoother appearance. See this
#'   [link](https://github.com/mbloch/mapshaper/wiki/Simplification-Tips)
#'   for more information.
#' @param weighting Coefficient for weighting Visvalingam simplification
#' (default is 0.7). Higher values produce smoother output. weighting=0 is
#' equivalent to unweighted Visvalingam simplification.
#' @param keep_shapes Prevent small polygon features from disappearing at high
#'   simplification (default \code{FALSE})
#' @param no_repair disable intersection repair after simplification (default
#'   \code{FALSE}).
#' @param snap Snap together vertices within a small distance threshold to fix
#'   small coordinate misalignment in adjacent polygons. Default \code{TRUE}.
#' @param explode Should multipart polygons be converted to singlepart polygons?
#'   This prevents small shapes from disappearing during simplification if
#'   \code{keep_shapes = TRUE}. Default \code{FALSE}
#' @param drop_null_geometries should Features with null geometries be dropped?
#'   Ignored for \code{Spatial*} objects, as it is always \code{TRUE}.
#' @param snap_interval Specify snapping distance in source units, must be a
#'   numeric. Default \code{NULL}
#' @inheritDotParams apply_mapshaper_commands force_FC sys sys_mem quiet gj2008
#'
#' @return a simplified representation of the geometry in the same class as the
#'   input
#' @examples
#' # With a simple geojson object
#' poly <- structure('{
#'  "type": "Feature",
#'  "properties": {},
#'  "geometry": {
#'    "type": "Polygon",
#'    "coordinates": [[
#'      [-70.603637, -33.399918],
#'      [-70.614624, -33.395332],
#'      [-70.639343, -33.392466],
#'      [-70.659942, -33.394759],
#'      [-70.683975, -33.404504],
#'      [-70.697021, -33.419406],
#'      [-70.701141, -33.434306],
#'      [-70.700454, -33.446339],
#'      [-70.694274, -33.458369],
#'      [-70.682601, -33.465816],
#'      [-70.668869, -33.472117],
#'      [-70.646209, -33.473835],
#'      [-70.624923, -33.472117],
#'      [-70.609817, -33.468107],
#'      [-70.595397, -33.458369],
#'      [-70.587158, -33.442901],
#'      [-70.587158, -33.426283],
#'      [-70.590591, -33.414248],
#'      [-70.594711, -33.406224],
#'      [-70.603637, -33.399918]
#'    ]]
#'  }
#' }', class = c("geojson", "json"))
#'
#' ms_simplify(poly, keep = 0.1)
#'
#' # With an sf object
#'
#' poly_sf <- geojsonsf::geojson_sf(poly)
#' ms_simplify(poly_sf, keep = 0.5)
#'
#' @export
ms_simplify <- function(
  input,
  keep = 0.05,
  method = NULL,
  weighting = 0.7,
  keep_shapes = FALSE,
  no_repair = FALSE,
  snap = TRUE,
  explode = FALSE,
  drop_null_geometries = TRUE,
  snap_interval = NULL,
  ...
) {
  UseMethod("ms_simplify")
}

#' @export
ms_simplify.character <- function(
  input,
  keep = 0.05,
  method = NULL,
  weighting = 0.7,
  keep_shapes = FALSE,
  no_repair = FALSE,
  snap = TRUE,
  explode = FALSE,
  drop_null_geometries = TRUE,
  snap_interval = NULL,
  ...
) {
  input <- check_character_input(input)

  ms_simplify_json(
    input = input,
    keep = keep,
    method = method,
    weighting = weighting,
    keep_shapes = keep_shapes,
    no_repair = no_repair,
    snap = snap,
    explode = explode,
    drop_null_geometries = drop_null_geometries,
    snap_interval = snap_interval,
    ...
  )
}

#' @export
ms_simplify.json <- function(
  input,
  keep = 0.05,
  method = NULL,
  weighting = 0.7,
  keep_shapes = FALSE,
  no_repair = FALSE,
  snap = TRUE,
  explode = FALSE,
  drop_null_geometries = TRUE,
  snap_interval = NULL,
  ...
) {
  ms_simplify_json(
    input = input,
    keep = keep,
    method = method,
    weighting = weighting,
    keep_shapes = keep_shapes,
    no_repair = no_repair,
    snap = snap,
    explode = explode,
    drop_null_geometries = drop_null_geometries,
    snap_interval = snap_interval,
    ...
  )
}

#' @export
ms_simplify.SpatialPolygons <- function(
  input,
  keep = 0.05,
  method = NULL,
  weighting = 0.7,
  keep_shapes = FALSE,
  no_repair = FALSE,
  snap = TRUE,
  explode = FALSE,
  drop_null_geometries = TRUE,
  snap_interval = NULL,
  ...
) {
  if (!is(input, "Spatial")) {
    stop("input must be a spatial object")
  }

  call <- make_simplify_call(
    keep = keep,
    method = method,
    weighting = weighting,
    keep_shapes = keep_shapes,
    no_repair = no_repair,
    snap = snap,
    explode = explode,
    drop_null_geometries = !keep_shapes,
    snap_interval = snap_interval
  )

  ms_sp(input, call, ...)
}

#' @export
ms_simplify.SpatialLines <- ms_simplify.SpatialPolygons

#' @export
ms_simplify.sf <- function(
  input,
  keep = 0.05,
  method = NULL,
  weighting = 0.7,
  keep_shapes = FALSE,
  no_repair = FALSE,
  snap = TRUE,
  explode = FALSE,
  drop_null_geometries = TRUE,
  snap_interval = NULL,
  ...
) {
  if (
    !all(
      sf::st_geometry_type(input) %in%
        c("LINESTRING", "MULTILINESTRING", "POLYGON", "MULTIPOLYGON")
    )
  ) {
    stop(
      "ms_simplify can only operate on (multi)polygons and (multi)linestrings",
      call. = FALSE
    )
  }

  call <- make_simplify_call(
    keep = keep,
    method = method,
    weighting = weighting,
    keep_shapes = keep_shapes,
    no_repair = no_repair,
    snap = snap,
    explode = explode,
    drop_null_geometries = !keep_shapes,
    snap_interval = snap_interval
  )

  ms_sf(input, call, ...)
}

#' @export
ms_simplify.sfc <- ms_simplify.sf

ms_simplify_json <- function(
  input,
  keep,
  method,
  weighting,
  keep_shapes,
  no_repair,
  snap,
  explode,
  drop_null_geometries,
  snap_interval,
  ...
) {
  call <- make_simplify_call(
    keep = keep,
    method = method,
    weighting = weighting,
    keep_shapes = keep_shapes,
    no_repair = no_repair,
    snap = snap,
    explode = explode,
    drop_null_geometries = drop_null_geometries,
    snap_interval = snap_interval
  )

  ret <- apply_mapshaper_commands(data = input, command = call, ...)

  ret
}

make_simplify_call <- function(
  keep,
  method,
  weighting,
  keep_shapes,
  no_repair,
  snap,
  explode,
  drop_null_geometries,
  snap_interval
) {
  if (keep > 1 || keep <= 0) {
    stop("keep must be > 0 and <= 1")
  }
  if (!is.null(snap_interval)) {
    if (!is.numeric(snap_interval)) {
      stop("snap_interval must be a numeric")
    }
    if (snap_interval < 0) stop("snap_interval must be >= 0")
  }
  if (is.null(method)) {
    method <- ""
  } else if (method == "vis") {
    method <- "visvalingam"
  } else if (!method == "dp") {
    stop(
      "method should be one of 'vis', 'dp', or NULL (to use the default weighted Visvalingam method)"
    )
  }

  if (!is.numeric(weighting)) {
    stop("weighting needs to be numeric.")
  }

  if (explode) {
    explode <- "-explode"
  } else {
    explode <- NULL
  }
  if (snap && !is.null(snap_interval)) {
    snap_interval <- paste0("interval=", snap_interval)
  }
  if (snap) {
    snap <- "-snap"
  } else {
    snap <- NULL
  }
  if (keep_shapes) {
    keep_shapes <- "keep-shapes"
  } else {
    keep_shapes <- NULL
  }
  if (no_repair) {
    no_repair <- "no-repair"
  } else {
    no_repair <- NULL
  }
  if (drop_null_geometries) {
    drop_null <- "-filter remove-empty"
  } else {
    drop_null <- NULL
  }

  call <- list(
    explode,
    snap,
    snap_interval,
    "-simplify",
    keep = format(keep, scientific = FALSE),
    method,
    weighting = paste0("weighting=", format(weighting, scientific = FALSE)),
    keep_shapes,
    no_repair,
    drop_null
  )

  call
}

ms_de_unit <- function(input) {
  input_columns_units <- vapply(
    input,
    inherits,
    "units",
    FUN.VALUE = logical(1)
  )
  if (any(input_columns_units)) {
    units_column_names <- names(input_columns_units)[input_columns_units]
    msg <- paste0(
      "Coercing these 'units' columns to class numeric: ",
      paste(units_column_names, collapse = ", ")
    )
    warning(msg)

    for (i in units_column_names) {
      input[[i]] <- as.numeric(input[[i]])
    }
  }
  input
}

#' Remove features or portions of features that fall outside a clipping area.
#'
#' Removes portions of the target layer that fall outside the clipping layer or bounding box.
#'
#' @param target the target layer from which to remove portions. One of:
#' \itemize{
#'  \item \code{geo_json} or \code{character} points, lines, or polygons;
#'  \item \code{SpatialPolygons}, \code{SpatialLines}, \code{SpatialPoints};
#'  \item \code{sf} or \code{sfc} points, lines, or polygons object
#'  }
#' @param clip the clipping layer (polygon). One of:
#' \itemize{
#'  \item \code{geo_json} or \code{character} polygons;
#'  \item \code{SpatialPolygons*};
#'  \item \code{sf} or \code{sfc} polygons object
#' }
#' @param bbox supply a bounding box instead of a clipping layer to extract from
#'   the target layer. Supply as a numeric vector: \code{c(minX, minY, maxX, maxY)}.
#' @param remove_slivers Remove tiny sliver polygons created by clipping. (Default \code{FALSE})
#' @inheritDotParams apply_mapshaper_commands force_FC sys sys_mem quiet gj2008
#'
#' @return clipped target in the same class as the input target
#'
#' @examples
#'
#' if (rmapshaper:::v8_version() >= "6") {
#'   library(geojsonsf, quietly = TRUE)
#'   library(sf)
#'
#'   poly <- structure("{\"type\":\"FeatureCollection\",
#'     \"features\":[{\"type\":\"Feature\",\"properties\":{},
#'     \"geometry\":{\"type\":\"Polygon\",\"coordinates\":
#'     [[[52.8658,-44.7219],[53.7702,-40.4873],[55.3204,-37.5579],
#'     [56.2757,-37.917],[56.184,-40.6443],[61.0835,-40.7529],
#'     [58.0202,-43.634],[61.6699,-45.0678],[62.737,-46.2841],
#'     [55.7763,-46.2637],[54.9742,-49.1184],[52.799,-45.9386],
#'     [52.0329,-49.5677],[50.1747,-52.1814],[49.0098,-52.3641],
#'     [52.7068,-45.7639],[43.2278,-47.1908],[48.4755,-45.1388],
#'     [50.327,-43.5207],[48.0804,-41.2784],[49.6307,-40.6159],
#'     [52.8658,-44.7219]]]}}]}", class = c("geojson", "json"))
#'   poly <- geojson_sf(poly)
#'   plot(poly)
#'
#'   clip_poly <- structure('{
#'   "type": "Feature",
#'   "properties": {},
#'   "geometry": {
#'   "type": "Polygon",
#'   "coordinates": [
#'   [
#'   [51, -40],
#'   [55, -40],
#'   [55, -45],
#'   [51, -45],
#'   [51, -40]
#'   ]
#'   ]
#'   }
#'   }', class = c("geojson", "json"))
#'   clip_poly <- geojson_sf(clip_poly)
#'   plot(clip_poly)
#'
#'   out <- ms_clip(poly, clip_poly)
#'   plot(out, add = TRUE)
#' }
#'
#' @export
ms_clip <- function(
  target,
  clip = NULL,
  bbox = NULL,
  remove_slivers = FALSE,
  ...
) {
  stop_for_old_v8()
  UseMethod("ms_clip")
}

#' @export
ms_clip.character <- function(
  target,
  clip = NULL,
  bbox = NULL,
  remove_slivers = FALSE,
  ...
) {
  target <- check_character_input(target)

  clip_erase_json(
    target = target,
    overlay_layer = clip,
    type = "clip",
    remove_slivers = remove_slivers,
    bbox = bbox,
    ...
  )
}

#' @export
ms_clip.json <- function(
  target,
  clip = NULL,
  bbox = NULL,
  remove_slivers = FALSE,
  ...
) {
  clip_erase_json(
    target = target,
    overlay_layer = clip,
    type = "clip",
    remove_slivers = remove_slivers,
    bbox = bbox,
    ...
  )
}

#' @export
ms_clip.SpatialPolygons <- function(
  target,
  clip = NULL,
  bbox = NULL,
  remove_slivers = FALSE,
  ...
) {
  clip_erase_sp(
    target = target,
    overlay_layer = clip,
    type = "clip",
    remove_slivers = remove_slivers,
    bbox = bbox,
    ...
  )
}

#' @export
ms_clip.SpatialLines <- function(
  target,
  clip = NULL,
  bbox = NULL,
  remove_slivers = FALSE,
  ...
) {
  clip_erase_sp(
    target = target,
    overlay_layer = clip,
    type = "clip",
    remove_slivers = remove_slivers,
    bbox = bbox,
    ...
  )
}

#' @export
ms_clip.SpatialPoints <- function(
  target,
  clip = NULL,
  bbox = NULL,
  remove_slivers = FALSE,
  ...
) {
  clip_erase_sp(
    target = target,
    overlay_layer = clip,
    type = "clip",
    remove_slivers = remove_slivers,
    bbox = bbox,
    ...
  )
}

#' @export
ms_clip.sf <- function(
  target,
  clip = NULL,
  bbox = NULL,
  remove_slivers = FALSE,
  ...
) {
  clip_erase_sf(
    target = target,
    overlay_layer = clip,
    type = "clip",
    remove_slivers = remove_slivers,
    bbox = bbox,
    ...
  )
}

#' @export
ms_clip.sfc <- function(
  target,
  clip = NULL,
  bbox = NULL,
  remove_slivers = FALSE,
  ...
) {
  clip_erase_sf(
    target = target,
    overlay_layer = clip,
    type = "clip",
    remove_slivers = remove_slivers,
    bbox = bbox,
    ...
  )
}

#' Remove features or portions of features that fall inside a specified area
#'
#' Removes portions of the target layer that fall inside the erasing layer or bounding box.
#'
#' @param target the target layer from which to remove portions. One of:
#' \itemize{
#'  \item \code{geo_json} or \code{character} points, lines, or polygons;
#'  \item \code{SpatialPolygons}, \code{SpatialLines}, \code{SpatialPoints}
#'  }
#' @param erase the erase layer (polygon). One of:
#' \itemize{
#'  \item \code{geo_json} or \code{character} polygons;
#'  \item \code{SpatialPolygons*}
#' }
#' @param bbox supply a bounding box instead of an erasing layer to remove from
#'   the target layer. Supply as a numeric vector: \code{c(minX, minY, maxX, maxY)}.
#' @param remove_slivers Remove tiny sliver polygons created by erasing. (Default \code{FALSE})
#' @inheritDotParams apply_mapshaper_commands force_FC sys sys_mem quiet gj2008
#'
#' @return erased target in the same format as the input target
#' @examples
#' if (rmapshaper:::v8_version() >= "6") {
#'   library(geojsonsf, quietly = TRUE)
#'   library(sf)
#'
#'   points <- structure("{\"type\":\"FeatureCollection\",
#'     \"features\":[{\"type\":\"Feature\",\"properties\":{},
#'     \"geometry\":{\"type\":\"Point\",\"coordinates\":
#'     [52.8658,-44.7219]}},{\"type\":\"Feature\",\"properties\":{},
#'     \"geometry\":{\"type\":\"Point\",\"coordinates\":
#'     [53.7702,-40.4873]}},{\"type\":\"Feature\",\"properties\":{},
#'     \"geometry\":{\"type\":\"Point\",\"coordinates\":[55.3204,-37.5579]}},
#'     {\"type\":\"Feature\",\"properties\":{},\"geometry\":
#'     {\"type\":\"Point\",\"coordinates\":[56.2757,-37.917]}},
#'     {\"type\":\"Feature\",\"properties\":{},\"geometry\":
#'     {\"type\":\"Point\",\"coordinates\":[56.184,-40.6443]}},
#'     {\"type\":\"Feature\",\"properties\":{},\"geometry\":
#'     {\"type\":\"Point\",\"coordinates\":[61.0835,-40.7529]}},
#'     {\"type\":\"Feature\",\"properties\":{},\"geometry\":
#'     {\"type\":\"Point\",\"coordinates\":[58.0202,-43.634]}}]}",
#'     class = c("geojson", "json"))
#'   points <- geojson_sf(points)
#'   plot(points)
#'
#'   erase_poly <- structure('{
#'   "type": "Feature",
#'   "properties": {},
#'   "geometry": {
#'   "type": "Polygon",
#'   "coordinates": [
#'   [
#'   [51, -40],
#'   [55, -40],
#'   [55, -45],
#'   [51, -45],
#'   [51, -40]
#'   ]
#'   ]
#'   }
#'   }', class = c("geojson", "json"))
#'   erase_poly <- geojson_sf(erase_poly)
#'
#'   out <- ms_erase(points, erase_poly)
#'   plot(out, add = TRUE)
#' }
#'
#'@export
ms_erase <- function(
  target,
  erase = NULL,
  bbox = NULL,
  remove_slivers = FALSE,
  ...
) {
  stop_for_old_v8()
  UseMethod("ms_erase")
}

#' @export
ms_erase.character <- function(
  target,
  erase = NULL,
  bbox = NULL,
  remove_slivers = FALSE,
  ...
) {
  target <- check_character_input(target)

  clip_erase_json(
    target = target,
    overlay_layer = erase,
    type = "erase",
    remove_slivers = remove_slivers,
    bbox = bbox,
    ...
  )
}

#' @export
ms_erase.json <- function(
  target,
  erase = NULL,
  bbox = NULL,
  remove_slivers = FALSE,
  ...
) {
  clip_erase_json(
    target = target,
    overlay_layer = erase,
    type = "erase",
    remove_slivers = remove_slivers,
    bbox = bbox,
    ...
  )
}

#' @export
ms_erase.SpatialPolygons <- function(
  target,
  erase = NULL,
  bbox = NULL,
  remove_slivers = FALSE,
  ...
) {
  clip_erase_sp(
    target = target,
    overlay_layer = erase,
    type = "erase",
    remove_slivers = remove_slivers,
    bbox = bbox,
    ...
  )
}

#' @export
ms_erase.SpatialLines <- function(
  target,
  erase = NULL,
  bbox = NULL,
  remove_slivers = FALSE,
  ...
) {
  clip_erase_sp(
    target = target,
    overlay_layer = erase,
    type = "erase",
    remove_slivers = remove_slivers,
    bbox = bbox,
    ...
  )
}

#' @export
ms_erase.SpatialPoints <- function(
  target,
  erase = NULL,
  bbox = NULL,
  remove_slivers = FALSE,
  ...
) {
  clip_erase_sp(
    target = target,
    overlay_layer = erase,
    type = "erase",
    remove_slivers = remove_slivers,
    bbox = bbox,
    ...
  )
}

#' @export
ms_erase.sf <- function(
  target,
  erase = NULL,
  bbox = NULL,
  remove_slivers = FALSE,
  ...
) {
  clip_erase_sf(
    target = target,
    overlay_layer = erase,
    type = "erase",
    remove_slivers = remove_slivers,
    bbox = bbox,
    ...
  )
}

#' @export
ms_erase.sfc <- function(
  target,
  erase = NULL,
  bbox = NULL,
  remove_slivers = FALSE,
  ...
) {
  clip_erase_sf(
    target = target,
    overlay_layer = erase,
    type = "erase",
    remove_slivers = remove_slivers,
    bbox = bbox,
    ...
  )
}

clip_erase_json <- function(
  target,
  overlay_layer,
  bbox,
  remove_slivers,
  type,
  force_FC = TRUE,
  sys = FALSE,
  sys_mem = getOption("mapshaper.sys_mem", default = 8),
  quiet = getOption("mapshaper.sys_quiet", default = FALSE),
  gj2008 = FALSE
) {
  check_overlay_bbox(overlay_layer = overlay_layer, bbox = bbox, type = type)

  if (!is.null(overlay_layer)) {
    overlay_layer <- check_character_input(overlay_layer)
  }

  mapshaper_clip_erase(
    target_layer = target,
    overlay_layer = overlay_layer,
    type = type,
    remove_slivers = remove_slivers,
    bbox = bbox,
    force_FC = force_FC,
    sys = sys,
    sys_mem = sys_mem,
    quiet = quiet,
    gj2008 = gj2008
  )
}

clip_erase_sp <- function(
  target,
  overlay_layer,
  bbox,
  type,
  remove_slivers,
  force_FC = TRUE,
  sys = FALSE,
  sys_mem = getOption("mapshaper.sys_mem", default = 8),
  quiet = getOption("mapshaper.sys_quiet", default = FALSE),
  gj2008 = FALSE
) {
  check_overlay_bbox(overlay_layer = overlay_layer, bbox = bbox, type = type)

  target_proj <- methods::slot(target, "proj4string")

  if (is.null(bbox)) {
    if (!is(overlay_layer, "SpatialPolygons"))
      stop(
        type,
        " must be of class SpatialPolygons or SpatialPolygonsDataFrame"
      )
    if (!sp::identicalCRS(target, overlay_layer)) {
      stop("target and ", type, " do not have identical CRS.", call. = FALSE)
    }
    overlay_geojson <- sp_to_GeoJSON(overlay_layer, file = sys)
  }

  target_geojson <- sp_to_GeoJSON(target, file = sys)

  result <- mapshaper_clip_erase(
    target_layer = target_geojson,
    overlay_layer = overlay_geojson,
    type = type,
    remove_slivers = remove_slivers,
    bbox = bbox,
    force_FC = TRUE,
    sys = sys,
    sys_mem = sys_mem,
    quiet = quiet,
    gj2008 = gj2008
  )

  ret <- GeoJSON_to_sp(result, target_proj)

  # remove data slot if input didn't have one
  if (!.hasSlot(target, "data")) {
    ret <- as(ret, class(target)[1])
  }

  ret
}

clip_erase_sf <- function(
  target,
  overlay_layer,
  bbox,
  type,
  remove_slivers,
  force_FC = TRUE,
  sys = FALSE,
  sys_mem = getOption("mapshaper.sys_mem", default = 8),
  quiet = getOption("mapshaper.sys_quiet", default = FALSE),
  gj2008 = FALSE
) {
  check_overlay_bbox(overlay_layer = overlay_layer, bbox = bbox, type = type)

  has_data <- is(target, "sf")
  if (has_data) {
    classes <- col_classes(target)
  }

  target_proj <- sf::st_crs(target)

  if (is.null(bbox)) {
    if (
      !(inherits(overlay_layer, c("sf", "sfc"))) ||
        !all(sf::st_is(overlay_layer, c("POLYGON", "MULTIPOLYGON")))
    ) {
      stop(
        type,
        " must be an sf or sfc object with POLYGON or MULTIPLOYGON geometry"
      )
    }
    if (target_proj != sf::st_crs(overlay_layer)) {
      stop("target and ", type, " do not have identical CRS.", call. = FALSE)
    }
    overlay_geojson <- sf_to_GeoJSON(overlay_layer, file = sys)
  }

  target_geojson <- sf_to_GeoJSON(target, file = sys)

  result <- mapshaper_clip_erase(
    target_layer = target_geojson,
    overlay_layer = overlay_geojson,
    bbox = bbox,
    type = type,
    remove_slivers = remove_slivers,
    force_FC = TRUE,
    sys = sys,
    sys_mem = sys_mem,
    quiet = quiet,
    gj2008 = gj2008
  )

  ret <- GeoJSON_to_sf(result, target_proj)

  # remove data slot if input didn't have one
  if (!has_data) {
    ret <- sf::st_geometry(ret)
  } else {
    ret <- restore_classes(ret, classes)
  }

  ret
}

check_overlay_bbox <- function(overlay_layer, bbox, type) {
  if (is.null(overlay_layer)) {
    if (is.null(bbox)) {
      stop("You must specificy either a bounding box or a ", type, " polygon.")
    }
    if (length(bbox) != 4 || !is.numeric(bbox)) {
      stop("bbox must be a numeric vector of length 4")
    }
  }

  if (!is.null(overlay_layer) && !is.null(bbox)) {
    stop("Please only specify either a bounding box or a ", type, " polygon.")
  }

  invisible(NULL)
}

mapshaper_clip_erase <- function(
  target_layer,
  overlay_layer,
  bbox,
  type,
  remove_slivers,
  force_FC = TRUE,
  sys = FALSE,
  sys_mem = getOption("mapshaper.sys_mem", default = 8),
  quiet = getOption("mapshaper.sys_quiet", default = FALSE),
  gj2008 = FALSE
) {
  remove_slivers <- ifelse(remove_slivers, "remove-slivers", "")

  if (!is.null(bbox)) {
    cmd <- paste0(
      "-",
      type,
      " bbox=",
      paste0(bbox, collapse = ","),
      " ",
      remove_slivers
    )
    out <- apply_mapshaper_commands(
      target_layer,
      cmd,
      force_FC = force_FC,
      sys = sys,
      sys_mem = sys_mem,
      quiet = quiet,
      gj2008 = gj2008
    )
  } else if (!is.null(overlay_layer)) {
    if (sys) {
      on.exit(unlink(c(target_layer, overlay_layer)), add = TRUE)
      cmd <- paste0("-", type)
      out <- sys_mapshaper(
        data = target_layer,
        data2 = overlay_layer,
        command = cmd,
        force_FC = force_FC,
        sys_mem = sys_mem,
        quiet = quiet,
        gj2008 = gj2008
      )
    } else {
      ms <- ms_make_ctx()
      ## Import the layers into the V8 session
      ms$assign("target_geojson", target_layer)
      ms$assign("overlay_geojson", overlay_layer)

      # Combine them into a single object
      ms$eval(
        "
      var data_layers = {'target.json': target_geojson, 'overlay.json': overlay_geojson}
    "
      )

      ## Construct the command string; referring to layer names as assigned above
      command <- paste0(
        "-i overlay.json -i target.json -",
        type,
        " overlay ",
        remove_slivers
      )

      command <- paste(command, "-o format=geojson")

      if (force_FC) {
        # this must come after -o format=geojson
        command <- paste(command, fc_command())
      }

      if (isTRUE(gj2008)) {
        command <- paste(command, "gj2008")
      }

      # Create an object to hold the return value
      ms$eval("var return_data;")

      # Run the commands on the data
      ms$eval(paste0(
        'mapshaper.applyCommands("',
        command,
        '", data_layers, ',
        callback(),
        ')'
      ))

      out <- ms$get("return_data")[["target.json"]]
      out <- class_geo_json(out)
    }
  }
  out
}

# Investigation: Why `-proj EPSG:3005 init=EPSG:4326` Returns Empty Response

## Problem

The command `apply_mapshaper_commands(geo, "-proj EPSG:3005 init=EPSG:4326")` returns an empty response with a warning.

## Root Cause

The bundled `mproj` library (v0.0.40) **does not support EPSG code lookups**. It only accepts full PROJ.4 definition strings.

### Technical Details

1. **mproj Library**: The package uses `mproj@0.0.40` bundled in the browserified JavaScript
2. **No EPSG Database**: Unlike modern PROJ (v6+), `mproj` is a lightweight port of PROJ.4 v4.9.3 that does **not** include EPSG code definitions
3. **Error Handling**: When an invalid projection string is passed:
   - mapshaper/mproj fails to parse it
   - The V8 callback logs the error to `console.error` (not visible in R)
   - Returns `null`/`undefined` data
   - R's `class_geo_json()` function detects this and returns an empty response with the warning: "The command returned an empty response. Please check your inputs"

## Test Results

Tested 8 different formats:

| Format | Works? | Example |
|--------|--------|---------|
| `EPSG:3005 init=EPSG:4326` | ✗ | EPSG codes not recognized |
| `epsg:3005 init=epsg:4326` | ✗ | Lowercase doesn't help |
| `+init=epsg:3005 from=+init=epsg:4326` | ✗ | Traditional PROJ.4 init syntax not supported |
| `+init=epsg:3005` | ✗ | Still tries EPSG lookup |
| `EPSG:3005` | ✗ | Same issue |
| **Full PROJ.4 string** | **✓** | **This is the only working method** |
| `crs=epsg:3005` | ✗ | Alternative syntax doesn't work |
| `crs=EPSG:3005` | ✗ | Case doesn't matter |

## Solution

**Use full PROJ.4 definition strings instead of EPSG codes:**

```r
# WGS84 (EPSG:4326) to BC Albers (EPSG:3005)
apply_mapshaper_commands(
  geo,
  "-proj '+proj=aea +lat_0=45 +lon_0=-126 +lat_1=50 +lat_2=58.5 +x_0=1000000 +y_0=0 +datum=NAD83 +units=m +no_defs'"
)
```

### Getting PROJ.4 Strings from EPSG Codes

If you need to convert EPSG codes to PROJ.4 strings in R:

```r
library(sf)

# Get PROJ.4 string for EPSG:3005
st_crs(3005)$proj4string
# Returns: "+proj=aea +lat_0=45 +lon_0=-126 +lat_1=50 +lat_2=58.5 +x_0=1000000 +y_0=0 +datum=NAD83 +units=m +no_defs"
```

## Why This Happens

1. Modern `mapshaper` (command-line version with Node.js) has access to the full PROJ database
2. The bundled V8/browser version uses `mproj`, a minimal JavaScript port
3. `mproj` intentionally excludes the large EPSG definition database to keep file size small
4. This makes EPSG codes unsupported in `rmapshaper` when using the bundled JavaScript (`sys = FALSE`)

## Workaround Options

### Option 1: Use Full PROJ.4 Strings (Works Now)

```r
apply_mapshaper_commands(geo, "-proj '<full PROJ.4 string>'")
```

### Option 2: Use System mapshaper (if installed)

If you have mapshaper installed via npm, use `sys = TRUE`:

```r
# System mapshaper has full PROJ support
apply_mapshaper_commands(geo, "-proj EPSG:3005 init=EPSG:4326", sys = TRUE)
```

### Option 3: Helper Function (Recommended)

Create a helper that converts EPSG codes to PROJ.4 strings:

```r
epsg_to_proj4 <- function(epsg_code) {
  sf::st_crs(epsg_code)$proj4string
}

# Usage:
target_crs <- epsg_to_proj4(3005)
apply_mapshaper_commands(geo, paste0("-proj '", target_crs, "'"))
```

## Recommendation for Package

Consider adding:
1. Documentation warning about EPSG code support
2. A wrapper function that auto-converts EPSG codes to PROJ.4 strings
3. Or recommend users install system mapshaper for projection operations

## Test Script

See `test_proj_formats.R` for the full test suite demonstrating all format attempts.

---
title: "rmapshaper Basics"
author: "Andy Teucher"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{rmapshaper Basics}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

rmapshaper is a package which is an R wrapper around the awesome [mapshaper](https://github.com/mbloch/mapshaper/) tool by Matthew Bloch, which has both a [Node.js command-line tool](https://github.com/mbloch/mapshaper/wiki/Introduction-to-the-Command-Line-Tool) as well as an [interactive web tool](https://mapshaper.org/).

The main advantage of the package is the availability of the topologically-aware simplification algorithm in `ms_simplify` (provided by the `simplify` tool in mapshaper). This means that shared boundaries between adjacent polygons are always kept intact, with no gaps or overlaps, even at high levels of simplification. It uses the  [Visvalingam](https://bost.ocks.org/mike/simplify/) simplification method.

At this time, `rmapshaper` provides the following functions:

- `ms_simplify` - simplify polygons or lines
- `ms_clip` - clip an area out of a layer using a polygon layer or a bounding box. Works on polygons, lines, and points
- `ms_erase` - erase an area from a layer using a polygon layer or a bounding box. Works on polygons, lines, and points
- `ms_dissolve` - aggregate polygon features, optionally specifying a field to aggregate on. If no field is specified, will merge all polygons into one.
- `ms_explode` - convert multipart shapes to single part. Works with polygons, lines, and points.
- `ms_lines` - convert polygons to topological boundaries (lines)
- `ms_innerlines` - convert polygons to shared inner boundaries (lines)
- `ms_points` - create points from a polygon layer
- `ms_filter_fields` - Remove fields from the attributes
- `ms_filter_islands` - Remove small detached polygons

This short vignette focuses on simplifying polygons with the `ms_simplify` function.

### Usage

rmapshaper works with `sf` objects as well as geojson strings (character objects of class `geo_json`). It also works with `Spatial` classes from the `sp` package, though this will likely be retired in the future; users are encouraged to use the more modern `sf` package.

We will use the `nc.gpkg` file (North Carolina county boundaries)
from the `sf` package and read it in as an `sf` object:

```{r}
library(rmapshaper)
library(sf)

file <- system.file("gpkg/nc.gpkg", package = "sf")
nc_sf <- read_sf(file)
```

Plot the original: 

```{r}
plot(nc_sf["FIPS"])
```

Now simplify using default parameters, then plot the simplified North Carolina counties:
```{r}
nc_simp <- ms_simplify(nc_sf)
plot(nc_simp["FIPS"])
```

You can see that even at very high levels of simplification, the mapshaper
simplification algorithm preserves the topology, including shared boundaries. The `keep` 
parameter specifies what proportion of vertices to keep:

```{r}
nc_very_simp <- ms_simplify(nc_sf, keep = 0.001)
plot(nc_very_simp["FIPS"])
```

Compare this to the output using `sf::st_simplify`, where overlaps and gaps are evident:

```{r}

nc_stsimp <- st_simplify(nc_sf, preserveTopology = TRUE, dTolerance = 10000) # dTolerance specified in meters
plot(nc_stsimp["FIPS"])
```

This time we'll demonstrate the `ms_innerlines` function:

```{r}
nc_sf_innerlines <- ms_innerlines(nc_sf)
plot(nc_sf_innerlines)
```

All of the functions are quite fast with `geojson` character objects. They are slower with the `sf` and 
`Spatial` classes due to internal conversion to/from json. If you are going to do multiple 
operations on large `sf` objects, 
it's recommended to first convert to json using `geojsonsf::sf_geojson()`, or `geojsonio::geojson_json()`. 
All of the functions have the input object as the first argument, 
and return the same class of object as the input. As such, they can be chained together. 
For a totally contrived example, using `nc_sf` as created above:

```{r eval=rmapshaper:::v8_version() >= '6'}
library(geojsonsf)
library(rmapshaper)
library(sf)

## First convert 'states' dataframe from geojsonsf pkg to json

nc_sf |> 
  sf_geojson() |> 
  ms_erase(bbox = c(-80, 35, -79, 35.5)) |>  # Cut a big hole in the middle
  ms_dissolve() |>  # Dissolve county borders
  ms_simplify(keep_shapes = TRUE, explode = TRUE) |> # Simplify polygon
  geojson_sf() |> # Convert to sf object
  plot(col = "blue") # plot
```

### Using the system mapshaper

Sometimes if you are dealing with a very large spatial object in R, `rmapshaper`
functions will take a very long time or not work at all. As of version `0.4.0`, 
you can make use of the system `mapshaper` library if you have it installed.
This will allow you to work with very large spatial objects.

First make sure you have mapshaper installed:

```{r eval=nzchar(Sys.which("mapshaper"))}
check_sys_mapshaper()
```

If you get an error, you will need to install mapshaper. First install node 
(https://nodejs.org/en) and then install mapshaper in a command prompt with:

```
$ npm install -g mapshaper
```

Then you can use the `sys` argument in any rmapshaper function:

```{r eval=nzchar(Sys.which("mapshaper"))}
nc_simp_sys <- ms_simplify(nc_sf, sys = TRUE)

plot(nc_simp_sys[, "FIPS"])
```

# NA values dealt with in sf_to_GeoJSON and GeoJSON_to_sf

    {
      "type": "character",
      "attributes": {
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["geojson", "json"]
        },
        "crs": {
          "type": "list",
          "attributes": {
            "names": {
              "type": "character",
              "attributes": {},
              "value": ["input", "wkt"]
            },
            "class": {
              "type": "character",
              "attributes": {},
              "value": ["crs"]
            }
          },
          "value": [
            {
              "type": "character",
              "attributes": {},
              "value": [null]
            },
            {
              "type": "character",
              "attributes": {},
              "value": [null]
            }
          ]
        }
      },
      "value": ["{\"type\":\"FeatureCollection\",\"features\":[{\"type\":\"Feature\",\"properties\":{\"a\":1.0},\"geometry\":{\"type\":\"Point\",\"coordinates\":[0.0,0.0]}},{\"type\":\"Feature\",\"properties\":{\"a\":null},\"geometry\":{\"type\":\"Point\",\"coordinates\":[1.0,1.0]}}]}"]
    }

# ms_innerlines works with all classes

    {
      "type": "character",
      "attributes": {
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["geojson", "json"]
        }
      },
      "value": ["{\"type\":\"FeatureCollection\", \"features\": [\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"LineString\",\"coordinates\":[[103,3],[103,2]]},\"properties\":null},\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"LineString\",\"coordinates\":[[103,2],[102,2]]},\"properties\":null},\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"LineString\",\"coordinates\":[[104,2],[103,2]]},\"properties\":null},\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"LineString\",\"coordinates\":[[103,2],[103,1]]},\"properties\":null}\n]}"]
    }

# ms_innerlines works with sys = TRUE

    {
      "type": "character",
      "attributes": {
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["geojson", "json"]
        }
      },
      "value": ["{\"type\":\"FeatureCollection\", \"features\": [\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"LineString\",\"coordinates\":[[103,3],[103,2]]},\"properties\":null},\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"LineString\",\"coordinates\":[[103,2],[102,2]]},\"properties\":null},\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"LineString\",\"coordinates\":[[104,2],[103,2]]},\"properties\":null},\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"LineString\",\"coordinates\":[[103,2],[103,1]]},\"properties\":null}\n]}"]
    }

---

    {
      "type": "character",
      "attributes": {
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["geojson", "json"]
        }
      },
      "value": ["{\"type\":\"GeometryCollection\", \"geometries\": [\n{\"type\":\"LineString\",\"coordinates\":[[103,3],[103,2]]},\n{\"type\":\"LineString\",\"coordinates\":[[103,2],[102,2]]},\n{\"type\":\"LineString\",\"coordinates\":[[104,2],[103,2]]},\n{\"type\":\"LineString\",\"coordinates\":[[103,2],[103,1]]}\n]}"]
    }

# ms_filter_islands works with min_area

    {
      "type": "character",
      "attributes": {
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["geojson", "json"]
        }
      },
      "value": ["{\"type\":\"FeatureCollection\", \"features\": [\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"Polygon\",\"coordinates\":[[[102,2],[104,2],[104,4],[102,4],[102,2]]]},\"properties\":null},\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"Polygon\",\"coordinates\":[[[100,2],[101.5,4],[98,4],[100,2]]]},\"properties\":null}\n]}"]
    }

# ms_filter_islands works with min_vertoces

    {
      "type": "character",
      "attributes": {
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["geojson", "json"]
        }
      },
      "value": ["{\"type\":\"FeatureCollection\", \"features\": [\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"Polygon\",\"coordinates\":[[[102,2],[104,2],[104,4],[102,4],[102,2]]]},\"properties\":null},\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"Polygon\",\"coordinates\":[[[100,0],[101,0],[101,1],[100,1],[100,0]]]},\"properties\":null}\n]}"]
    }

# ms_filter_islands works drop_null_geometries = FALSE

    {
      "type": "character",
      "attributes": {
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["geojson", "json"]
        }
      },
      "value": ["{\"type\":\"FeatureCollection\", \"features\": [\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"Polygon\",\"coordinates\":[[[102,2],[104,2],[104,4],[102,4],[102,2]]]},\"properties\":null},\n{\"type\":\"Feature\",\"geometry\":null,\"properties\":null},\n{\"type\":\"Feature\",\"geometry\":null,\"properties\":null}\n]}"]
    }

# ms_filter_islands works with sys = TRUE

    {
      "type": "character",
      "attributes": {
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["geojson", "json"]
        }
      },
      "value": ["{\"type\":\"FeatureCollection\", \"features\": [\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"Polygon\",\"coordinates\":[[[102,2],[104,2],[104,4],[102,4],[102,2]]]},\"properties\":null},\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"Polygon\",\"coordinates\":[[[100,2],[101.5,4],[98,4],[100,2]]]},\"properties\":null},\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"Polygon\",\"coordinates\":[[[100,0],[101,0],[101,1],[100,1],[100,0]]]},\"properties\":null}\n]}"]
    }

# ms_clip.geojson works

    {
      "type": "character",
      "attributes": {
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["geojson", "json"]
        }
      },
      "value": ["{\"type\":\"FeatureCollection\", \"features\": [\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"Polygon\",\"coordinates\":[[[53,-42],[53,-45],[55,-45],[55,-42],[53,-42]]]},\"properties\":null}\n]}"]
    }

---

    {
      "type": "character",
      "attributes": {
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["geojson", "json"]
        }
      },
      "value": ["{\"type\":\"GeometryCollection\", \"geometries\": [\n{\"type\":\"Polygon\",\"coordinates\":[[[53,-42],[53,-45],[55,-45],[55,-42],[53,-42]]]}\n]}"]
    }

---

    {
      "type": "character",
      "attributes": {
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["geojson", "json"]
        }
      },
      "value": ["{\"type\":\"FeatureCollection\", \"features\": [\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"Polygon\",\"coordinates\":[[[53,-42],[53,-45],[55,-45],[55,-42],[53,-42]]]},\"properties\":null}\n]}"]
    }

---

    {
      "type": "character",
      "attributes": {
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["geojson", "json"]
        }
      },
      "value": ["{\"type\":\"GeometryCollection\", \"geometries\": [\n{\"type\":\"Polygon\",\"coordinates\":[[[53,-42],[53,-45],[55,-45],[55,-42],[53,-42]]]}\n]}"]
    }

# ms_clip.character works

    {
      "type": "character",
      "attributes": {
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["geojson", "json"]
        }
      },
      "value": ["{\"type\":\"FeatureCollection\", \"features\": [\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"Polygon\",\"coordinates\":[[[53,-42],[53,-45],[55,-45],[55,-42],[53,-42]]]},\"properties\":null}\n]}"]
    }

# ms_erase.geojson works

    {
      "type": "character",
      "attributes": {
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["geojson", "json"]
        }
      },
      "value": ["{\"type\":\"FeatureCollection\", \"features\": [\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"Polygon\",\"coordinates\":[[[55,-42],[55,-45],[53,-45],[53,-47],[57,-47],[57,-42],[55,-42]]]},\"properties\":null}\n]}"]
    }

---

    {
      "type": "character",
      "attributes": {
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["geojson", "json"]
        }
      },
      "value": ["{\"type\":\"GeometryCollection\", \"geometries\": [\n{\"type\":\"Polygon\",\"coordinates\":[[[55,-42],[55,-45],[53,-45],[53,-47],[57,-47],[57,-42],[55,-42]]]}\n]}"]
    }

# ms_erase.character works

    {
      "type": "character",
      "attributes": {
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["geojson", "json"]
        }
      },
      "value": ["{\"type\":\"FeatureCollection\", \"features\": [\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"Polygon\",\"coordinates\":[[[55,-42],[55,-45],[53,-45],[53,-47],[57,-47],[57,-42],[55,-42]]]},\"properties\":null}\n]}"]
    }

# ms_clip works with lines

    {
      "type": "character",
      "attributes": {
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["geojson", "json"]
        }
      },
      "value": ["{\"type\":\"FeatureCollection\", \"features\": [\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"LineString\",\"coordinates\":[[55,-40.125],[52,-42]]},\"properties\":null}\n]}"]
    }

---

    {
      "type": "character",
      "attributes": {
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["geojson", "json"]
        }
      },
      "value": ["{\"type\":\"FeatureCollection\", \"features\": [\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"LineString\",\"coordinates\":[[55,-40.125],[52,-42]]},\"properties\":null}\n]}"]
    }

# ms_erase works with lines

    {
      "type": "character",
      "attributes": {
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["geojson", "json"]
        }
      },
      "value": ["{\"type\":\"FeatureCollection\", \"features\": [\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"LineString\",\"coordinates\":[[60,-37],[55,-40.125]]},\"properties\":null}\n]}"]
    }

---

    {
      "type": "character",
      "attributes": {
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["geojson", "json"]
        }
      },
      "value": ["{\"type\":\"FeatureCollection\", \"features\": [\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"LineString\",\"coordinates\":[[60,-37],[55,-40.125]]},\"properties\":null}\n]}"]
    }

# ms_clip works with points

    {
      "type": "character",
      "attributes": {
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["geojson", "json"]
        }
      },
      "value": ["{\"type\":\"FeatureCollection\", \"features\": [\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"Point\",\"coordinates\":[53,-42]},\"properties\":null}\n]}"]
    }

---

    {
      "type": "character",
      "attributes": {
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["geojson", "json"]
        }
      },
      "value": ["{\"type\":\"FeatureCollection\", \"features\": [\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"Point\",\"coordinates\":[53,-42]},\"properties\":null}\n]}"]
    }

# ms_erase works with points

    {
      "type": "character",
      "attributes": {
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["geojson", "json"]
        }
      },
      "value": ["{\"type\":\"FeatureCollection\", \"features\": [\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"Point\",\"coordinates\":[57,-42]},\"properties\":null}\n]}"]
    }

---

    {
      "type": "character",
      "attributes": {
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["geojson", "json"]
        }
      },
      "value": ["{\"type\":\"FeatureCollection\", \"features\": [\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"Point\",\"coordinates\":[57,-42]},\"properties\":null}\n]}"]
    }

# bbox works

    {
      "type": "character",
      "attributes": {
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["geojson", "json"]
        }
      },
      "value": ["{\"type\":\"FeatureCollection\", \"features\": [\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"Polygon\",\"coordinates\":[[[55,-42],[55,-45],[53,-45],[53,-47],[57,-47],[57,-42],[55,-42]]]},\"properties\":null}\n]}"]
    }

---

    {
      "type": "character",
      "attributes": {
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["geojson", "json"]
        }
      },
      "value": ["{\"type\":\"FeatureCollection\", \"features\": [\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"Polygon\",\"coordinates\":[[[53,-42],[53,-45],[55,-45],[55,-42],[53,-42]]]},\"properties\":null}\n]}"]
    }

# drop_null_geometries_works

    {
      "type": "character",
      "attributes": {
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["geojson", "json"]
        }
      },
      "value": ["{\"type\":\"FeatureCollection\", \"features\": [\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"Point\",\"coordinates\":[1,1]},\"properties\":{\"x\":1}}\n]}"]
    }

# ms_explode.geojson works

    {
      "type": "character",
      "attributes": {
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["geojson", "json"]
        }
      },
      "value": ["{\"type\":\"FeatureCollection\", \"features\": [\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"Polygon\",\"coordinates\":[[[102,2],[103,2],[103,3],[102,3],[102,2]]]},\"properties\":null},\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"Polygon\",\"coordinates\":[[[100,0],[101,0],[101,1],[100,1],[100,0]]]},\"properties\":null}\n]}"]
    }

# ms_explode.character works

    {
      "type": "character",
      "attributes": {
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["geojson", "json"]
        }
      },
      "value": ["{\"type\":\"FeatureCollection\", \"features\": [\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"Polygon\",\"coordinates\":[[[102,2],[103,2],[103,3],[102,3],[102,2]]]},\"properties\":null},\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"Polygon\",\"coordinates\":[[[100,0],[101,0],[101,1],[100,1],[100,0]]]},\"properties\":null}\n]}"]
    }

# ms_explode works with lines

    {
      "type": "character",
      "attributes": {
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["geojson", "json"]
        }
      },
      "value": ["{\"type\":\"FeatureCollection\", \"features\": [\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"LineString\",\"coordinates\":[[102,2],[103,2],[103,3],[102,3]]},\"properties\":null},\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"LineString\",\"coordinates\":[[100,0],[101,0],[101,1],[100,1]]},\"properties\":null}\n]}"]
    }

# ms_explode works with points

    {
      "type": "character",
      "attributes": {
        "class": {
          "type": "character",
          "attributes": {},
          "value": ["geojson", "json"]
        }
      },
      "value": ["{\"type\":\"FeatureCollection\", \"features\": [\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"Point\",\"coordinates\":[100,0]},\"properties\":null},\n{\"type\":\"Feature\",\"geometry\":{\"type\":\"Point\",\"coordinates\":[101,1]},\"properties\":null}\n]}"]
    }

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/points.R
\name{ms_points}
\alias{ms_points}
\title{Create points from a polygon layer}
\usage{
ms_points(input, location = NULL, x = NULL, y = NULL, ...)
}
\arguments{
\item{input}{input polygons object to convert to points. One of:
\itemize{
\item \code{geo_json} or \code{character} polygons;
\item \code{SpatialPolygons*};
\item \code{sf} or \code{sfc} polygons object
}}

\item{location}{either \code{"centroid"} or \code{"inner"}. If
\code{"centroid"}, creates points at the centroid of the largest ring of
each polygon feature. if \code{"inner"}, creates points in the interior of
the largest ring of each polygon feature. Inner points are located away
from polygon boundaries. Must be \code{NULL} if \code{x} and \code{y} are
specified. If left as \code{NULL} (default), will use centroids.}

\item{x}{name of field containing x coordinate values. Must be \code{NULL} if
\code{location} is specified.}

\item{y}{name of field containing y coordinate values. Must be \code{NULL} if
\code{location} is specified.}

\item{...}{
  Arguments passed on to \code{\link{apply_mapshaper_commands}}
  \describe{
    \item{\code{force_FC}}{should the output be forced to be a FeatureCollection (or sf
object or Spatial*DataFrame) even if there are no attributes? Default
\code{TRUE}. If FALSE and there are no attributes associated with the
geometries, a GeometryCollection (or Spatial object with no dataframe, or
sfc) will be output.}
    \item{\code{sys}}{Should the system mapshaper be used instead of the bundled
mapshaper? Gives better performance on large files. Requires the mapshaper
node package to be installed and on the PATH.}
    \item{\code{sys_mem}}{How much memory (in GB) should be allocated if using the
system mapshaper (\code{sys = TRUE})? Default 8. Ignored if \code{sys = FALSE}. This
can also be set globally with the option \code{"mapshaper.sys_mem"}}
    \item{\code{quiet}}{If \code{sys = TRUE}, should the mapshaper messages be silenced?
Default \code{FALSE}. This can also be set globally with the option
\code{"mapshaper.sys_quiet"}}
    \item{\code{gj2008}}{Generate output that is consistent with the pre-RFC 7946
GeoJSON spec (dating to 2008). Polygon rings are CW and holes are CCW,
which is the opposite of the default RFC 7946-compatible output. This should
be rarely needed, though may be useful when preparing data for D3-based
data visualizations (such as \code{plotly::plot_ly()}). Default \code{FALSE}}
  }}
}
\value{
points in the same class as the input.
}
\description{
Can be generated from the polygons by specifying \code{location} to be
\code{"centroid"} or \code{"inner"}, OR by specifying fields in the
attributes of the layer containing \code{x} and \code{y} coordinates.
}
\examples{
library(geojsonsf)
library(sf)

poly <- structure("{\"type\":\"FeatureCollection\",
           \"features\":[{\"type\":\"Feature\",\"properties\":
           {\"x_pos\": 1, \"y_pos\": 2},
           \"geometry\":{\"type\":\"Polygon\",
           \"coordinates\":[[[102,2],[102,4],[104,4],[104,2],[102,2]]]}},
           {\"type\":\"Feature\",\"properties\":{\"x_pos\": 3, \"y_pos\": 4},
           \"geometry\":{\"type\":\"Polygon\",
           \"coordinates\":[[[100,2],[98,4],[101.5,4],[100,2]]]}},
           {\"type\":\"Feature\",\"properties\":{\"x_pos\": 5, \"y_pos\": 6},
           \"geometry\":{\"type\":\"Polygon\",
           \"coordinates\":[[[100,0],[100,1],[101,1],[101,0],[100,0]]]}}]}",
           class = c("geojson", "json"))

poly <- geojson_sf(poly)
summary(poly)
plot(poly)

# Convert to points using centroids
out <- ms_points(poly, location = "centroid")
summary(out)
plot(out)

# Can also specify locations using attributes in the data
out <- ms_points(poly, x = "x_pos", y = "y_pos")
summary(out)
plot(out)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dissolve.R
\name{ms_dissolve}
\alias{ms_dissolve}
\title{Aggregate shapes in a polygon or point layer.}
\usage{
ms_dissolve(
  input,
  field = NULL,
  sum_fields = NULL,
  copy_fields = NULL,
  weight = NULL,
  snap = TRUE,
  ...
)
}
\arguments{
\item{input}{spatial object to dissolve. One of:
\itemize{
\item \code{geo_json} or \code{character} points or polygons;
\item \code{SpatialPolygons}, or \code{SpatialPoints}
}}

\item{field}{the field to dissolve on}

\item{sum_fields}{fields to sum}

\item{copy_fields}{fields to copy. The first instance of each field will be
copied to the aggregated feature.}

\item{weight}{Name of an attribute field for generating weighted centroids (points only).}

\item{snap}{Snap together vertices within a small distance threshold to fix
small coordinate misalignment in adjacent polygons. Default \code{TRUE}.}

\item{...}{
  Arguments passed on to \code{\link{apply_mapshaper_commands}}
  \describe{
    \item{\code{force_FC}}{should the output be forced to be a FeatureCollection (or sf
object or Spatial*DataFrame) even if there are no attributes? Default
\code{TRUE}. If FALSE and there are no attributes associated with the
geometries, a GeometryCollection (or Spatial object with no dataframe, or
sfc) will be output.}
    \item{\code{sys}}{Should the system mapshaper be used instead of the bundled
mapshaper? Gives better performance on large files. Requires the mapshaper
node package to be installed and on the PATH.}
    \item{\code{sys_mem}}{How much memory (in GB) should be allocated if using the
system mapshaper (\code{sys = TRUE})? Default 8. Ignored if \code{sys = FALSE}. This
can also be set globally with the option \code{"mapshaper.sys_mem"}}
    \item{\code{quiet}}{If \code{sys = TRUE}, should the mapshaper messages be silenced?
Default \code{FALSE}. This can also be set globally with the option
\code{"mapshaper.sys_quiet"}}
    \item{\code{gj2008}}{Generate output that is consistent with the pre-RFC 7946
GeoJSON spec (dating to 2008). Polygon rings are CW and holes are CCW,
which is the opposite of the default RFC 7946-compatible output. This should
be rarely needed, though may be useful when preparing data for D3-based
data visualizations (such as \code{plotly::plot_ly()}). Default \code{FALSE}}
  }}
}
\value{
the same class as the input
}
\description{
Aggregates using specified field, or all shapes if no field is given. For point layers,
replaces a group of points with their centroid.
}
\examples{
library(geojsonsf)
library(sf)

poly <- structure('{"type":"FeatureCollection",
  "features":[
  {"type":"Feature",
  "properties":{"a": 1, "b": 2},
  "geometry":{"type":"Polygon","coordinates":[[
  [102,2],[102,3],[103,3],[103,2],[102,2]
  ]]}}
  ,{"type":"Feature",
  "properties":{"a": 5, "b": 3},
  "geometry":{"type":"Polygon","coordinates":[[
  [100,0],[100,1],[101,1],[101,0],[100,0]
  ]]}}]}', class = c("geojson", "json"))
poly <- geojson_sf(poly)
plot(poly)
length(poly)
poly

# Dissolve the polygon
out <- ms_dissolve(poly)
plot(out)
length(out)
out

# Dissolve and summing columns
out <- ms_dissolve(poly, sum_fields = c("a", "b"))
plot(out)
out

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{check_sys_mapshaper}
\alias{check_sys_mapshaper}
\title{Check the system mapshaper}
\usage{
check_sys_mapshaper(command = "mapshaper-xl", verbose = TRUE)
}
\arguments{
\item{command}{either "mapshaper-xl" (default) or "mapshaper"}

\item{verbose}{Print a message stating mapshaper's current version? Default \code{TRUE}}
}
\value{
character path to mapshaper executable if appropriate version is installed, otherwise throws an error
}
\description{
Check the system mapshaper
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/explode.R
\name{ms_explode}
\alias{ms_explode}
\title{Convert multipart lines or polygons to singlepart}
\usage{
ms_explode(input, ...)
}
\arguments{
\item{input}{One of:
\itemize{
\item \code{geo_json} or \code{character} multipart lines, or polygons;
\item multipart \code{SpatialPolygons}, \code{SpatialLines};
\item \code{sf} or \code{sfc} multipart lines, or polygons object
}}

\item{...}{
  Arguments passed on to \code{\link{apply_mapshaper_commands}}
  \describe{
    \item{\code{force_FC}}{should the output be forced to be a FeatureCollection (or sf
object or Spatial*DataFrame) even if there are no attributes? Default
\code{TRUE}. If FALSE and there are no attributes associated with the
geometries, a GeometryCollection (or Spatial object with no dataframe, or
sfc) will be output.}
    \item{\code{sys}}{Should the system mapshaper be used instead of the bundled
mapshaper? Gives better performance on large files. Requires the mapshaper
node package to be installed and on the PATH.}
    \item{\code{sys_mem}}{How much memory (in GB) should be allocated if using the
system mapshaper (\code{sys = TRUE})? Default 8. Ignored if \code{sys = FALSE}. This
can also be set globally with the option \code{"mapshaper.sys_mem"}}
    \item{\code{quiet}}{If \code{sys = TRUE}, should the mapshaper messages be silenced?
Default \code{FALSE}. This can also be set globally with the option
\code{"mapshaper.sys_quiet"}}
    \item{\code{gj2008}}{Generate output that is consistent with the pre-RFC 7946
GeoJSON spec (dating to 2008). Polygon rings are CW and holes are CCW,
which is the opposite of the default RFC 7946-compatible output. This should
be rarely needed, though may be useful when preparing data for D3-based
data visualizations (such as \code{plotly::plot_ly()}). Default \code{FALSE}}
  }}
}
\value{
same class as input
}
\description{
For objects of class \code{Spatial} (e.g., \code{SpatialPolygonsDataFrame}),
you may find it faster to use \code{sp::disaggregate}.
}
\details{
There is currently no method for SpatialMultiPoints
}
\examples{
library(geojsonsf)
library(sf)

poly <- "{\"type\":\"FeatureCollection\",\"features\":
          [\n{\"type\":\"Feature\",\"geometry\":{\"type\":
          \"MultiPolygon\",\"coordinates\":[[[[102,2],[102,3],
          [103,3],[103,2],[102,2]]],[[[100,0],[100,1],[101,1],
          [101,0],[100,0]]]]},\"properties\":{\"a\":0}}\n]}"

poly <- geojson_sf(poly)
plot(poly)
length(poly)
poly

# Explode the polygon
out <- ms_explode(poly)
plot(out)
length(out)
out

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter_islands.R
\name{ms_filter_islands}
\alias{ms_filter_islands}
\title{Remove small detached polygons (islands)}
\usage{
ms_filter_islands(
  input,
  min_area = NULL,
  min_vertices = NULL,
  drop_null_geometries = TRUE,
  ...
)
}
\arguments{
\item{input}{spatial object to filter. One of:
\itemize{
\item \code{geo_json} or \code{character} polygons;
\item \code{SpatialPolygons*};
\item \code{sf} or \code{sfc} polygons object
}}

\item{min_area}{minimum area of polygons to retain. Area is calculated using
planar geometry, except for the area of unprojected polygons, which is
calculated using spherical geometry in units of square meters.}

\item{min_vertices}{minimum number of vertices in polygons to retain.}

\item{drop_null_geometries}{should features with empty geometries be dropped?
Default \code{TRUE}. Ignored for \code{SpatialPolyons*}, as it is always
\code{TRUE}.}

\item{...}{
  Arguments passed on to \code{\link{apply_mapshaper_commands}}
  \describe{
    \item{\code{force_FC}}{should the output be forced to be a FeatureCollection (or sf
object or Spatial*DataFrame) even if there are no attributes? Default
\code{TRUE}. If FALSE and there are no attributes associated with the
geometries, a GeometryCollection (or Spatial object with no dataframe, or
sfc) will be output.}
    \item{\code{sys}}{Should the system mapshaper be used instead of the bundled
mapshaper? Gives better performance on large files. Requires the mapshaper
node package to be installed and on the PATH.}
    \item{\code{sys_mem}}{How much memory (in GB) should be allocated if using the
system mapshaper (\code{sys = TRUE})? Default 8. Ignored if \code{sys = FALSE}. This
can also be set globally with the option \code{"mapshaper.sys_mem"}}
    \item{\code{quiet}}{If \code{sys = TRUE}, should the mapshaper messages be silenced?
Default \code{FALSE}. This can also be set globally with the option
\code{"mapshaper.sys_quiet"}}
    \item{\code{gj2008}}{Generate output that is consistent with the pre-RFC 7946
GeoJSON spec (dating to 2008). Polygon rings are CW and holes are CCW,
which is the opposite of the default RFC 7946-compatible output. This should
be rarely needed, though may be useful when preparing data for D3-based
data visualizations (such as \code{plotly::plot_ly()}). Default \code{FALSE}}
  }}
}
\value{
object with only specified features retained, in the same class as
the input
}
\description{
Remove small detached polygons, keeping those with a minimum area and/or a
minimum number of vertices. Optionally remove null geometries.
}
\examples{
library(geojsonsf)
library(sf)

poly <- structure("{\"type\":\"FeatureCollection\",
           \"features\":[{\"type\":\"Feature\",\"properties\":{},
           \"geometry\":{\"type\":\"Polygon\",
           \"coordinates\":[[[102,2],[102,4],[104,4],[104,2],[102,2]]]}},
           {\"type\":\"Feature\",\"properties\":{},
           \"geometry\":{\"type\":\"Polygon\",
           \"coordinates\":[[[100,2],[98,4],[101.5,4],[100,2]]]}},
           {\"type\":\"Feature\",\"properties\":{},
           \"geometry\":{\"type\":\"Polygon\",
           \"coordinates\":[[[100,0],[100,1],[101,1],[101,0],[100,0]]]}}]}",
           class = c("geojson", "json"))

poly <- geojson_sf(poly)
plot(poly)

out <- ms_filter_islands(poly, min_area = 12391399903)
plot(out)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{apply_mapshaper_commands}
\alias{apply_mapshaper_commands}
\title{Apply a mapshaper command string to a geojson object}
\usage{
apply_mapshaper_commands(
  data,
  command,
  force_FC = TRUE,
  sys = FALSE,
  sys_mem = getOption("mapshaper.sys_mem", default = 8),
  quiet = getOption("mapshaper.sys_quiet", default = FALSE),
  gj2008 = FALSE
)
}
\arguments{
\item{data}{character containing geojson or path to geojson file. If a file
path, \code{sys} must be true.}

\item{command}{valid mapshaper command string}

\item{force_FC}{should the output be forced to be a FeatureCollection (or sf
object or Spatial*DataFrame) even if there are no attributes? Default
\code{TRUE}. If FALSE and there are no attributes associated with the
geometries, a GeometryCollection (or Spatial object with no dataframe, or
sfc) will be output.}

\item{sys}{Should the system mapshaper be used instead of the bundled
mapshaper? Gives better performance on large files. Requires the mapshaper
node package to be installed and on the PATH.}

\item{sys_mem}{How much memory (in GB) should be allocated if using the
system mapshaper (\code{sys = TRUE})? Default 8. Ignored if \code{sys = FALSE}. This
can also be set globally with the option \code{"mapshaper.sys_mem"}}

\item{quiet}{If \code{sys = TRUE}, should the mapshaper messages be silenced?
Default \code{FALSE}. This can also be set globally with the option
\code{"mapshaper.sys_quiet"}}

\item{gj2008}{Generate output that is consistent with the pre-RFC 7946
GeoJSON spec (dating to 2008). Polygon rings are CW and holes are CCW,
which is the opposite of the default RFC 7946-compatible output. This should
be rarely needed, though may be useful when preparing data for D3-based
data visualizations (such as \code{plotly::plot_ly()}). Default \code{FALSE}}
}
\value{
geojson
}
\description{
Apply a mapshaper command string to a geojson object
}
\examples{

nc <- sf::read_sf(system.file("gpkg/nc.gpkg", package = "sf"))
rmapshaper::apply_mapshaper_commands(geojsonsf::sf_geojson(nc), "-clean")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter_fields.R
\name{ms_filter_fields}
\alias{ms_filter_fields}
\title{Delete fields in the attribute table}
\usage{
ms_filter_fields(input, fields, ...)
}
\arguments{
\item{input}{spatial object to filter fields on. One of:
\itemize{
\item \code{geo_json} or \code{character} points, lines, or polygons;
\item \code{SpatialPolygonsDataFrame}, \code{SpatialLinesDataFrame}, \code{SpatialPointsDataFrame};
\item \code{sf} object
}}

\item{fields}{character vector of fields to retain.}

\item{...}{
  Arguments passed on to \code{\link{apply_mapshaper_commands}}
  \describe{
    \item{\code{sys}}{Should the system mapshaper be used instead of the bundled
mapshaper? Gives better performance on large files. Requires the mapshaper
node package to be installed and on the PATH.}
    \item{\code{sys_mem}}{How much memory (in GB) should be allocated if using the
system mapshaper (\code{sys = TRUE})? Default 8. Ignored if \code{sys = FALSE}. This
can also be set globally with the option \code{"mapshaper.sys_mem"}}
    \item{\code{quiet}}{If \code{sys = TRUE}, should the mapshaper messages be silenced?
Default \code{FALSE}. This can also be set globally with the option
\code{"mapshaper.sys_quiet"}}
  }}
}
\value{
object with only specified attributes retained, in the same class as
the input
}
\description{
Removes all fields except those listed in the \code{fields} parameter
}
\examples{
library(geojsonsf)
library(sf)

poly <- structure("{\"type\":\"FeatureCollection\",
                  \"features\":[{\"type\":\"Feature\",
                  \"properties\":{\"a\": 1, \"b\":2, \"c\": 3},
                  \"geometry\":{\"type\":\"Polygon\",
                  \"coordinates\":[[[102,2],[102,4],[104,4],[104,2],[102,2]]]}}]}",
                  class = c("geojson", "json"))
poly <- geojson_sf(poly)
poly

# Filter (keep) fields a and b, drop c
out <- ms_filter_fields(poly, c("a", "b"))
out

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/inner_lines.R
\name{ms_innerlines}
\alias{ms_innerlines}
\title{Create a line layer consisting of shared boundaries with no attribute data}
\usage{
ms_innerlines(input, ...)
}
\arguments{
\item{input}{input polygons object to convert to inner lines. One of:
\itemize{
\item \code{geo_json} or \code{character} polygons;
\item \code{SpatialPolygons*};
\item \code{sf} or \code{sfc} polygons object
}}

\item{...}{
  Arguments passed on to \code{\link{apply_mapshaper_commands}}
  \describe{
    \item{\code{force_FC}}{should the output be forced to be a FeatureCollection (or sf
object or Spatial*DataFrame) even if there are no attributes? Default
\code{TRUE}. If FALSE and there are no attributes associated with the
geometries, a GeometryCollection (or Spatial object with no dataframe, or
sfc) will be output.}
    \item{\code{sys}}{Should the system mapshaper be used instead of the bundled
mapshaper? Gives better performance on large files. Requires the mapshaper
node package to be installed and on the PATH.}
    \item{\code{sys_mem}}{How much memory (in GB) should be allocated if using the
system mapshaper (\code{sys = TRUE})? Default 8. Ignored if \code{sys = FALSE}. This
can also be set globally with the option \code{"mapshaper.sys_mem"}}
    \item{\code{quiet}}{If \code{sys = TRUE}, should the mapshaper messages be silenced?
Default \code{FALSE}. This can also be set globally with the option
\code{"mapshaper.sys_quiet"}}
    \item{\code{gj2008}}{Generate output that is consistent with the pre-RFC 7946
GeoJSON spec (dating to 2008). Polygon rings are CW and holes are CCW,
which is the opposite of the default RFC 7946-compatible output. This should
be rarely needed, though may be useful when preparing data for D3-based
data visualizations (such as \code{plotly::plot_ly()}). Default \code{FALSE}}
  }}
}
\value{
lines in the same class as the input layer, but without attributes
}
\description{
Create a line layer consisting of shared boundaries with no attribute data
}
\examples{
library(geojsonsf)
library(sf)

poly <- structure('{"type":"FeatureCollection",
            "features":[
              {"type":"Feature",
                "properties":{"foo": "a"},
                "geometry":{"type":"Polygon","coordinates":[[
                  [102,2],[102,3],[103,3],[103,2],[102,2]
                  ]]}}
              ,{"type":"Feature",
                "properties":{"foo": "a"},
                "geometry":{"type":"Polygon","coordinates":[[
                  [103,3],[104,3],[104,2],[103,2],[103,3]
                  ]]}},
              {"type":"Feature",
                "properties":{"foo": "b"},
                "geometry":{"type":"Polygon","coordinates":[[
                  [102,1],[102,2],[103,2],[103,1],[102,1]
                  ]]}},
              {"type":"Feature",
                "properties":{"foo": "b"},
                "geometry":{"type":"Polygon","coordinates":[[
                  [103,1],[103,2],[104,2],[104,1],[103,1]
                  ]]}}]}', class = c("geojson", "json"))

poly <- geojson_sf(poly)
plot(poly)

out <- ms_innerlines(poly)
plot(out)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rmapshaper-package.R
\docType{package}
\name{rmapshaper-package}
\alias{rmapshaper}
\alias{rmapshaper-package}
\title{rmapshaper: Client for 'mapshaper' for 'Geospatial' Operations}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Edit and simplify 'geojson', 'Spatial', and 'sf' objects. This is wrapper around the 'mapshaper' 'JavaScript' library by Matthew Bloch \url{https://github.com/mbloch/mapshaper/} to perform topologically-aware polygon simplification, as well as other operations such as clipping, erasing, dissolving, and converting 'multi-part' to 'single-part' geometries.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/ateucher/rmapshaper}
  \item \url{http://andyteucher.ca/rmapshaper/}
  \item Report bugs at \url{https://github.com/ateucher/rmapshaper/issues}
}

}
\author{
\strong{Maintainer}: Andy Teucher \email{andy.teucher@gmail.com} (\href{https://orcid.org/0000-0002-7840-692X}{ORCID})

Authors:
\itemize{
  \item Andy Teucher \email{andy.teucher@gmail.com} (\href{https://orcid.org/0000-0002-7840-692X}{ORCID})
  \item Kenton Russell \email{kent.russell@timelyportfolio.com} (JavaScript support)
}

Other contributors:
\itemize{
  \item Matthew Bloch (mapshaper Javascript library) [copyright holder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/drop_null_geometries.R
\name{drop_null_geometries}
\alias{drop_null_geometries}
\title{Drop features from a \code{geo_json} FeatureCollection with null geometries}
\usage{
drop_null_geometries(x)
}
\arguments{
\item{x}{a \code{geo_json} FeatureCollection}
}
\value{
a \code{geo_json} FeatureCollection with Features with null geometries
removed
}
\description{
Drop features from a \code{geo_json} FeatureCollection with null geometries
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simplify.R
\name{ms_simplify}
\alias{ms_simplify}
\title{Topologically-aware geometry simplification.}
\usage{
ms_simplify(
  input,
  keep = 0.05,
  method = NULL,
  weighting = 0.7,
  keep_shapes = FALSE,
  no_repair = FALSE,
  snap = TRUE,
  explode = FALSE,
  drop_null_geometries = TRUE,
  snap_interval = NULL,
  ...
)
}
\arguments{
\item{input}{spatial object to simplify. One of:
\itemize{
\item \code{geo_json} or \code{character} polygons or lines;
\item \code{SpatialPolygons*} or \code{SpatialLines*};
\item \code{sf} or \code{sfc} polygons or lines object
}}

\item{keep}{proportion of points to retain (0-1; default 0.05)}

\item{method}{simplification method to use: \code{"vis"} for Visvalingam
algorithm, or \code{"dp"} for Douglas-Peuker algorithm. If left as
\code{NULL} (default), uses Visvalingam simplification but modifies the
area metric by underweighting the effective area of points at the vertex of
more acute angles, resulting in a smoother appearance. See this
\href{https://github.com/mbloch/mapshaper/wiki/Simplification-Tips}{link}
for more information.}

\item{weighting}{Coefficient for weighting Visvalingam simplification
(default is 0.7). Higher values produce smoother output. weighting=0 is
equivalent to unweighted Visvalingam simplification.}

\item{keep_shapes}{Prevent small polygon features from disappearing at high
simplification (default \code{FALSE})}

\item{no_repair}{disable intersection repair after simplification (default
\code{FALSE}).}

\item{snap}{Snap together vertices within a small distance threshold to fix
small coordinate misalignment in adjacent polygons. Default \code{TRUE}.}

\item{explode}{Should multipart polygons be converted to singlepart polygons?
This prevents small shapes from disappearing during simplification if
\code{keep_shapes = TRUE}. Default \code{FALSE}}

\item{drop_null_geometries}{should Features with null geometries be dropped?
Ignored for \code{Spatial*} objects, as it is always \code{TRUE}.}

\item{snap_interval}{Specify snapping distance in source units, must be a
numeric. Default \code{NULL}}

\item{...}{
  Arguments passed on to \code{\link{apply_mapshaper_commands}}
  \describe{
    \item{\code{force_FC}}{should the output be forced to be a FeatureCollection (or sf
object or Spatial*DataFrame) even if there are no attributes? Default
\code{TRUE}. If FALSE and there are no attributes associated with the
geometries, a GeometryCollection (or Spatial object with no dataframe, or
sfc) will be output.}
    \item{\code{sys}}{Should the system mapshaper be used instead of the bundled
mapshaper? Gives better performance on large files. Requires the mapshaper
node package to be installed and on the PATH.}
    \item{\code{sys_mem}}{How much memory (in GB) should be allocated if using the
system mapshaper (\code{sys = TRUE})? Default 8. Ignored if \code{sys = FALSE}. This
can also be set globally with the option \code{"mapshaper.sys_mem"}}
    \item{\code{quiet}}{If \code{sys = TRUE}, should the mapshaper messages be silenced?
Default \code{FALSE}. This can also be set globally with the option
\code{"mapshaper.sys_quiet"}}
    \item{\code{gj2008}}{Generate output that is consistent with the pre-RFC 7946
GeoJSON spec (dating to 2008). Polygon rings are CW and holes are CCW,
which is the opposite of the default RFC 7946-compatible output. This should
be rarely needed, though may be useful when preparing data for D3-based
data visualizations (such as \code{plotly::plot_ly()}). Default \code{FALSE}}
  }}
}
\value{
a simplified representation of the geometry in the same class as the
input
}
\description{
Uses \href{https://github.com/mbloch/mapshaper}{mapshaper} to simplify
polygons.
}
\examples{
# With a simple geojson object
poly <- structure('{
 "type": "Feature",
 "properties": {},
 "geometry": {
   "type": "Polygon",
   "coordinates": [[
     [-70.603637, -33.399918],
     [-70.614624, -33.395332],
     [-70.639343, -33.392466],
     [-70.659942, -33.394759],
     [-70.683975, -33.404504],
     [-70.697021, -33.419406],
     [-70.701141, -33.434306],
     [-70.700454, -33.446339],
     [-70.694274, -33.458369],
     [-70.682601, -33.465816],
     [-70.668869, -33.472117],
     [-70.646209, -33.473835],
     [-70.624923, -33.472117],
     [-70.609817, -33.468107],
     [-70.595397, -33.458369],
     [-70.587158, -33.442901],
     [-70.587158, -33.426283],
     [-70.590591, -33.414248],
     [-70.594711, -33.406224],
     [-70.603637, -33.399918]
   ]]
 }
}', class = c("geojson", "json"))

ms_simplify(poly, keep = 0.1)

# With an sf object

poly_sf <- geojsonsf::geojson_sf(poly)
ms_simplify(poly_sf, keep = 0.5)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clip_erase.R
\name{ms_erase}
\alias{ms_erase}
\title{Remove features or portions of features that fall inside a specified area}
\usage{
ms_erase(target, erase = NULL, bbox = NULL, remove_slivers = FALSE, ...)
}
\arguments{
\item{target}{the target layer from which to remove portions. One of:
\itemize{
\item \code{geo_json} or \code{character} points, lines, or polygons;
\item \code{SpatialPolygons}, \code{SpatialLines}, \code{SpatialPoints}
}}

\item{erase}{the erase layer (polygon). One of:
\itemize{
\item \code{geo_json} or \code{character} polygons;
\item \code{SpatialPolygons*}
}}

\item{bbox}{supply a bounding box instead of an erasing layer to remove from
the target layer. Supply as a numeric vector: \code{c(minX, minY, maxX, maxY)}.}

\item{remove_slivers}{Remove tiny sliver polygons created by erasing. (Default \code{FALSE})}

\item{...}{
  Arguments passed on to \code{\link{apply_mapshaper_commands}}
  \describe{
    \item{\code{force_FC}}{should the output be forced to be a FeatureCollection (or sf
object or Spatial*DataFrame) even if there are no attributes? Default
\code{TRUE}. If FALSE and there are no attributes associated with the
geometries, a GeometryCollection (or Spatial object with no dataframe, or
sfc) will be output.}
    \item{\code{sys}}{Should the system mapshaper be used instead of the bundled
mapshaper? Gives better performance on large files. Requires the mapshaper
node package to be installed and on the PATH.}
    \item{\code{sys_mem}}{How much memory (in GB) should be allocated if using the
system mapshaper (\code{sys = TRUE})? Default 8. Ignored if \code{sys = FALSE}. This
can also be set globally with the option \code{"mapshaper.sys_mem"}}
    \item{\code{quiet}}{If \code{sys = TRUE}, should the mapshaper messages be silenced?
Default \code{FALSE}. This can also be set globally with the option
\code{"mapshaper.sys_quiet"}}
    \item{\code{gj2008}}{Generate output that is consistent with the pre-RFC 7946
GeoJSON spec (dating to 2008). Polygon rings are CW and holes are CCW,
which is the opposite of the default RFC 7946-compatible output. This should
be rarely needed, though may be useful when preparing data for D3-based
data visualizations (such as \code{plotly::plot_ly()}). Default \code{FALSE}}
  }}
}
\value{
erased target in the same format as the input target
}
\description{
Removes portions of the target layer that fall inside the erasing layer or bounding box.
}
\examples{
if (rmapshaper:::v8_version() >= "6") {
  library(geojsonsf, quietly = TRUE)
  library(sf)

  points <- structure("{\"type\":\"FeatureCollection\",
    \"features\":[{\"type\":\"Feature\",\"properties\":{},
    \"geometry\":{\"type\":\"Point\",\"coordinates\":
    [52.8658,-44.7219]}},{\"type\":\"Feature\",\"properties\":{},
    \"geometry\":{\"type\":\"Point\",\"coordinates\":
    [53.7702,-40.4873]}},{\"type\":\"Feature\",\"properties\":{},
    \"geometry\":{\"type\":\"Point\",\"coordinates\":[55.3204,-37.5579]}},
    {\"type\":\"Feature\",\"properties\":{},\"geometry\":
    {\"type\":\"Point\",\"coordinates\":[56.2757,-37.917]}},
    {\"type\":\"Feature\",\"properties\":{},\"geometry\":
    {\"type\":\"Point\",\"coordinates\":[56.184,-40.6443]}},
    {\"type\":\"Feature\",\"properties\":{},\"geometry\":
    {\"type\":\"Point\",\"coordinates\":[61.0835,-40.7529]}},
    {\"type\":\"Feature\",\"properties\":{},\"geometry\":
    {\"type\":\"Point\",\"coordinates\":[58.0202,-43.634]}}]}",
    class = c("geojson", "json"))
  points <- geojson_sf(points)
  plot(points)

  erase_poly <- structure('{
  "type": "Feature",
  "properties": {},
  "geometry": {
  "type": "Polygon",
  "coordinates": [
  [
  [51, -40],
  [55, -40],
  [55, -45],
  [51, -45],
  [51, -40]
  ]
  ]
  }
  }', class = c("geojson", "json"))
  erase_poly <- geojson_sf(erase_poly)

  out <- ms_erase(points, erase_poly)
  plot(out, add = TRUE)
}

}
